/*
 * File: cmd_DHCOMupdate.c
 *
 * (C) Copyright 2011-2012
 * Andreas Geisreiter, DH electronics GmbH , ageisreiter@dh-electronics.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* ------------------------------------------------------------------------------
 *
 * Abstract: Command line function to update the flash content from storage device.
 *
 * Notes:
 *     Created by: 	Andreas Geisreiter (ageisreiter@dh-electronics.de)
 *     Created:     January 03, 2012
 *
 * Description:
 *     The following command line calls are possible:
 *     1. "update auto":
 *        Is called after the start of the bootloader with the env variable "bootcmd". This function call is
 *        used for automatic update at the system start. The function tries to update the flash content via
 *        the DHupdate.ini file. The update mechanism is searching for the update files on the specified
 *        storage devices. The update devices must be defined at the settings block. With the DHupdate.ini
 *        file you have the possibility to run more then one updates, e.g. OS image and settings block.
 *
 *     2. "update":
 *        Is used form the command line. This function call tries to update the flash content via the
 *        DHupdate.ini file. In contrast to 1. the update mechanism is searching on every available storage
 *        device for the update files.
 *
 *     3. "update <type> [filename]":
 *        Is used form the command line. This function call doesn't use the DHupdate.ini file. It
 *        allows only one update and the update mechanism is searching on every available storage device for
 *        the update files. With the <type> parameter you have to specify the update type.
 *
 *     Possible Updates:
 *     Argument = '1' ==> Update only WinCE Flash Image (nk.bin or nk.gz)
 *     Argument = '2' ==> Update only Linux Flash Image
 *     Argument = '3' ==> Update Bootloader Flash Image
 *     Argument = '4' ==> Update splash.bmp (Start Screen Picture)
 *     Argument = '5' ==> Update settings.bin file
 * 	   Argument = '6' ==> Refresh DH settings and LCD controller
 *     Argument = '7' ==> Display adapter EEPROM settings update
 * 	   Argument = '8' ==> Reset Board
 * 	   Argument = '9' ==> Update rootfs in NAND-Flash
 */

#include <config.h>
#include <common.h>
#include <command.h>
#include <i2c.h>
#include <linux/ctype.h>
#include <asm/io.h>

#ifdef CONFIG_MX25
#include <asm/arch/mx25.h>
#include <asm/arch/mx25-regs.h>
#include <asm/arch/mx25_pins.h>
#include <asm/arch/iomux.h>
#endif

#include <asm/arch/gpio.h>
#include <dh_settings.h>
#include <dh_update.h>
#include <nand.h>

#if defined(CONFIG_CMD_DHCOM_UPDATE)

DECLARE_GLOBAL_DATA_PTR;

// ===============================================================================================================
// Variables
// ===============================================================================================================

// ===============================================================================================================
// Extern defined variables
// ===============================================================================================================

// ===============================================================================================================
// Extern defined functions
// ===============================================================================================================
extern int do_mmcinit (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int do_usb(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[]);
extern int do_fat_fsload (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int do_bmp (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);

#ifdef DH_NAND_FLASH    
extern int do_nand(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[]);
#else
extern int do_spi_flash(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
#endif /* DH_NAND_FLASH  */

extern int do_mmu(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int do_mem_mw (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int do_source (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int DHCOMUpdateLED_Init(updateinfo_t *p_stDHupdateINI);
extern void DHCOMUpdateDelayMs(unsigned long msec);
extern void DHCOMUpdateLED_SetHigh(void);
extern void DHCOMUpdateLED_SetLow(void);
extern int  get_DHCOMUpdateLED(void);
extern void lcd_ctrl_init (void *lcdbase);
extern void lcd_enable (void);

// ===============================================================================================================
// Functions
// ===============================================================================================================

//------------------------------------------------------------------------------
//
//  Function:  update_flash_content
//
//  Update Flash content on the specified address.
//
//  Commit values:  - ulFlashAddress    = FLash Block address
//                  - ulSDRAMBufferAddress  = SDRAM Buffer address
//                  - ulBlocks              = Number of block that should be updated
//                  - ulFlashBlockSize      = Flash block size
//
//  Return value:   0 = No error
//                  1 = Flash erase error (the error block number is specified from bit 3 to bit 31)
//                  2 = Flash write error (the error block number is specified from bit 3 to bit 31)
//
int update_flash_content (unsigned long ulFlashAddress, unsigned long ulSDRAMBufferAddress, unsigned long ulBlocks, unsigned long ulFlashBlockSize)
{
#ifdef DH_NAND_FLASH   
    char *p_cEraseFlashBlock[4]         = {"nand", "erase", "12345671", "12345672"};  
    char *p_cWriteFlashBlock[5]         = {"nand", "write", "12345673", "12345674", "12345675"};  
#else /* DH_SPI_FLASH */
    char *p_cEraseFlashBlock[4]         = {"sf", "erase", "12345671", "12345672"};  
    char *p_cWriteFlashBlock[5]         = {"sf", "write", "12345673", "12345674", "12345675"};
#endif /* DH_NAND_FLASH  */
    char *mmu_on[2]                     = {"mmu","on"};
    char *mmu_off[2]                    = {"mmu","off"};
    
    unsigned long i,j;
    int ret_value = 0;
    
    // Set FLash Block size
    sprintf (p_cEraseFlashBlock[3], "%08x", (unsigned int)ulFlashBlockSize);
    sprintf (p_cWriteFlashBlock[4], "%08x", (unsigned int)ulFlashBlockSize);     
    
    // Enable MMU
    do_mmu(NULL, 0, 2, mmu_on);
    
    for(j = 1, i = 0; i < ulBlocks; j++, i++)
    {
        if(j == 51)
        {
            printf ("\n                         ");
            j = 1;
        }
        
        // Set block address
        sprintf (p_cEraseFlashBlock[2], "%08x", (unsigned int)(ulFlashAddress + i * ulFlashBlockSize));
        sprintf (p_cWriteFlashBlock[3], "%08x", (unsigned int)(ulFlashAddress + i * ulFlashBlockSize));
        
        // Erase flash block
        printf ("e");
        
        DISABLE_PRINTF()
#ifdef DH_NAND_FLASH 
        ret_value = do_nand(NULL, 0, 4, p_cEraseFlashBlock);
#else /* DH_SPI_FLASH */
        ret_value = do_spi_flash(NULL, 0, 4, p_cEraseFlashBlock);
#endif /* DH_NAND_FLASH  */
        SESSION_DEPENDED_PRINTF_ENABLE()
        
        // Exit on Erase ERROR
        if(ret_value != 0)
        {
            do_mmu(NULL, 0, 2, mmu_off);
            ret_value = (i << 2) | 1;
            return ret_value;  
        }

        // Set SDRAM Buffer address
        sprintf (p_cWriteFlashBlock[2], "%08x", (unsigned int)(ulSDRAMBufferAddress + i * ulFlashBlockSize));
        
        // Write flash block
        printf ("\bw");
        DISABLE_PRINTF()
#ifdef DH_NAND_FLASH 
        ret_value = do_nand(NULL, 0, 5, p_cWriteFlashBlock);
#else /* DH_SPI_FLASH */
        ret_value = do_spi_flash(NULL, 0, 5, p_cWriteFlashBlock);
#endif /* DH_NAND_FLASH  */

        SESSION_DEPENDED_PRINTF_ENABLE()
        
        // Exit on Write ERROR
        if(ret_value != 0)
        {
            do_mmu(NULL, 0, 2, mmu_off);
            ret_value = (i << 2) | 2;
            return ret_value;    
        }
        
        printf ("\bd");
    }
    // Disable MMU
    do_mmu(NULL, 0, 2, mmu_off);
    
    printf ("]");      
    return 0;
}


//------------------------------------------------------------------------------
//
//  Function:  update_eeprom_content
//
//  Update EEProm content on the specified address.
//
//  Commit values:  - ulEepromRegisterAddress = Eeprom register offset
//                  - ulSDRAMBufferAddress  = SDRAM Buffer address
//                  - ulBytes               = Number of bytes that should be updated
//
//  Return value:   0 = No error
//                  1 = EEPROM write error
//
int update_eeprom_content (unsigned long ulEepromRegisterAddress, unsigned long ulSDRAMBufferAddress, unsigned long ulBytes)
{
    unsigned int i,j;
    int ret_value = 0;
    uchar ucBuffer;

#if defined(CONFIG_AM3517)
    /* Set i2c driver to use i2c bus 3  */
	i2c_set_bus_num(3);
#endif

    for(j = 1, i = 0; i < ulBytes; j++, i++)
    {
        if(j == 51)
        {
            printf ("\n                         ");
            j = 1;
        }

        // Write eeprom content
        printf ("w");
        DISABLE_PRINTF()
        ret_value = i2c_write(DISPLAY_ADAPTER_EEPROM_ADDR, i, 1, (uchar*)(ulSDRAMBufferAddress+i), 1);
        DHCOMUpdateDelayMs(8);
        SESSION_DEPENDED_PRINTF_ENABLE()

        // Exit on Write ERROR
        if(ret_value != 0)
        {

		#if defined(CONFIG_AM3517)
			/* Set i2c driver back to use i2c bus 1 */
			i2c_set_bus_num(1);
		#endif

            return 1;
        }

        printf ("\bd");
    }

    // Verify EEPROM data
    for(i = 0; i < ulBytes; i++)
    {
        DISABLE_PRINTF()
        ret_value = i2c_read(DISPLAY_ADAPTER_EEPROM_ADDR, i, 1, &ucBuffer, 1);
        SESSION_DEPENDED_PRINTF_ENABLE()

        // Exit on Write ERROR
        if((ret_value != 0) || (ucBuffer != *(uchar*)(ulSDRAMBufferAddress+i)))
        {

		#if defined(CONFIG_AM3517)
			/* Set i2c driver back to use i2c bus 1 */
			i2c_set_bus_num(1);
		#endif

            return 1;
        }
    }

    printf ("]");

#if defined(CONFIG_AM3517)
    /* Set i2c driver back to use i2c bus 1 */
    i2c_set_bus_num(1);
#endif

    return 0;
}


//------------------------------------------------------------------------------
//
//  Function:  check_imagesize
//
//  Update Flash content on the specified address.
//
//  Commit values:  - ulFilesize    		= Size of the image to write
//                  - ulFlashPartitionSize  = available partitionsize
//                  - cUpdatetype           = Identifier for updatetype (Customizing Error-Messages)
//
//  Return value:   0 = No error
//                  1 = image is to large
//
int check_imagesize (unsigned long ulFilesize, unsigned long ulFlashPartitionSize, char cUpdatetype, char* pcErrorString )
{
	if (ulFilesize > ulFlashPartitionSize)
	{
		switch (cUpdatetype)
		{
			case WINCE_IMAGE_UPDATE:
				sprintf (pcErrorString, "\n==> Update ERROR: WinCE image to large: max %li bytes\n", ulFlashPartitionSize);
			break;
			case LINUX_IMAGE_UPDATE:
				sprintf (pcErrorString, "\n==> Update ERROR: Linux kernel to large: max %li bytes\n", ulFlashPartitionSize);
			break;
			case BOOTLOADER_FLASH_UPDATE:
				sprintf (pcErrorString, "\n==> Update ERROR: u-boot image to large: max %li bytes\n", ulFlashPartitionSize);
			break;
			case SPLASHIMAGE_UPDATE:
				sprintf (pcErrorString, "\n==> Update ERROR: Splashimage to large: max %li bytes\n", ulFlashPartitionSize);
			break;
			case SETTINGS_BIN_UPDATE:
				sprintf (pcErrorString, "\n==> Update ERROR: Settings.bin to large: max %li bytes\n", ulFlashPartitionSize);
			break;
			case EEPROM_SETTINGS_UPDATE:
				sprintf (pcErrorString, "\n==> Update ERROR: Eeprom.bin to large: max %li bytes\n", ulFlashPartitionSize);
			break;
			default:
				sprintf (pcErrorString, "\n==> Update ERROR: Image to large: max %li bytes\n", ulFlashPartitionSize);
		}

		return -1;
	}

	return 0; /* no error*/
}

//------------------------------------------------------------------------------
//
//  Function:  ReadDHupdateFile
//
//  Read DHupdate.ini file content.
//
//  Commit values:  - *p_stDHupdateINI = Pointer to DHupdateINI struct
//                  - ulDHUpdateIniSDRAMAddress = DHupdate.ini file SDRAM address
//
//  Return value:   0 = No error
//                  1 = error
//
int ReadDHupdateFile(updateinfo_t *p_stDHupdateINI, unsigned long ulDHUpdateIniSDRAMAddress)
{
    void *p_vFilePointer = (void*)ulDHUpdateIniSDRAMAddress;    // Pointer to DHupdate.ini file
    int iPositionPointer = 0;                                   // Act. position in DHupdate.ini file
    int iPositionPointerMemorize = 0;                           // Old position in DHupdate.ini file
    int iDifference = 0;                                        // differenc between act. and old position
    int iSpecialFilename = 0;
    char *p_cCompareStringDisplay = {"display\0"};
    char *p_cCompareStringLED = {"led\0"};
    char *p_cCompareStringUpdate = {"update\0"};
    char *p_cCompareStringEnd = {"end\0"};
    
    // Set to invalid
    p_stDHupdateINI->iDisplayInfo = 0;
    p_stDHupdateINI->iLedInfo = 0;
    p_stDHupdateINI->iUpdateCounter = 0;    
    
    // Search in DHupdate.ini file for valid mask
    for(iPositionPointer = 0; *(char*)(p_vFilePointer + iPositionPointer) != '\n'; iPositionPointer++);
    *(char*)(p_vFilePointer + iPositionPointer) = '\0';
    p_stDHupdateINI->p_cValidMarker = (char*)p_vFilePointer;
    
    iPositionPointer++;
    // Search for '['']' content
    while(*(char*)(p_vFilePointer + iPositionPointer) == '[')
    {
        iPositionPointer++;
        iPositionPointerMemorize = iPositionPointer;
        while(*(char*)(p_vFilePointer + iPositionPointer) != ']') 
        {
            iPositionPointer++;
        }
        *(char*)(p_vFilePointer + iPositionPointer) = '\0';
        iDifference = iPositionPointer - iPositionPointerMemorize;
        
        if(!(memcmp ( (char*)(p_vFilePointer+iPositionPointer-iDifference), p_cCompareStringDisplay, iDifference)))
        {
            // **********************************************************
            // ***************** Update Block [display] ***************** 
            // **********************************************************
            
            p_stDHupdateINI->iDisplayInfo = 1;
            // Progress Bmp
            iPositionPointer+=2; // +2, because ']' and '\n'
            iPositionPointerMemorize = iPositionPointer;
            while(*(char*)(p_vFilePointer + iPositionPointer) != '\n') 
            {
                iPositionPointer++;
            }
            *(char*)(p_vFilePointer + iPositionPointer) = '\0';
            p_stDHupdateINI->p_cFileNameProgressBmp = (char*)(p_vFilePointer+iPositionPointerMemorize);
            // Ok Bmp
            iPositionPointer++;
            iPositionPointerMemorize = iPositionPointer;
            if(*(char*)(p_vFilePointer + iPositionPointer) == '[') 
            {
                // DHupdate.ini display file missing   
                return 1;
            }
            while(*(char*)(p_vFilePointer + iPositionPointer) != '\n') 
            {
                iPositionPointer++;
            }
            *(char*)(p_vFilePointer + iPositionPointer) = '\0';
            p_stDHupdateINI->p_cFileNameOkBmp = (char*)(p_vFilePointer+iPositionPointerMemorize);
            // Error Bmp
            iPositionPointer++;
            iPositionPointerMemorize = iPositionPointer;
            if(*(char*)(p_vFilePointer + iPositionPointer) == '[') 
            {
                // DHupdate.ini display file missing  
                return 1;
            }            
            while(*(char*)(p_vFilePointer + iPositionPointer) != '\n') 
            {
                iPositionPointer++;
            }
            *(char*)(p_vFilePointer + iPositionPointer) = '\0';
            p_stDHupdateINI->p_cFileNameErrorBmp = (char*)(p_vFilePointer+iPositionPointerMemorize);
            iPositionPointer++;
        }
        else if(!(memcmp ( (char*)(p_vFilePointer+iPositionPointer-iDifference), p_cCompareStringLED, iDifference)))
        {
            // **********************************************************
            // ******************* Update Block [led] ******************* 
            // **********************************************************
            p_stDHupdateINI->iLedInfo = 1;
            // Update GPIO
            iPositionPointer+=2; // +2, because ']' and '\n'
            iPositionPointerMemorize = iPositionPointer;
            while((*(char*)(p_vFilePointer + iPositionPointer) != '\n') && (*(char*)(p_vFilePointer + iPositionPointer) != ' '))
            {
                iPositionPointer++;
            }
            
            if(*(char*)(p_vFilePointer + iPositionPointer) == '\n')
            {
                // Active state is missing --> Don't set Update GPIO, because we don't know the active state of the connected LED
                p_stDHupdateINI->iLedInfo = 0;
                iPositionPointer++;
            }
            else
            {
                // GPIO name
                *(char*)(p_vFilePointer + iPositionPointer) = '\0';
                p_stDHupdateINI->p_cUpdateGpioName = (char*)(p_vFilePointer+iPositionPointerMemorize);
                iPositionPointer++;
                
                // GPIO Active state
                iPositionPointerMemorize = iPositionPointer;
                while(*(char*)(p_vFilePointer + iPositionPointer) != '\n')
                {
                    iPositionPointer++;
                }
                
                *(char*)(p_vFilePointer + iPositionPointer) = '\0';
                
                if (strcmp((char*)(p_vFilePointer+iPositionPointerMemorize), "high") == 0)
                {
                    p_stDHupdateINI->iUpdateGpioActiveState = 1;
                }
                else if (strcmp((char*)(p_vFilePointer+iPositionPointerMemorize), "low") == 0)
                {
                    p_stDHupdateINI->iUpdateGpioActiveState = 0;
                }
                else
                {
                    // Wrong active state argument
                    p_stDHupdateINI->iLedInfo = 0;
                }
                
                iPositionPointer++;
            }
        }
        else if(!(memcmp ( (char*)(p_vFilePointer+iPositionPointer-iDifference), p_cCompareStringUpdate, iDifference)))
        {
            // **********************************************************
            // ***************** Update Block [update] ****************** 
            // **********************************************************
            iPositionPointer+=2; // +2, because ']' and '\n'
            
            while(*(char*)(p_vFilePointer + iPositionPointer) != '[')
            {
                if(p_stDHupdateINI->iUpdateCounter >= DHUPDATEINI_MAX_UPDATES)
                {
                    // Too much update arguments in DHupdate.ini file  
                    return 1;
                }
                
                p_stDHupdateINI->iUpdateCounter++;
                // Update Type
                iPositionPointerMemorize = iPositionPointer;

                // read text until '\n' or ' '
                while((*(char*)(p_vFilePointer + iPositionPointer) != '\n') && (*(char*)(p_vFilePointer + iPositionPointer) != ' '))
                {
                    iPositionPointer++;
                }

                // Only if ' ' after update type a custom filename should be available
                if(*(char*)(p_vFilePointer + iPositionPointer) == ' ')
                {
                    iSpecialFilename = 1;
                }
                else
                {
                    iSpecialFilename = 0;
                }
                
                *(char*)(p_vFilePointer + iPositionPointer) = '\0'; // Set '\0' to mark end of update-type string

                // set pointer to update-type string in stDHUpdateInfo array
                p_stDHupdateINI->stDHUpdateInfo[p_stDHupdateINI->iUpdateCounter - 1].p_cUpdateType = (char*)(p_vFilePointer+iPositionPointerMemorize);
                iPositionPointer++;
                
                if(iSpecialFilename)
                {
                    // Update Type
                    iPositionPointerMemorize = iPositionPointer;
                    while(*(char*)(p_vFilePointer + iPositionPointer) != '\n') 
                    {
                        iPositionPointer++;
                    }
                    *(char*)(p_vFilePointer + iPositionPointer) = '\0'; // Set '\0' to mark end of filename string

                    // set pointer to filename string in stDHUpdateInfo array
                    p_stDHupdateINI->stDHUpdateInfo[p_stDHupdateINI->iUpdateCounter - 1].p_cFilename = (char*)(p_vFilePointer+iPositionPointerMemorize);
                    iPositionPointer++;
                }
                else
                {
                    p_stDHupdateINI->stDHUpdateInfo[p_stDHupdateINI->iUpdateCounter - 1].p_cFilename = NULL;
                }
            }
        }
        else if(!(memcmp ( (char*)(p_vFilePointer+iPositionPointer-iDifference), p_cCompareStringEnd, iDifference)))
        {
            return 0;
        }
        else
        {
            return 1;
        }
    }
    
    return 1;
}

//------------------------------------------------------------------------------
//
//  Function:  ShowBitmap
//
//  Shows the specified Update Bitmap on the Display
//
//  Commit values:  - *p_stDHupdateINI = Pointer to DHupdateINI struct
//                  - eBitmapType = enum to specify Bitmap (PROGRESS_BITMAP, END_BITMAP and ERROR_BITMAP)
//                  - p_cStorageDevice = Specify the act. storage device (e.g. "usb")
//                  - p_cDevicePartitionNumber = Specify the act. partition number of the storage device
//
//  Return value:   0 = No error
//                  1 = error
//
int ShowBitmap(updateinfo_t *p_stDHupdateINI, enum BitmapTypeEnum eBitmapType, char *p_cStorageDevice, char *p_cDevicePartitionNumber)
{
    char cBmpSDRAMAddress[9]        = {UPDATE_BMP_SDRAM_ADDRESS};
    char *p_cLoadBmpToSDRAM[5]      = {"fatload","","",cBmpSDRAMAddress,""};      
    char *p_cDisplayBmpOnScreen[5]  = {"bmp","display",cBmpSDRAMAddress,"32767","32767"};
    int ret_value                   = 0;

    // Set current device (mmc or usb)
    p_cLoadBmpToSDRAM[1] = p_cStorageDevice;

    // Set current device number and partition
    p_cLoadBmpToSDRAM[2] = p_cDevicePartitionNumber;  
    
    switch(eBitmapType)
    {
        case PROGRESS_BITMAP:
            // Set "progress bmp" filename from DHupdate.ini file
            p_cLoadBmpToSDRAM[4] = p_stDHupdateINI->p_cFileNameProgressBmp;
        break;
        case END_BITMAP:
            // Set "done bmp" filename from DHupdate.ini file
            p_cLoadBmpToSDRAM[4] = p_stDHupdateINI->p_cFileNameOkBmp; 
        break;
        case ERROR_BITMAP:
            // Set "error bmp" filename from DHupdate.ini file
            p_cLoadBmpToSDRAM[4] = p_stDHupdateINI->p_cFileNameErrorBmp;
        break;
        default:
            return 1;
        break;
    }
    
    // Load bitmap from MMC/SD - Card to SDRAM.
    DISABLE_PRINTF()
    ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadBmpToSDRAM);
    SESSION_DEPENDED_PRINTF_ENABLE()
    if(ret_value == 0)
    {
        do_bmp(NULL, 0, 5, p_cDisplayBmpOnScreen);
        return 0;
    }
    return 1; 

}

//------------------------------------------------------------------------------
//
//  Function:  AcitvateUpdateGPIO
//
//  Activate the Update GPIO
//
//  Commit values:  - *p_stDHupdateINI = Pointer to DHupdateINI struct
//
//  Return value:   -
//
void AcitvateUpdateGPIO(updateinfo_t *p_stDHupdateINI)
{
    // Activate Update GPIO
    if(p_stDHupdateINI->iUpdateGpioActiveState == 1)
    {
        DHCOMUpdateLED_SetHigh();
    }
    else
    {
        DHCOMUpdateLED_SetLow();
    }
}

//------------------------------------------------------------------------------
//
//  Function:  DeacitvateUpdateGPIO
//
//  Deactivate the Update GPIO
//
//  Commit values:  - *p_stDHupdateINI = Pointer to DHupdateINI struct
//
//  Return value:   -
//
void DeacitvateUpdateGPIO(updateinfo_t *p_stDHupdateINI)
{
    // Activate Update GPIO
    if(p_stDHupdateINI->iUpdateGpioActiveState == 1)
    {
        DHCOMUpdateLED_SetLow();
    }
    else
    {
        DHCOMUpdateLED_SetHigh();
    }
}

//------------------------------------------------------------------------------
//
//  Function:  UpdateGPIOBlinkInterval
//
//  Let the update LED blink.
//
//  Commit values:  - *p_stDHupdateINI = Pointer to DHupdateINI struct
//                  - iNumber = Blinking number
//
//  Return value:   -
//
void UpdateGPIOBlinkInterval(updateinfo_t *p_stDHupdateINI, int iNumber)
{
    for( ;iNumber > 0; iNumber--)
    {
        DHCOMUpdateDelayMs(250);
        AcitvateUpdateGPIO(p_stDHupdateINI);
        DHCOMUpdateDelayMs(250);
        DeacitvateUpdateGPIO(p_stDHupdateINI);
    }
}

//------------------------------------------------------------------------------
//
//  Function:  ShowUpdateError
//
//  Shows the specified Update Error String on the console, loads error bitmap
//  and sets the LED GPIO.
//
//  Commit values:  - *p_stDHupdateINI = Pointer to DHupdateINI struct
//                  - p_cErrorStringPointer = pointer to the error string
//                  - eDHCOMUpdateError = Specifies the error for the LED blink algorithm
//                  - iUpdateViaDHupdateIniFile = Specifies if Update is started 
//                                                form console or from DHupdate.ini file
//                  - p_cStorageDevice = Specify the act. storage device (e.g. "usb")
//                  - p_cDevicePartitionNumber = Specify the act. partition number of the storage device
//
//  Return value:   -
//
void ShowUpdateError(updateinfo_t *p_stDHupdateINI, char *p_cErrorStringPointer, enum UpdateErrorEnum eDHCOMUpdateError, int iUpdateViaDHupdateIniFile, char *p_cStorageDevice, char *p_cDevicePartitionNumber)
{
    printf ("%s", p_cErrorStringPointer);
    if(iUpdateViaDHupdateIniFile == 1)
    {
        // Check if DHupdate.ini contains [display] section
        if(p_stDHupdateINI->iDisplayInfo == 1)
        {
            if(ShowBitmap(p_stDHupdateINI, ERROR_BITMAP, p_cStorageDevice, p_cDevicePartitionNumber) != 0)
            {
                printf ("\n--> Update INFO: Error bitmap %s not found\n", p_stDHupdateINI->p_cFileNameErrorBmp);
            }
        }
        // Don't Start OS after Update error
        CLEAR_DH_GD_FLAG(BOOT_AFTER_UPDATE_FLAG);
        
        if(p_stDHupdateINI->iLedInfo == 1)
        {
            switch(eDHCOMUpdateError)
            {
                case DHUPDATE_INI_ERROR:
                    DeacitvateUpdateGPIO(p_stDHupdateINI);
                    while(1)
                    {
                        UpdateGPIOBlinkInterval(p_stDHupdateINI, 1);
                        DHCOMUpdateDelayMs(1750);
                    }
                break;
                case FILE_NOT_FOUND_ERROR:
                    DeacitvateUpdateGPIO(p_stDHupdateINI);
                    while(1)
                    {
                        UpdateGPIOBlinkInterval(p_stDHupdateINI, 2);
                        DHCOMUpdateDelayMs(1750);
                    }
                break;
                case FLASH_ERROR:
                    DeacitvateUpdateGPIO(p_stDHupdateINI);
                    while(1)
                    {
                        UpdateGPIOBlinkInterval(p_stDHupdateINI, 3);
                        DHCOMUpdateDelayMs(1750);
                    }
                break;
                case IMAGE_TYPE_ERROR:
                    DeacitvateUpdateGPIO(p_stDHupdateINI);
                    while(1)
                    {
                        UpdateGPIOBlinkInterval(p_stDHupdateINI, 4);
                        DHCOMUpdateDelayMs(1750);
                    }
                break;
                case INVALID_FILE_ERROR:
                    DeacitvateUpdateGPIO(p_stDHupdateINI);
                    while(1)
                    {
                        UpdateGPIOBlinkInterval(p_stDHupdateINI, 5);
                        DHCOMUpdateDelayMs(1750);
                    }
                break;
                case IMAGE_SIZE_ERROR:
                    DeacitvateUpdateGPIO(p_stDHupdateINI);
                    while(1)
                    {
                        UpdateGPIOBlinkInterval(p_stDHupdateINI, 6);
                        DHCOMUpdateDelayMs(1750);
                    }
                break;
                case INTERNAL_ERROR:
                    DeacitvateUpdateGPIO(p_stDHupdateINI);
                    while(1)
                    {
                        UpdateGPIOBlinkInterval(p_stDHupdateINI, 7);
                        DHCOMUpdateDelayMs(1750);
                    }
                break;
                case NO_ERROR:
                	/* This case is to suppress warnings */
                break;
            }

        }
    }
    
    return;
}

//------------------------------------------------------------------------------
//
//  Function:  SetENVScriptFilename
//
//  Set the ENV script filename for Linux and WinCE, if we don't use the default filename
//
//  Commit values:  - *file_name = OS image filename
//                  - *p_cENVScriptFileName = Pointer to the ENV script filename
//
//  Return value:   -
//
void SetENVScriptFilename(char *file_name, char *p_cENVScriptFileName)
{
    int i = 0;
    
    // Set ENV Script File name
    for(i = 0; (i < 50) && (*(file_name+i) != '.') && (*(file_name+i) != '\0'); i++)
    {
        *(p_cENVScriptFileName+i) = *(file_name+i);
    }
    *(p_cENVScriptFileName+i) = '.';
    *(p_cENVScriptFileName+i+1) = 'e';
    *(p_cENVScriptFileName+i+2) = 'n';
    *(p_cENVScriptFileName+i+3) = 'v';
    *(p_cENVScriptFileName+i+4) = '\0';
}

//------------------------------------------------------------------------------
//
//  Function:  CopyAddressStringToCharArray
//
//  Copy address string content to local char array
//
//  Commit values:  - *p_cCharArray = Pointer to the Char Array
//                  - *p_cPointer = Copy Pointer
//
//  Return value:   -
//
void CopyAddressStringToCharArray(char *p_cCharArray, char *p_cPointer)
{
    int i,j;
    for(i = 0, j = 0; i < 8; i++, j++)
    {
        // Check if string contains "0x"
        if(p_cPointer[j] == 'x')
        {
            i = 0;
            j++;
        }
        p_cCharArray[i] = p_cPointer[j];
    }
    
    p_cCharArray[i] = '\0';
}


//------------------------------------------------------------------------------
//
//  Function:  UpdateGlobalDataDHSettings
//
//  Update the DHsettings global data structure from ram
//
//  Commit values:  - addr = Address of the new settings file.
//
//  Return value:   0 = No error
//                  1 = error
//
int UpdateGlobalDataDHSettings (ulong addr)
{
    // settings.bin file Valid Mask should be "DH" = 0x4844
    if((((*(ulong *)addr) >> 16) & 0xFFFF) == 0x4844)
    {
        gd->dh_board_settings.cLength = (readl(addr) & 0xFF);
        gd->dh_board_settings.cDisplayID = ((readl(addr) & 0xFF00) >> 8);

        gd->dh_board_settings.wYResolution = (readl(addr+4) & 0xFFFF);
        gd->dh_board_settings.wXResolution = ((readl(addr+4) & 0xFFFF0000) >> 16);

        gd->dh_board_settings.wLCDConfigFlags = (readl(addr+8) & 0xFFFF);
        gd->dh_board_settings.wPixelClock = ((readl(addr+8) & 0xFFFF0000) >> 16);

        gd->dh_board_settings.wVPulseWidth = (readl(addr+12) & 0xFFFF);
        gd->dh_board_settings.wHPulseWidth = ((readl(addr+12) & 0xFFFF0000) >> 16);

        gd->dh_board_settings.wHBackPorch = (readl(addr+16) & 0xFFFF);
        gd->dh_board_settings.wHFrontPorch = ((readl(addr+16) & 0xFFFF0000) >> 16);

        gd->dh_board_settings.wVBackPorch = (readl(addr+20) & 0xFFFF);
        gd->dh_board_settings.wVFrontPorch = ((readl(addr+20) & 0xFFFF0000) >> 16);

        gd->dh_board_settings.cACBiasTrans = (readl(addr+24) & 0xFF);
        gd->dh_board_settings.cACBiasFreq = ((readl(addr+24) & 0xFF00) >> 8);
        gd->dh_board_settings.cDatalines = ((readl(addr+24) & 0xFFFF0000) >> 16);

        gd->dh_board_settings.wGPIODir = (readl(addr+32) & 0xFFFF);
        gd->dh_board_settings.wGPIOState = (readl(addr+36) & 0xFFFF);

        gd->dh_board_settings.wHWConfigFlags = (readl(addr+40) & 0xFFFF);

        return 0;
    }
    return 1; /* no valid settings.bin available */
}


//------------------------------------------------------------------------------
//
//  Function:  DHCOMupdate
//
//  DHCOM U-Boot Update function
//
//  Commit values:  - *cmdtp = Command Type, used for cmd_usage()
//                  - argc = Number of command line arguments
//                  - *argv[] = Command line argument strings
//                  - *p_stDHupdateINI = Pointer to DHupdateINI struct
//                  - p_cStorageDevice = Specify the act. storage device (e.g. "usb")
//                  - p_cDevicePartitionNumber = Specify the act. partition number of the storage device
//
//  Return value:   0 = No error
//                  1 = error
//
int DHCOMupdate (cmd_tbl_t *cmdtp, int argc, char *argv[], updateinfo_t *p_stDHupdateINI, char *p_cStorageDevice, char *p_cDevicePartitionNumber)
{
    char cENVSDRAMBufferAddress[9];
    CopyAddressStringToCharArray(&cENVSDRAMBufferAddress[0], getenv ("rambufferaddr"));
    char cENVOSImageFlashAddress[9];
    CopyAddressStringToCharArray(&cENVOSImageFlashAddress[0], getenv ("nandflashOSaddr"));
    char cENVBootloaderFlashAddress[9];
    CopyAddressStringToCharArray(&cENVBootloaderFlashAddress[0], getenv ("bootloaderflashaddr"));
    char cENVSplashimageFlashAddress[9];
    CopyAddressStringToCharArray(&cENVSplashimageFlashAddress[0], getenv ("splashimage"));        
    char cENVSettingsBinFlashAddress[9];
    CopyAddressStringToCharArray(&cENVSettingsBinFlashAddress[0], getenv ("settingsflashaddr")); 
    char cENVSDRAMKernelLoadAddr[9];
    sprintf(cENVSDRAMKernelLoadAddr,"%08x",CONFIG_SYS_LOAD_ADDR);
    

    char cDHUpdateIniSDRAMAddress[9]        = {UPDATE_DHUPDATE_INI_SDRAM_ADDRESS};
    char cImageFileSize[9]                  = {"12345677\0"};
    char cTempBuffer[9]                  	= {"12345677\0"};
    char cENVScriptFileName[50];
    char cErrorString[100];
    char cBootargsString[256];

    nand_info_t *nand;
    nand = &nand_info[nand_curr_device];		/* Get nand-info structure from current device */

    /* get address of copy-buffer */
    unsigned long ulSDRAMBufferAddress      = simple_strtoul (cENVSDRAMBufferAddress, NULL, 16);

    /* get OS-Image flash offset */
    unsigned long ulOSImageFlashAddress     = simple_strtoul (cENVOSImageFlashAddress, NULL, 16);

	/* get wince flash partition size */
    CopyAddressStringToCharArray(cTempBuffer, getenv ("wincepartitionsize"));
    unsigned long ulWinCEFlashPartitionSize = simple_strtoul (cTempBuffer, NULL, 16);

	/* get linux-kernel flash partition size */
    CopyAddressStringToCharArray(cTempBuffer, getenv ("linuxkernelpartitionsize"));
    unsigned long ulLinuxFlashPartitionSize = simple_strtoul (cTempBuffer, NULL, 16);

	/* get u-boot flash offset and partition size */
    unsigned long ulBootloaderFlashAddress  = simple_strtoul (cENVBootloaderFlashAddress, NULL, 16);
    unsigned long ulBootloaderFlashPartitionSize = CONFIG_SYS_MONITOR_LEN;

	/* get settings.bin flash offset and partition size */
    unsigned long ulSettingsBinFlashAddress = simple_strtoul (cENVSettingsBinFlashAddress, NULL, 16);
    unsigned long ulSettingsBinFlashPartitionSize = ulOSImageFlashAddress - ulSettingsBinFlashAddress;

	/* get spashimage flash offset and partition size */
    unsigned long ulSplashimageFlashAddress = simple_strtoul (cENVSplashimageFlashAddress, NULL, 16);
    unsigned long ulSplashimagePartitionSize = ulSettingsBinFlashAddress - ulSplashimageFlashAddress;


    
    void *p_vUpdateArgument;  
    char *cmd, *file_name, *file_type;
    char *p_cSDRAMBufferAddress             = (char*)ulSDRAMBufferAddress;
    char *p_cCompareString[1]               = {"##DHCOMupdate##\0"};
    char *p_cENVCompareString[1]            = {"Set_ENV"}; 
    char *p_cLoadDHUpdateIniToSDRAM[5]      = {"fatload","","",cDHUpdateIniSDRAMAddress,"DHupdate.ini"};
    char *p_cLoadNkBinToSDRAM[5]            = {"fatload","","",cENVSDRAMBufferAddress,"nk.bin"};
    char *p_cLoadNkGzToSDRAM[5]             = {"fatload","","",cENVSDRAMBufferAddress,"nk.gz"};
    char *p_cLoadNkENVToSDRAM[5]            = {"fatload","","",cENVSDRAMBufferAddress,"nk.env"};
    char *p_cLoaduImageToSDRAM[5]           = {"fatload","","",cENVSDRAMBufferAddress,"uImage"};
    char *p_cLoaduImageENVToSDRAM[5]        = {"fatload","","",cENVSDRAMBufferAddress,"uImage.env"};
    char *p_cRunENVScript[2]                = {"source",cENVSDRAMBufferAddress};    
    char *p_cLoadUBootBinToSDRAM[5]         = {"fatload","","",cENVSDRAMBufferAddress,"u-boot.bin"};
    char *p_cLoadSplashBmpToSDRAM[5]        = {"fatload","","",cENVSDRAMBufferAddress,"splash.bmp"};    
    char *p_cLoadSettingsBinToSDRAM[5]      = {"fatload","","",cENVSDRAMBufferAddress,"settings.bin"};
    char *p_cLoadEepromBinToSDRAM[5]        = {"fatload","","",cENVSDRAMBufferAddress,"eeprom.bin"};
    char *p_cLoadUpdateKernelToSDRAM[5]     = {"fatload","","",cENVSDRAMKernelLoadAddr,"uImage.update"};
    char *p_cRootfsFilename        			= {"fsroot.tar.gz"};
    
    int ret_value = 0;
    int i;
    int iUpdateViaDHupdateIniFile = 1;      // 0 = Command line Update with more then 1 arguments (e.g. "update wince dh_nk.gz")
                                            // 1 = DHupdate.ini file Update (autostart or command line call "update")
                                            // 2 = error
    int iUpdateLoopCounter = 0;
    int iRefreshStringFound = 0;
    int iDoReset = 0;						// if set to 1 do a reset if 0 do not execute a reset (used as paremeter for update-kernel [rootfs])
    int iLedGPIO = -1;						// if set to -1 there is no LED-gpio (used as paremeter for update-kernel [rootfs])
    char cUpdateArgument = 0;
    char cOSFileType = NO_SPECIAL_FILE_TYPE;
    unsigned long ulFilesize;
    unsigned long ulBlocks;    
    unsigned long ulFlashBlockSize;
#ifdef DH_NAND_FLASH
    // unsigned long ulNandPagesPerBlock;
#endif /* DH_NAND_FLASH  */

	/* at least one arguments please */
	if (argc < 1)
		goto usage;

#ifdef DH_NAND_FLASH

    // Determine NAND Flash block size
    ulFlashBlockSize = nand->erasesize;

#else /* DH_SPI_FLASH */
    ulFlashBlockSize = SPI_FLASH_SECTOR_SIZE;
#endif /* DH_NAND_FLASH  */        
    
    // Set current device (mmc or usb)
    p_cLoadDHUpdateIniToSDRAM[1] = p_cStorageDevice;
    p_cLoadNkBinToSDRAM[1] = p_cStorageDevice;
    p_cLoadNkGzToSDRAM[1] = p_cStorageDevice;
    p_cLoadNkENVToSDRAM[1] = p_cStorageDevice; 
    p_cLoaduImageToSDRAM[1] = p_cStorageDevice;    
    p_cLoaduImageENVToSDRAM[1] = p_cStorageDevice;
    p_cLoadUBootBinToSDRAM[1] = p_cStorageDevice;  
    p_cLoadSplashBmpToSDRAM[1] = p_cStorageDevice;  
    p_cLoadSettingsBinToSDRAM[1] = p_cStorageDevice; 
    p_cLoadEepromBinToSDRAM[1] = p_cStorageDevice;
    p_cLoadUpdateKernelToSDRAM[1] = p_cStorageDevice;

    // Set current device number and partition
    p_cLoadDHUpdateIniToSDRAM[2] = p_cDevicePartitionNumber;   
    p_cLoadNkBinToSDRAM[2] = p_cDevicePartitionNumber;
    p_cLoadNkGzToSDRAM[2] = p_cDevicePartitionNumber;   
    p_cLoadNkENVToSDRAM[2] = p_cDevicePartitionNumber;    
    p_cLoaduImageToSDRAM[2] = p_cDevicePartitionNumber;       
    p_cLoaduImageENVToSDRAM[2] = p_cDevicePartitionNumber;
    p_cLoadUBootBinToSDRAM[2] = p_cDevicePartitionNumber;  
    p_cLoadSplashBmpToSDRAM[2] = p_cDevicePartitionNumber;    
    p_cLoadSettingsBinToSDRAM[2] = p_cDevicePartitionNumber;
    p_cLoadEepromBinToSDRAM[2] = p_cDevicePartitionNumber;
    p_cLoadUpdateKernelToSDRAM[2] = p_cDevicePartitionNumber;
    
    // Update without DHupdate.ini file --> Command line update
    if(argc > 1)
    {
        cmd = argv[1];
        
        //
        // StartUp automatic update with DHupdate.ini file
        if (strcmp(cmd, "auto") == 0) 
        {
            iUpdateViaDHupdateIniFile = 1;    // Set indicator for automatic update
        }
        // Identify the specified command line update
        else if (strcmp(cmd, "wince") == 0) 
        {
            cUpdateArgument = WINCE_IMAGE_UPDATE;
            p_vUpdateArgument = &cUpdateArgument;
            iUpdateViaDHupdateIniFile = 0;
            // Load File name from command line
            if(argc > 2)
            {
                file_name = argv[2];
                
                // Set ENV Script File name
                SetENVScriptFilename(file_name, &cENVScriptFileName[0]);
                p_cLoadNkENVToSDRAM[4] = &cENVScriptFileName[0];                
            
                // Search for image type (*.gz or *.bin)
                for(i = 0; *(file_name+i) != '.'; i++);
                
                file_type = file_name + i + 1;

#ifdef DH_NAND_FLASH                
                if (strcmp(file_type, "gz") == 0) 
                {
                    cOSFileType = UPDATE_FILE_TYPE_GZ;
                    p_cLoadNkGzToSDRAM[4] = file_name;
                }
                else if (strcmp(file_type, "bin") == 0) 
                {
                    cOSFileType = UPDATE_FILE_TYPE_BIN;
                    p_cLoadNkBinToSDRAM[4] = file_name;
                }
                else if (strcmp(file_type, "env") == 0) 
                {
                    // Update only ENV settings via ENV script File
                    cOSFileType = UPDATE_FILE_TYPE_ENV;
                }                
                else
                {
                    printf ("\n--> Update ERROR: Wrong WinCE image type (Filename must end with .gz or .bin) \n");        
                    return 1;
                }
#else /* DH_SPI_FLASH */
                // If we are using SPI Flash, we could only update ENV
                // --> Show ERROR message, if *.env file is missing
                cOSFileType = UPDATE_FILE_TYPE_ENV;
#endif /* DH_NAND_FLASH */
            }
        }
        else if (strcmp(cmd, "linux") == 0) 
        {
            cUpdateArgument = LINUX_IMAGE_UPDATE;
            p_vUpdateArgument = &cUpdateArgument;
            iUpdateViaDHupdateIniFile = 0;
            // Load File name from command line
            if(argc > 2)
            {
                file_name = argv[2];
                p_cLoaduImageToSDRAM[4] = file_name;
            
                // Set ENV Script File name
                SetENVScriptFilename(file_name, &cENVScriptFileName[0]);
                p_cLoaduImageENVToSDRAM[4] = &cENVScriptFileName[0];
                
                for(i = 0; (i < 50) && (*(file_name+i) != '.') && (*(file_name+i) != '\0'); i++);
                
                file_type = file_name + i + 1;
                
#ifdef DH_NAND_FLASH                  
                if (strcmp(file_type, "env") == 0) 
                {
                    // Update only ENV settings via ENV script File
                    cOSFileType = UPDATE_FILE_TYPE_ENV;
                }
#else /* DH_SPI_FLASH */
                // If we are using SPI Flash, we could only update ENV
                // --> Show ERROR message, if *.env file is missing
                cOSFileType = UPDATE_FILE_TYPE_ENV;
#endif /* DH_NAND_FLASH */
            }
        }
        else if (strcmp(cmd, "bootloader") == 0) 
        {
            cUpdateArgument = BOOTLOADER_FLASH_UPDATE;
            p_vUpdateArgument = &cUpdateArgument;
            iUpdateViaDHupdateIniFile = 0;
            // Load File name from command line
            if(argc > 2)
            {
                file_name = argv[2];
                p_cLoadUBootBinToSDRAM[4] = file_name;
            }
        }
        else if (strcmp(cmd, "splash") == 0) 
        {
            cUpdateArgument = SPLASHIMAGE_UPDATE;
            p_vUpdateArgument = &cUpdateArgument;
            iUpdateViaDHupdateIniFile = 0;
            // Load File name from command line
            if(argc > 2)
            {
                file_name = argv[2];
                p_cLoadSplashBmpToSDRAM[4] = file_name;
            }
        }
        else if (strcmp(cmd, "settings") == 0) 
        {
            cUpdateArgument = SETTINGS_BIN_UPDATE;
            p_vUpdateArgument = &cUpdateArgument;
            iUpdateViaDHupdateIniFile = 0;
            // Load File name from command line
            if(argc > 2)
            {
                file_name = argv[2];
                p_cLoadSettingsBinToSDRAM[4] = file_name;
            }
        }
        else if (strcmp(cmd, "eeprom") == 0)
        {
            cUpdateArgument = EEPROM_SETTINGS_UPDATE;
            p_vUpdateArgument = &cUpdateArgument;
            iUpdateViaDHupdateIniFile = 0;
            // Load File name from command line
            if(argc > 2)
            {
                file_name = argv[2];
                p_cLoadEepromBinToSDRAM[4] = file_name;
            }
        }
        else if (strcmp(cmd, "rootfs") == 0)
        {
            cUpdateArgument = NAND_ROOTFS_UPDATE;
            p_vUpdateArgument = &cUpdateArgument;
            iUpdateViaDHupdateIniFile = 0;
            // Load File name from command line
            if(argc > 2)
            {
                file_name = argv[2];
                p_cRootfsFilename = file_name;
            }
        }
        else
        {
            // Wrong argument
            iUpdateViaDHupdateIniFile = 2;

            printf ("\n--> Update ERROR: Wrong command line update argument\n");
            goto usage;
        }
    }    

	// *************************************************************************************************************************************************
    // Handle DHupdate.ini File
    // (if we run update in auto mode or just update without any options)
    if(iUpdateViaDHupdateIniFile == 1)
    {
        // Try to load DHupdate.ini file to SDRAM.
        DISABLE_PRINTF()
        ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadDHUpdateIniToSDRAM);
        SESSION_DEPENDED_PRINTF_ENABLE()

        // Check if DHupdate.ini was found on Storage Device
        if((ret_value != 0))
        {
            printf ("\n--> Update INFO: No DHupdate.ini file found on Storage Device\n");
            return 1;
        }

		// Read DHupdate.ini file content
		if(ReadDHupdateFile(p_stDHupdateINI, simple_strtoul (cDHUpdateIniSDRAMAddress, NULL, 16)))
		{
			printf("\n--> Update ERROR: Wrong DHupdate.ini file content \n");
			return 1;
		}

		// Check if DHupdate.ini is a valid update file.
		if(memcmp ( (char*)p_stDHupdateINI->p_cValidMarker, p_cCompareString[0], 15))
		{
			printf ("\n--> Update ERROR: No valid DHupdate.ini file found on Storage Device\n");
			// Don't Start OS after Update error
			CLEAR_DH_GD_FLAG(BOOT_AFTER_UPDATE_FLAG);
			return 1;
		}

		printf ("\n--> Update: DHupdate.ini (%d bytes) found on Storage Device", (int)simple_strtoul (getenv("filesize"), NULL, 16));
    }

	// *************************************************************************************************************************************************
	// Check if DHupdate.ini contains [display] section
	// If so, load progress bitmap.
	if(p_stDHupdateINI->iDisplayInfo == 1)
	{
		for(i = 0; i < p_stDHupdateINI->iUpdateCounter; i++)
		{
			if (strcmp(p_stDHupdateINI->stDHUpdateInfo[i].p_cUpdateType, "refresh") == 0)
			{
				iRefreshStringFound = 1;
				break;
			}
		}

		if(iRefreshStringFound == 0)
		{
			if(ShowBitmap(p_stDHupdateINI, PROGRESS_BITMAP, p_cStorageDevice, p_cDevicePartitionNumber) != 0)
			{
				printf ("\n--> Update INFO: Progress bitmap %s not found", p_stDHupdateINI->p_cFileNameProgressBmp);
			}
		}
	}

	// *************************************************************************************************************************************************
	// Check if DHupdate.ini contains [led] section
	// If so, init. update led
	if(p_stDHupdateINI->iLedInfo == 1)
	{
		if(DHCOMUpdateLED_Init(p_stDHupdateINI) != 0)
		{
			printf ("\n--> Update INFO: Can't initialize update LED Pin %s", p_stDHupdateINI->p_cUpdateGpioName);
			p_stDHupdateINI->iLedInfo = 0;
		}
		else
		{
			// Activate Update GPIO
			AcitvateUpdateGPIO(p_stDHupdateINI);
		}
	}

	printf ("\n");

	// *************************************************************************************************************************************************
	// ==> Start UPDATE LOOP
	do
	{
		// Identify the next specified update in DHupdate.ini file
		if(iUpdateViaDHupdateIniFile == 1)
		{
			// Don't enable printf after update, if DIS_PRINTF in settings.bin file is set.
			SET_DH_GD_FLAG(UPDATE_ACTIVE_FLAG);

			if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "wince") == 0)
			{
				cUpdateArgument = WINCE_IMAGE_UPDATE;
				p_vUpdateArgument = &cUpdateArgument;

				// Check if filename is specified in DHupdate.ini file
				if(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename != NULL)
				{
					// Load File name from DHupdate.ini file
					file_name = p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename;

					// Set ENV Script File name
					SetENVScriptFilename(file_name, &cENVScriptFileName[0]);
					p_cLoadNkENVToSDRAM[4] = &cENVScriptFileName[0];

					// Search for image type (*.gz or *.bin)
					for(i = 0; *(file_name+i) != '.'; i++);

					file_type = file_name + i + 1;

#ifdef DH_NAND_FLASH                            
					if (strcmp(file_type, "gz") == 0)
					{
						cOSFileType = UPDATE_FILE_TYPE_GZ;
						p_cLoadNkGzToSDRAM[4] = file_name;
					}
					else if (strcmp(file_type, "bin") == 0)
					{
						cOSFileType = UPDATE_FILE_TYPE_BIN;
						p_cLoadNkBinToSDRAM[4] = file_name;
					}
					else if (strcmp(file_type, "env") == 0)
					{
						// Update only ENV settings via ENV script File
						cOSFileType = UPDATE_FILE_TYPE_ENV;
					}
					else
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: Wrong WinCE image type (Filename must end with .gz or .bin) \n");
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], IMAGE_TYPE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}
#else /* DH_SPI_FLASH */
					// If we are using SPI Flash, we could only update ENV
					// --> Show ERROR message, if *.env file is missing
					cOSFileType = UPDATE_FILE_TYPE_ENV;
#endif /* DH_NAND_FLASH */
				}
			}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "linux") == 0)
			{
				cUpdateArgument = LINUX_IMAGE_UPDATE;
				p_vUpdateArgument = &cUpdateArgument;

				// Check if filename is specified in DHupdate.ini file
				if(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename != NULL)
				{
					file_name = p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename;
					p_cLoaduImageToSDRAM[4] = file_name;

					// Set ENV Script File name
					SetENVScriptFilename(file_name, &cENVScriptFileName[0]);
					p_cLoaduImageENVToSDRAM[4] = &cENVScriptFileName[0];

					for(i = 0; (i < 50) && (*(file_name+i) != '.') && (*(file_name+i) != '\0'); i++);

					file_type = file_name + i + 1;

#ifdef DH_NAND_FLASH                             
					if (strcmp(file_type, "env") == 0)
					{
						// Update only ENV settings via ENV script File
						cOSFileType = UPDATE_FILE_TYPE_ENV;
					}
#else /* DH_SPI_FLASH */
					// If we are using SPI Flash, we could only update ENV
					// --> Show ERROR message, if *.env file is missing
					cOSFileType = UPDATE_FILE_TYPE_ENV;
#endif /* DH_NAND_FLASH */
				}
			}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "bootloader") == 0)
			{
				cUpdateArgument = BOOTLOADER_FLASH_UPDATE;
				p_vUpdateArgument = &cUpdateArgument;

				// Check if filename is specified in DHupdate.ini file
				if(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename != NULL)
				{
					p_cLoadUBootBinToSDRAM[4] = p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename;
				}
			}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "splash") == 0)
			{
				cUpdateArgument = SPLASHIMAGE_UPDATE;
				p_vUpdateArgument = &cUpdateArgument;

				// Check if filename is specified in DHupdate.ini file
				if(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename != NULL)
				{
					p_cLoadSplashBmpToSDRAM[4] = p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename;
				}
			}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "settings") == 0)
			{
				cUpdateArgument = SETTINGS_BIN_UPDATE;
				p_vUpdateArgument = &cUpdateArgument;

				// Check if filename is specified in DHupdate.ini file
				if(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename != NULL)
				{
					p_cLoadSettingsBinToSDRAM[4] = p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename;
				}
			}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "eeprom") == 0)
			{
				cUpdateArgument = EEPROM_SETTINGS_UPDATE;
				p_vUpdateArgument = &cUpdateArgument;

				// Check if filename is specified in DHupdate.ini file
				if(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename != NULL)
				{
					p_cLoadEepromBinToSDRAM[4] = p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename;
				}
			}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "refresh") == 0)
				{
					cUpdateArgument = REFRESH_DH_SETTINGS;
					p_vUpdateArgument = &cUpdateArgument;
				}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "reset") == 0)
				{
					cUpdateArgument = EXECUTE_RESET;
					p_vUpdateArgument = &cUpdateArgument;
				}
			else if (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cUpdateType, "rootfs") == 0)
				{
					cUpdateArgument = NAND_ROOTFS_UPDATE;
					p_vUpdateArgument = &cUpdateArgument;

					// Check if filename of rootfs.tar.gz is specified in DHupdate.ini file
					if(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename != NULL)
					{
						p_cRootfsFilename = p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter].p_cFilename;
					}
				}
			else
			{
				// Wrong argument
				sprintf (&cErrorString[0], "\n--> Update ERROR: Wrong or no Update Type in DHupdate.ini file\n");
				ShowUpdateError(p_stDHupdateINI, &cErrorString[0], DHUPDATE_INI_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
				return 1;
			}
		}
		else
		{
			// Run only one command line update
			p_stDHupdateINI->iUpdateCounter = 1;
		}

		// Argument = '1' ==> Update only WinCE Flash Image (nk.bin or nk.gz)
		// Argument = '2' ==> Update only Linux Flash Image
		// Argument = '3' ==> Update Bootloader Flash Image
		// Argument = '4' ==> Update splash.bmp (Start Screen Picture)
		// Argument = '5' ==> Update settings.bin file
		// Argument = '6' ==> Refresh DH settings and LCD controller
		// Argument = '7' ==> Display adapter EEPROM settings update
		// Argument = '8' ==> Reset Board
		// Argument = '9' ==> Update rootfs in NAND-Flash
		switch (*(char*)p_vUpdateArgument)
		{
			// ***************************************************************************************
			// ******************* Update only WinCE Image in the Flash memory ***********************
			// ***************************************************************************************
			case WINCE_IMAGE_UPDATE:
#ifdef DH_NAND_FLASH
				printf ("\n==> Update: Start to Update WinCE Flash Image");

				// if FileType = UPDATE_FILE_TYPE_BIN or UPDATE_FILE_TYPE_GZ update flash image
				// otherwise update only env settings.
				if(cOSFileType != UPDATE_FILE_TYPE_ENV)
				{
					// Load nk.bin or nk.gz file from Storage Device to SDRAM.
					if((argc < 3) || (cOSFileType == UPDATE_FILE_TYPE_BIN))
					{
						// Try to load nk.bin or command line specified *.bin file
						DISABLE_PRINTF()
						ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadNkBinToSDRAM);
						SESSION_DEPENDED_PRINTF_ENABLE()
					}
					else
					{
						// Can't load *.bin file or File Type is UPDATE_FILE_TYPE_GZ
						ret_value = 1;
					}

					if(ret_value == 1)
					{
						if((argc < 3) || (cOSFileType == UPDATE_FILE_TYPE_GZ))
						{
							// Try to load nk.gz or command line specified *.gz file
							DISABLE_PRINTF()
							ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadNkGzToSDRAM);
							SESSION_DEPENDED_PRINTF_ENABLE()
						}
						else
						{
							// Can't load *.gz file
							ret_value = 1;
						}

						if(ret_value == 0)
						{
							printf ("\n--> Update: Load %s to SDRAM (%d bytes)", (char*)p_cLoadNkGzToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));
							// Save NAND Flash Image Type for do_bootwince command
							DISABLE_PRINTF()
							setenv("winceimagetype", "nk.gz");
							saveenv();
							SESSION_DEPENDED_PRINTF_ENABLE()
						}
					}
					else
					{
						printf ("\n--> Update: Load %s to SDRAM (%d bytes)", (char*)p_cLoadNkBinToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));
						// Save NAND Flash Image Type for do_bootwince command
						DISABLE_PRINTF()
						setenv("winceimagetype", "nk.bin");
						saveenv();
						SESSION_DEPENDED_PRINTF_ENABLE()
					}

					if(ret_value == 0)
					{
						// Calculate necessary sectors in flash for the nk.bin file.
						ulFilesize = simple_strtoul (getenv("filesize"), NULL, 16);
						ulBlocks = ulFilesize / ulFlashBlockSize + 0x1;

						// Check if the file fits into to the flash-partition
						if (check_imagesize (ulFilesize, ulWinCEFlashPartitionSize, WINCE_IMAGE_UPDATE, cErrorString ))
						{
							// ERROR: file is to large for the specified partition
							ShowUpdateError(p_stDHupdateINI, cErrorString, IMAGE_SIZE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							return 1;
						}

						printf ("\n--> Update: The new WinCE image File needs %lu Flash blocks\n", ulBlocks);
						printf ("    e = Erase Block\n");
						printf ("    w = Write to Block\n");
						printf ("    d = Done\n");
						printf ("    Write File to Flash:[");

						ret_value = update_flash_content (ulOSImageFlashAddress, ulSDRAMBufferAddress, ulBlocks, ulFlashBlockSize);

						if((ret_value & 0x3) == 1)
						{
							sprintf (&cErrorString[0], "\n--> Update ERROR: Erase error on block 0x%08x\n", (unsigned int)(ulOSImageFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
							ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							// Save NAND Flash Image Size for do_bootwince command
							DISABLE_PRINTF()
							setenv("winceimagesize", "00000000");
							setenv("winceimagetype", "no_wince");
							saveenv();
							SESSION_DEPENDED_PRINTF_ENABLE()
							return 1;
						}
						else if((ret_value & 0x3) == 2)
						{
							sprintf (&cErrorString[0], "\n--> Update ERROR: Write error on block 0x%08x\n", (unsigned int)(ulOSImageFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
							ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							// Save NAND Flash Image Size for do_bootwince command
							DISABLE_PRINTF()
							setenv("winceimagesize", "00000000");
							setenv("winceimagetype", "no_wince");
							saveenv();
							SESSION_DEPENDED_PRINTF_ENABLE()
							return 1;
						}

						// Windows Embedded CE Flash Update done.
						printf ("\n--> Update: WinCE Update done");

						// Save NAND Flash Image Size for do_bootwince command
						DISABLE_PRINTF()
						sprintf (&cImageFileSize[0], "%08x", (unsigned int)ulFilesize);
						setenv("winceimagesize", &cImageFileSize[0]);
						SESSION_DEPENDED_PRINTF_ENABLE()
					}
					else
					{
						if(cOSFileType == UPDATE_FILE_TYPE_GZ)
						{
							// Can't load *.gz file from Storage Device to SDRAM.
							sprintf (&cErrorString[0], "\n--> Update ERROR: No %s found on Storage Device\n", (char*)p_cLoadNkGzToSDRAM[4]);
							ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							return 1;
						}
						else
						{
							// Can't load *.bin file from Storage Device to SDRAM.
							sprintf (&cErrorString[0], "\n--> Update ERROR: No %s found on Storage Device\n", (char*)p_cLoadNkBinToSDRAM[4]);
							ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							return 1;
						}
					}

					// Setup default U-Boot NAND Flash ENV for WinCE
					printf ("\n--> Update: Setup ENV for WinCE");
					//setenv("loadaddr", "00200000");
					setenv("bootcmd", "run nandbootwince");
					DISABLE_PRINTF()
					saveenv();
					SESSION_DEPENDED_PRINTF_ENABLE()
				}
#else /* DH_SPI_FLASH */
				printf ("\n==> Update INFO: No WinCE Update possible, because we are running on SPI Flash");
#endif /* DH_NAND_FLASH */
				// Try to load WinCE ENV setup file
				DISABLE_PRINTF()
				ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadNkENVToSDRAM);
				SESSION_DEPENDED_PRINTF_ENABLE()

				if(ret_value == 0)
				{
					printf ("\n--> Update: Load ENV setup file %s to SDRAM (%d bytes)", (char*)p_cLoadNkENVToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));

					// Setup U-Boot ENV for WinCE
					printf ("\n--> Update: Setup ENV for WinCE");

					// Check if *.env file contains the script name "Set_ENV"
					if(!(memcmp ( (char*)(p_cSDRAMBufferAddress+0x20), p_cENVCompareString[0], 7)))
					{
						printf ("\n\n----------------- Print ENV Script Info: -----------------\n");
						ret_value = do_source(NULL, 0, 2, p_cRunENVScript);
						printf ("\n----------------------------------------------------------\n");
					}
					else
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: No valid %s file found on Storage Device\n", (char*)p_cLoadNkENVToSDRAM[4]);
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], INVALID_FILE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					DISABLE_PRINTF()
					saveenv();
					SESSION_DEPENDED_PRINTF_ENABLE()
				}
				else
				{
					if(cOSFileType == UPDATE_FILE_TYPE_ENV)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: No ENV setup file %s found on Storage Device\n", (char*)p_cLoadNkENVToSDRAM[4]);
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}
					else
					{
						printf ("\n--> Update INFO: No ENV setup file %s found on Storage Device\n", (char*)p_cLoadNkENVToSDRAM[4]);
					}
				}
			break;

			// ***************************************************************************************
			// ******************* Update only Linux Image in the Flash memory ***********************
			// ***************************************************************************************
			case LINUX_IMAGE_UPDATE:
#ifdef DH_NAND_FLASH
			   printf ("\n==> Update: Start to Update Linux Flash Image");

				// if FileType != UPDATE_FILE_TYPE_ENV update flash image
				// otherwise update only env settings.
				if(cOSFileType != UPDATE_FILE_TYPE_ENV)
				{
					// Load Linux image file from Storage Device to SDRAM.
					// Try to load Linux image
					DISABLE_PRINTF()
					ret_value = do_fat_fsload(NULL, 0, 5, p_cLoaduImageToSDRAM);
					SESSION_DEPENDED_PRINTF_ENABLE()
					if(ret_value == 0)
					{
						// Get linux image size and save it to ENV
						ulFilesize = simple_strtoul (getenv("filesize"), NULL, 16);
						setenv("linuxfilesize", getenv("filesize"));
						DISABLE_PRINTF()
						saveenv();
						SESSION_DEPENDED_PRINTF_ENABLE()

						printf ("\n--> Update: Load %s to SDRAM (%d bytes)", (char*)p_cLoaduImageToSDRAM[4], (int)ulFilesize);

						// Check if the file fits into to the flash-partition
						if (check_imagesize (ulFilesize, ulLinuxFlashPartitionSize, LINUX_IMAGE_UPDATE, cErrorString ))
						{
							// ERROR: file is to large for the specified partition
							ShowUpdateError(p_stDHupdateINI, cErrorString, IMAGE_SIZE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							return 1;
						}

						// Calculate necessary sectors in flash for the Linux image file.
						ulBlocks = ulFilesize / ulFlashBlockSize + 0x1;

						printf ("\n--> Update: The new Linux image File needs %lu Flash blocks\n", ulBlocks);
						printf ("    e = Erase Block\n");
						printf ("    w = Write to Block\n");
						printf ("    w = Write to Block\n");
						printf ("    d = Done\n");
						printf ("    Write File to Flash:[");

						ret_value = update_flash_content (ulOSImageFlashAddress, ulSDRAMBufferAddress, ulBlocks, ulFlashBlockSize);

						if((ret_value & 0x3) == 1)
						{
							sprintf (&cErrorString[0], "\n--> Update ERROR: Erase error on block 0x%08x\n", (unsigned int)(ulOSImageFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
							ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							return 1;
						}
						else if((ret_value & 0x3) == 2)
						{
							sprintf (&cErrorString[0], "\n--> Update ERROR: Write error on block 0x%08x\n", (unsigned int)(ulOSImageFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
							ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
							return 1;
						}

						// Linux Flash Update done.
						printf ("\n--> Update: Linux Update done\n");
					}
					else
					{
						// Can't load Linux image file from Storage Device to SDRAM.
						sprintf (&cErrorString[0], "\n--> Update ERROR: No %s file found on Storage Device\n", (char*)p_cLoaduImageToSDRAM[4]);
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					// LZ 31.05.2012
					// Setup default U-Boot NAND Flash ENV for Linux
					printf ("\n--> Update: Setup ENV for Linux");
					setenv("bootcmd", "run nandbootlinux");
					DISABLE_PRINTF()
					saveenv();
					SESSION_DEPENDED_PRINTF_ENABLE()
				}
#else /* DH_SPI_FLASH */
				printf ("\n==> Update INFO: No Linux Update possible, because we are running on SPI Flash");
#endif /* DH_NAND_FLASH */
				// Try to load linux ENV setup file
				DISABLE_PRINTF()
				ret_value = do_fat_fsload(NULL, 0, 5, p_cLoaduImageENVToSDRAM);
				SESSION_DEPENDED_PRINTF_ENABLE()

				if(ret_value == 0)
				{
					printf ("\n--> Update: Load ENV setup file %s to SDRAM (%d bytes)", (char*)p_cLoaduImageENVToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));

					// Setup U-Boot ENV for Linux
					printf ("\n--> Update: Setup ENV for Linux");

					// Check if *.env file contains the script name "Set_ENV"
					if(!(memcmp ( (char*)(p_cSDRAMBufferAddress+0x20), p_cENVCompareString[0], 7)))
					{
						printf ("\n\n----------------- Print ENV Script Info: -----------------\n");
						ret_value = do_source(NULL, 0, 2, p_cRunENVScript);
						printf ("\n----------------------------------------------------------\n");
					}
					else
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: No valid %s file found on Storage Device\n", (char*)p_cLoaduImageENVToSDRAM[4]);
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], INVALID_FILE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					DISABLE_PRINTF()
					saveenv();
					SESSION_DEPENDED_PRINTF_ENABLE()
				}
				else
				{
					if(cOSFileType == UPDATE_FILE_TYPE_ENV)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: No ENV setup file %s found on Storage Device\n", (char*)p_cLoaduImageENVToSDRAM[4]);
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}
					else
					{
						printf ("\n--> Update INFO: No ENV setup file %s found on Storage Device\n", (char*)p_cLoaduImageENVToSDRAM[4]);
					}
				}
			break;

			// ***************************************************************************************
			// **************************** Bootloader Flash Image ***********************************
			// ***************************************************************************************
			case BOOTLOADER_FLASH_UPDATE:
				printf ("\n==> Update: Start to Update Bootloader Flash Image");

				// Load u-boot image file  from Storage Device to SDRAM.
				printf("\n\nLoad u-boot with: %s, %s, %s, %s, %s \n", p_cLoadUBootBinToSDRAM[0], p_cLoadUBootBinToSDRAM[1], p_cLoadUBootBinToSDRAM[2], p_cLoadUBootBinToSDRAM[3], p_cLoadUBootBinToSDRAM[4]);

				DISABLE_PRINTF()
				ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadUBootBinToSDRAM);
				SESSION_DEPENDED_PRINTF_ENABLE()

				if(ret_value == 0)
				{
					printf ("\n--> Update: Load %s to SDRAM (%d bytes)", (char*)p_cLoadUBootBinToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));

					// Calculate necessary sectors in flash for the u-boot image file.
					ulFilesize = simple_strtoul (getenv("filesize"), NULL, 16);
					ulBlocks = ulFilesize / ulFlashBlockSize + 0x1;

					// Check if the file fits into to the flash-partition
					if (check_imagesize (ulFilesize, ulBootloaderFlashPartitionSize, BOOTLOADER_FLASH_UPDATE, cErrorString ))
					{
						// ERROR: file is to large for the specified partition
						ShowUpdateError(p_stDHupdateINI, cErrorString, IMAGE_SIZE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					printf ("\n--> Update: The new u-boot image File needs %lu Flash blocks\n", ulBlocks);
					printf ("    e = Erase Block\n");
					printf ("    w = Write to Block\n");
					printf ("    d = Done\n");
					printf ("    Write File to Flash:[");

					ret_value = update_flash_content (ulBootloaderFlashAddress, ulSDRAMBufferAddress, ulBlocks, ulFlashBlockSize);

					if((ret_value & 0x3) == 1)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: Erase error on block 0x%08x\n", (unsigned int)(ulBootloaderFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}
					else if((ret_value & 0x3) == 2)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: Write error on block 0x%08x\n", (unsigned int)(ulBootloaderFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					// Bootloader Flash Update done.
					printf ("\n--> Update: U-Boot Update done\n");
				}
				else
				{
					// Can't load u-boot image file from Storage Device to SDRAM.
					sprintf (&cErrorString[0], "\n--> Update ERROR: No %s file found on Storage Device\n", (char*)p_cLoadUBootBinToSDRAM[4]);
					ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}
			break;

			// ***************************************************************************************
			// ****************************** Update SplashImage *************************************
			// ***************************************************************************************
			case SPLASHIMAGE_UPDATE:
				printf ("\n==> Update: Start to Update Splashimage Bitmap");
				// Load splash bitmap file from Storage Device to SDRAM.
				DISABLE_PRINTF()
				ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadSplashBmpToSDRAM);
				SESSION_DEPENDED_PRINTF_ENABLE()

				if(ret_value == 0)
				{
					printf ("\n--> Update: Load %s to SDRAM (%d bytes)", (char*)p_cLoadSplashBmpToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));

					// Calculate necessary sectors in flash for the splash bitmap file.
					ulFilesize = simple_strtoul (getenv("filesize"), NULL, 16);
					ulBlocks = ulFilesize / ulFlashBlockSize + 0x1;

					// Check if the file fits into to the flash-partition
					if (check_imagesize (ulFilesize, ulSplashimagePartitionSize, SPLASHIMAGE_UPDATE, cErrorString ))
					{
						// ERROR: file is to large for the specified partition
						ShowUpdateError(p_stDHupdateINI, cErrorString, IMAGE_SIZE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					printf ("\n--> Update: The new splash bitmap File needs %lu Flash blocks\n", ulBlocks);
					printf ("    e = Erase Block\n");
					printf ("    w = Write to Block\n");
					printf ("    d = Done\n");
					printf ("    Write File to Flash:[");

					ret_value = update_flash_content (ulSplashimageFlashAddress, ulSDRAMBufferAddress, ulBlocks, ulFlashBlockSize);

					if((ret_value & 0x3) == 1)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: Erase error on block 0x%08x\n", (unsigned int)(ulSplashimageFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}
					else if((ret_value & 0x3) == 2)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: Write error on block 0x%08x\n", (unsigned int)(ulSplashimageFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					// Splashimage Flash Update done.
					printf ("\n--> Update: Splashimage Update done\n");
				}
				else
				{
					// Can't load splash bitmap file from Storage Device to SDRAM.
					sprintf (&cErrorString[0], "\n--> Update ERROR: No %s file found on Storage Device\n", (char*)p_cLoadSplashBmpToSDRAM[4]);
					ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}
			break;

			// ***************************************************************************************
			// **************************** Update settings.bin File *********************************
			// ***************************************************************************************
			case SETTINGS_BIN_UPDATE:
				 printf ("\n==> Update: Start to Update settings file");
				// Load settings file from Storage Device to SDRAM.
				DISABLE_PRINTF()
				ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadSettingsBinToSDRAM);
				SESSION_DEPENDED_PRINTF_ENABLE()

				if(ret_value == 0)
				{
					printf ("\n--> Update: Load %s to SDRAM (%d bytes)", (char*)p_cLoadSettingsBinToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));

					// Calculate necessary sectors in flash for the settings file.
					ulFilesize = simple_strtoul (getenv("filesize"), NULL, 16);
					ulBlocks = ulFilesize / ulFlashBlockSize + 0x1;

					// Check if the file fits into to the flash-partition
					if (check_imagesize (ulFilesize, ulSettingsBinFlashPartitionSize, SETTINGS_BIN_UPDATE, cErrorString ))
					{
						// ERROR: file is to large for the specified partition
						ShowUpdateError(p_stDHupdateINI, cErrorString, IMAGE_SIZE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					printf ("\n--> Update: The new settings File needs %lu Flash blocks\n", ulBlocks);
					printf ("    e = Erase Block\n");
					printf ("    w = Write to Block\n");
					printf ("    d = Done\n");
					printf ("    Write File to Flash:[");

					ret_value = update_flash_content (ulSettingsBinFlashAddress, ulSDRAMBufferAddress, ulBlocks, ulFlashBlockSize);

					if((ret_value & 0x3) == 1)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: Erase error on block 0x%08x\n", (unsigned int)(ulSettingsBinFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}
					else if((ret_value & 0x3) == 2)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: Write error on block 0x%08x\n", (unsigned int)(ulSettingsBinFlashAddress + (ret_value >> 2) * ulFlashBlockSize));
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					// Settings Flash Update done.
					printf ("\n--> Update: Settings Update done\n");
				}
				else
				{
					// Can't load settings file from Storage Device to SDRAM.
					sprintf (&cErrorString[0], "\n--> Update ERROR: No %s file found on Storage Device\n", (char*)p_cLoadSettingsBinToSDRAM[4]);
					ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}
			break;


			// ***************************************************************************************
			// ****************************** Refresh DH settings ***********************************
			// ***************************************************************************************
			case REFRESH_DH_SETTINGS:
				printf ("\n==> Update: Refresh DH settings");

				if(UpdateGlobalDataDHSettings(ulSDRAMBufferAddress) == 0)
				{
					lcd_ctrl_init ((void *)(gd->fb_base));

					// Check if DHupdate.ini contains [display] section
					// If so, load progress bitmap.
					if(p_stDHupdateINI->iDisplayInfo == 1)
					{
						if(ShowBitmap(p_stDHupdateINI, PROGRESS_BITMAP, p_cStorageDevice, p_cDevicePartitionNumber) != 0)
						{
							printf ("\n--> Update INFO: Progress bitmap %s not found", p_stDHupdateINI->p_cFileNameProgressBmp);
						}
					}

					lcd_enable();

					// Refresh DH settings done.
					printf ("\n--> Update: Refresh DH settings done\n");
				}
				else
				{
					// No valid DH settings FIle found on SDRAM
					sprintf (&cErrorString[0], "\n--> Update ERROR: No valid DH settings file found on RAM (Note: Set \"refresh\" string after \"settings\" update)\n");
					ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}
			break;
			// ***************************************************************************************
			// **************************** Update eeprom.bin File *********************************
			// ***************************************************************************************
			case EEPROM_SETTINGS_UPDATE:
				 printf ("\n==> Update: Start to Update display adapter eeprom content");
				// Load eeprom file from Storage Device to SDRAM.
				DISABLE_PRINTF()
				ret_value = do_fat_fsload(NULL, 0, 5, p_cLoadEepromBinToSDRAM);
				SESSION_DEPENDED_PRINTF_ENABLE()

				if(ret_value == 0)
				{
					printf ("\n--> Update: Load %s to SDRAM (%d bytes)", (char*)p_cLoadEepromBinToSDRAM[4], (int)simple_strtoul (getenv("filesize"), NULL, 16));

					// Get eeprom.bin filesize
					ulFilesize = simple_strtoul (getenv("filesize"), NULL, 16);

					// Check if the file fits into to the eeprom-partition
					if (check_imagesize (ulFilesize, DHCOM_DISPLAY_SETTINGS_SIZE, EEPROM_SETTINGS_UPDATE, cErrorString ))
					{
						// ERROR: file is to large for the specified partition
						ShowUpdateError(p_stDHupdateINI, cErrorString, IMAGE_SIZE_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					printf ("\n--> Update: The new eeprom File needs %lu Bytes\n", ulFilesize);
					printf ("    w = Write to addr\n");
					printf ("    d = Done\n");
					printf ("    Write File to EEPROM:[");


					ret_value = update_eeprom_content (0x00, ulSDRAMBufferAddress, ulFilesize);

					if(ret_value != 0)
					{
						sprintf (&cErrorString[0], "\n--> Update ERROR: EEPROM write error\n");
						ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FLASH_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}

					// Settings Flash Update done.
					printf ("\n--> Update: Settings Update done\n");
				}
				else
				{
					// Can't load settings file from Storage Device to SDRAM.
					sprintf (&cErrorString[0], "\n--> Update ERROR: No %s file found on Storage Device\n", (char*)p_cLoadEepromBinToSDRAM[4]);
					ShowUpdateError(p_stDHupdateINI, &cErrorString[0], FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}
			break;
			// ***************************************************************************************
			// ****************************** Reset Board ********************************************
			// ***************************************************************************************
			case EXECUTE_RESET:
				printf ("\n==> Update: Reset Board");
				if(p_stDHupdateINI->iUpdateCounter != 1)
				{
					printf ("\n==> Update ERROR: Skipped reset!!!");

					// Can't do reset.
					sprintf (cErrorString, "\n==> Update ERROR: Reset has to be the last action in DHupdate.ini (Danger of infinite Update-Loop)!!!");
					ShowUpdateError(p_stDHupdateINI, cErrorString, DHUPDATE_INI_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}
				// do reset
				run_command("reset",0);
			break;
		   // ***************************************************************************************
		   // ****************************** Update root-filesystem *********************************
		   // ***************************************************************************************
			case NAND_ROOTFS_UPDATE:
				 printf ("\n==> Update: Update NAND-Flash linux rootfile-system\n");

				if(p_stDHupdateINI->iUpdateCounter != 1)
				{
					// Check if last update is to execute a reset
					if((p_stDHupdateINI->iUpdateCounter == 2) && (strcmp(p_stDHupdateINI->stDHUpdateInfo[iUpdateLoopCounter + 1].p_cUpdateType, "reset") == 0))
					{
						iDoReset = 1;
					}
					else
					{
						printf ("\n==> Update ERROR: Skipped root-filesystem update!!!");

						// Can't do rootfs-update.
						sprintf (cErrorString, "\n==> Update ERROR: The root-filesystem update has to be the last update to run (except reset)!!!");
						ShowUpdateError(p_stDHupdateINI, cErrorString, DHUPDATE_INI_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
						return 1;
					}
				}

				// Setup kernel command line to run update kernel - Do not save to NAND !!
				// Required informations: updatemedia=usbhost1, mtd=6, rootfs.tar.gz, error.bmp, progress.bmp, done.bmp,

		        // Check if DHupdate.ini contains [display] section
		        if(p_stDHupdateINI->iDisplayInfo != 1)
		        {
		        	p_stDHupdateINI->p_cFileNameErrorBmp = "none";
		        	p_stDHupdateINI->p_cFileNameProgressBmp = "none";
		        	p_stDHupdateINI->p_cFileNameOkBmp = "none";
		        }

		        // Check if DHupdate.ini contains [display] section
		        if(p_stDHupdateINI->iLedInfo == 1)
		        {
		        	iLedGPIO = get_DHCOMUpdateLED();
		        }
		        else
		        {
			        iLedGPIO = -1;
			    }

				sprintf(cBootargsString, "media=%s;dev=%s;rootfs=%s;err=%s;prog=%s;done=%s;gpio=%i;reset=%i", p_cStorageDevice, p_cDevicePartitionNumber,
						p_cRootfsFilename, p_stDHupdateINI->p_cFileNameErrorBmp, p_stDHupdateINI->p_cFileNameProgressBmp, p_stDHupdateINI->p_cFileNameOkBmp, iLedGPIO, iDoReset);

				if(0 != setenv("bootargs", cBootargsString ))
				{
					sprintf (cErrorString, "\n==> Update ERROR: Setup update-kernel commandline error!");
					ShowUpdateError(p_stDHupdateINI, cErrorString, INTERNAL_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}

				// Setup 'bootcmd' to boot update kernel - Do not save to NAND !!
				if(0 != setenv ("bootcmd", "bootm"))
				{
					sprintf (cErrorString, "\n==> Update ERROR: Setup update-kernel bootcmd error!");
					ShowUpdateError(p_stDHupdateINI, cErrorString, INTERNAL_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}

				// Setup 'bootcmd' to boot update kernel - Do not save to NAND !!
				if(0 != setenv ("loadaddr", p_cLoadUpdateKernelToSDRAM[3]))
				{
					sprintf (cErrorString, "\n==> Update ERROR: Setup update-kernel loadaddr error!");
					ShowUpdateError(p_stDHupdateINI, cErrorString, INTERNAL_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}

				printf("\n\nLoad Kernel with: %s, %s, %s, %s, %s \n", p_cLoadUpdateKernelToSDRAM[0], p_cLoadUpdateKernelToSDRAM[1], p_cLoadUpdateKernelToSDRAM[2], p_cLoadUpdateKernelToSDRAM[3], p_cLoadUpdateKernelToSDRAM[4]);

				// load update kernel to sdram
				if(0 != do_fat_fsload(NULL, 0, 5, p_cLoadUpdateKernelToSDRAM))
				{
					printf ("\n==> Update ERROR: Image of update-kernel not found!");

					// Can't find update-kernel.
					sprintf (cErrorString, "\n==> Update ERROR: The update-kernel has to be named \"uImage.update\"!");
					ShowUpdateError(p_stDHupdateINI, cErrorString, FILE_NOT_FOUND_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}

				// TODO: Check if rootfs is available



				// boot update kernel
				printf ("\n==> Update: Booting update-kernel ... \n \n");
				run_command("boot",0);
				// After booting the update kernel we do not return !!!

			break;



			// ***************************************************************************************
			// *** No Update Event is defined for the Update Argument in the DHupdate.ini file ***
			// ***************************************************************************************
			default:
				if(iUpdateViaDHupdateIniFile == 1)
				{
					sprintf (&cErrorString[0], "\n--> Update ERROR: No Update Event defined for the Argument: %s\n", (char*)p_vUpdateArgument);
					ShowUpdateError(p_stDHupdateINI, &cErrorString[0], DHUPDATE_INI_ERROR, iUpdateViaDHupdateIniFile, p_cStorageDevice, p_cDevicePartitionNumber);
					return 1;
				}
				else
				{
					printf ("\n--> Update ERROR: No Update Event defined for this command line Argument\n");
					// Don't Start OS after Update error
					goto usage;
				}
		}

		if(iUpdateViaDHupdateIniFile == 1)
		{
			// Start next update
			iUpdateLoopCounter++;
		}

		// Number of updates
		p_stDHupdateINI->iUpdateCounter--;

	} while(p_stDHupdateINI->iUpdateCounter > 0);
	// ==> End UPDATE LOOP
	// *************************************************************************************************************************************************

	// Check if DHupdate.ini contains [display] section
	if(p_stDHupdateINI->iDisplayInfo == 1)
	{
		if(ShowBitmap(p_stDHupdateINI, END_BITMAP, p_cStorageDevice, p_cDevicePartitionNumber) != 0)
		{
			printf ("\n--> Update INFO: End bitmap %s not found\n", p_stDHupdateINI->p_cFileNameOkBmp);
		}
	}
    
    if(iUpdateViaDHupdateIniFile == 1)
    {
        // Don't Start OS after Update
        CLEAR_DH_GD_FLAG(BOOT_AFTER_UPDATE_FLAG);
    }
    return 0;
    
usage:
	cmd_usage(cmdtp);  
    return 1;     
}

//------------------------------------------------------------------------------
//
//  Function:  do_DHCOMupdate
//
//  DHCOM U-Boot Update function
//  Description:	Scan for available storage devices
//
//  Return value:   0 = No error
//                  1 = error
//
int do_DHCOMupdate (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	#ifdef CONFIG_AM3517
	    char *p_cMicroSDCardInit[2]             = {"mmcinit","2"};
	    char *p_cMMCCardInit[2]                 = {"mmcinit","1"};
		#define ARGC_COUNT						2
	#else
	    char *p_cMicroSDCardInit[1]             = {"mmcinit"};
	    char *p_cMMCCardInit[1]                 = {"mmcinit"};
		#define ARGC_COUNT						1
	#endif
    char *p_cUSBInit[2]                     = {"usb","start"};
    char *p_cCompareString[1]               = {"##DHCOMupdate##\0"};
    char *cmd;
    
    char cSDCardStorageDevice[4]            = {"mmc\0"};
    char cSDCardDevicePartitionNumber[4]    = {"1:1\0"};    
    char cUSBStorageDevice[4]               = {"usb\0"};    
    char cUSBHost1DevicePartitionNumber[4]  = {"0:1\0"};   
    char cDHupdateIniInvalid[8]             = {"INVALID\0"};    
    
    int ret_value = 0;
    int iUpdateDeviceCounter = 0;
    int iUpdateSuccess = 0;
    int iDHupdateIniFound = 0;
    int iStartUpAutomaticallyUpdate = 0;
    
    updateinfo_t stDHupdateINI;
    
    // Don't Show anything on the display and GPIO's
    stDHupdateINI.iDisplayInfo = 0;
    stDHupdateINI.iLedInfo = 0;    
    
    if(argc > 1)
    {
        // Check if this is a StartUp automatically Update
        // In this case, check which storage devices are enabled in the settings block
        cmd = argv[1];
            
        if (strcmp(cmd, "auto") == 0) 
        {
            iStartUpAutomaticallyUpdate = 1;
        }
    }
    
    // Check for available storage devices depending on the flags of the settings block
    for(iUpdateDeviceCounter = 1; (iUpdateDeviceCounter <= 4) && (iUpdateSuccess == 0) && (iDHupdateIniFound == 0); iUpdateDeviceCounter++)
    {
        stDHupdateINI.p_cValidMarker = &cDHupdateIniInvalid[0];
    
        switch(iUpdateDeviceCounter)
        {
            // MicroSD Card Slot
            case 1:
                if(((iStartUpAutomaticallyUpdate == 1) && (gd->dh_board_settings.wHWConfigFlags & UPDATE_VIA_MICROSD_SLOT)) || (iStartUpAutomaticallyUpdate == 0))
                {
                    // Initialize MMC/SD -Card.
                    printf ("--> Update: Try to initialize MicroSD - Card:"); 
                    //DISABLE_PRINTF()
                    ret_value = do_mmcinit(NULL, 0, ARGC_COUNT, p_cMicroSDCardInit);
                    SESSION_DEPENDED_PRINTF_ENABLE()
                    
                    // Call update function, if MicroSD init returns success
                    if(ret_value == 0)
                    {
                        ret_value = DHCOMupdate (cmdtp, argc, argv, &stDHupdateINI, &cSDCardStorageDevice[0], &cSDCardDevicePartitionNumber[0]);
                        
                        // Update success
                        if(ret_value == 0)
                        {
                            // Don't try to start update from another storage device
                            iUpdateSuccess = 1;
                        }
                        
                        // DHupdate.ini File found on storage device, but Update fails with error
                        else if(!(memcmp ( (char*)stDHupdateINI.p_cValidMarker, p_cCompareString[0], 15)))
                        {
                            // Don't try to start update from another storage device
                            iDHupdateIniFound = 1;
                        }

                        // Command line Update with error --> New line is neccessary
                        else
                        {
                            printf ("\n");
                        }
                    }
                    else
                    {
                        printf ("\n--> Update INFO: No MicroSD - Card detected!\n");
                    }
                }
            break;
            
            // SD/MMC Card Slot
            case 2:
                if(((iStartUpAutomaticallyUpdate == 1) && (gd->dh_board_settings.wHWConfigFlags & UPDATE_VIA_SD_MMC_SLOT)) ||
                   (iStartUpAutomaticallyUpdate == 0))
                {
					#ifdef CONFIG_MX25
                    // Select extern MMC/SD Slot
                    mxc_set_gpio_dataout(MX25_PIN_UPLL_BYPCLK, 1);      // Set Pin to High
					#endif

                    // Initialize MMC/SD -Card.
                    printf ("--> Update: Try to initialize MMC/SD - Card:"); 
                    DISABLE_PRINTF()
                    ret_value = do_mmcinit(NULL, 0, ARGC_COUNT, p_cMMCCardInit);
                    SESSION_DEPENDED_PRINTF_ENABLE()
                    
                    // Call update function, if MMC/SD init returns success
                    if(ret_value == 0)
                    {
                        ret_value = DHCOMupdate (cmdtp, argc, argv, &stDHupdateINI, &cSDCardStorageDevice[0], &cSDCardDevicePartitionNumber[0]);
                        
                        // Update success
                        if(ret_value == 0)
                        {
                            // Don't try to start update from another storage device
                            iUpdateSuccess = 1;
                        }
                        
                        // DHupdate.ini File found on storage device, but Update fails with error
                        else if(!(memcmp ( (char*)stDHupdateINI.p_cValidMarker, p_cCompareString[0], 15)))
                        {
                            // Don't try to start update from another storage device
                            iDHupdateIniFound = 1;
                        }

                        // Command line Update with error --> New line is neccessary
                        else
                        {
                            printf ("\n");
                        }
                    }      
                    else
                    {
                        printf ("\n--> Update INFO: No MMC/SD - Card detected!\n");
                    }
                    
					#ifdef CONFIG_MX25
                    // Select MicroSD Slot (U-Boot default)
                    mxc_set_gpio_dataout(MX25_PIN_UPLL_BYPCLK, 0);      // Set Pin to Low
					#endif
                }
            break;
            
            // USB Host 1 Port (on AM35x 436-200 and newer USB Host 2 can be used too)
            case 3:
                if(((iStartUpAutomaticallyUpdate == 1) && (gd->dh_board_settings.wHWConfigFlags & UPDATE_VIA_USB_HOST_1_PORT)) ||
                   (iStartUpAutomaticallyUpdate == 0))
                {
                // NOTE: On AM3517 you can choose between Host and OTG
				#if defined(CONFIG_MX25) || (defined(CONFIG_AM3517) && defined(CONFIG_USB_EHCI))
                   // Initialize USB Stick
                    printf ("--> Update: Try to initialize USB Stick on Host Port:");
                    DISABLE_PRINTF()
                    ret_value = do_usb(NULL, 0, 2, p_cUSBInit);
                    SESSION_DEPENDED_PRINTF_ENABLE()

                    // Call update function, if USB init returns success
                    if(ret_value == 0)
                    {
                        ret_value = DHCOMupdate (cmdtp, argc, argv, &stDHupdateINI, &cUSBStorageDevice[0], &cUSBHost1DevicePartitionNumber[0]);
                        
                        // Update success
                        if(ret_value == 0)
                        {
                            // Don't try to start update from another storage device
                            iUpdateSuccess = 1;
                        }
                        
                        // DHupdate.ini File found on storage device, but Update fails with error
                        else if(!(memcmp ( (char*)stDHupdateINI.p_cValidMarker, p_cCompareString[0], 15)))
                        {
                            // Don't try to start update from another storage device
                            iDHupdateIniFound = 1;
                        }

                        // Command line Update with error --> New line is neccessary
                        else
                        {
                            printf ("\n");
                        }
                    } 
                    else
                    {
                        printf ("\n--> Update INFO: No USB Stick detected!\n");
                    }
				#endif
                }
            break;
            
            // USB OTG Port
            case 4:
                if(((iStartUpAutomaticallyUpdate == 1) && (gd->dh_board_settings.wHWConfigFlags & UPDATE_VIA_USB_OTG_PORT)) ||
                   (iStartUpAutomaticallyUpdate == 0))
                {
                // NOTE: Not implemented on DHCOM i.MX25 Core Module

				#if defined(CONFIG_AM3517) && defined(CONFIG_MUSB_HCD)
                	// Initialize USB Stick
                	printf ("--> Update: Try to initialize USB Stick on OTG:");
                	DISABLE_PRINTF()
                	ret_value = do_usb(NULL, 0, 2, p_cUSBInit);
                	SESSION_DEPENDED_PRINTF_ENABLE()

                	// Call update function, if USB init returns success
                	if(ret_value == 0)
                	{
                		ret_value = DHCOMupdate (cmdtp, argc, argv, &stDHupdateINI, &cUSBStorageDevice[0], &cUSBHost1DevicePartitionNumber[0]);

                		// Update success
                		if(ret_value == 0)
                		{
                			// Don't try to start update from another storage device
                	        iUpdateSuccess = 1;
                		}

                		// DHupdate.ini File found on storage device, but Update fails with error
                	    else if(!(memcmp ( (char*)stDHupdateINI.p_cValidMarker, p_cCompareString[0], 15)))
                	    {
							// Don't try to start update from another storage device
                	    	iDHupdateIniFound = 1;
                	    }

                        // Command line Update with error --> New line is neccessary
                        else
                        {
                            printf ("\n");
                        }
                	}
                	else
                	{
                		printf ("\n--> Update INFO: No USB Stick detected!\n");
                	}
				#endif
                }
            break;
        } 
    }
    
    // Check if DHupdate.ini contains [led] section
    // If so, init. update led
    if(stDHupdateINI.iLedInfo == 1)
    {
        // Activate Update GPIO
        DeacitvateUpdateGPIO(&stDHupdateINI);
    }

    return 0;
}

/* ====================================================================== */

U_BOOT_CMD(
  update,      3,      0,      do_DHCOMupdate,
  "Update Flash content from Storage Device.",
  "- Starts update with DHupdate.ini file\n"
  "         Necessary files on Storage Device:\n"
  "           - DHupdate.ini File\n"
  "           - nk.bin or flash.bin File, ...\n"
  "update <type> [filename] - Starts update without DHupdate.ini file\n"
  "         Types:\n"
  "           - wince = WinCE image update (default file names nk.gz or nk.bin)\n"
  "           - linux = Linux image update (default file name uImage)\n"
  "           - bootloader = Bootloader update (default file name u-boot.bin)\n"  
  "           - splash = Bootbitmap update (default file name splash.bmp)\n"
  "           - settings = Settings block update (default file name settings.bin)\n"
);
#endif/* CONFIG_CMD_UPDATECE */
