#include <common.h>
#include <command.h>
#include <asm/arch/sys_proto.h>

/*
 * Some information to understand MMU operations
 *
 *    MCR    <co-pro>, <op>, <ARM reg>, <co-pro reg>, <co-pro reg2>, <op2>
 *    -> The instruction MCR transfers an ARM register to a co-processor register.
 *
 *    MRC    <co-pro>, <op>, <ARM reg>, <co-pro reg>, <co-pro reg2>, <op2>
 *    -> The instruction MRC transfers a co-processor register to an ARM register.
 *
 *    -----------------------------------------------------------------------------------
 *
 *    extended assembler syntax:
 *
 *    volatile asm ( assembler template
 *         			: output operands                  // optional
 *        			: input operands                   // optional
 *         			: list of clobbered registers      // optional (clobbered <-> overwritten)
 *         			);
 *
 *    -----------------------------------------------------------------------------------
 *
 *    Register 	CP15 (Co-Processor 15)			Cross reference
 *
 *    TLB Type Register 						c0, TLB Type Register on page 3-21
 *    Control Register 							c1, Control Register on page 3-44
 *    Nonsecure Access Control Register 		c1, Nonsecure Access Control Register on page 3-56
 *    Translation Table Base Register 0 		c2, Translation Table Base Register 0 on page 3-57
 *    Translation Table Base Register 1 		c2, Translation Table Base Register 1 on page 3-59
 *    Translation Table Base Control Register	c2, Translation Table Base Control Register on page 3-60
 *    Domain Access Control Register 			c3, Domain Access Control Register on page 3-62
 *    Data Fault Status Register (DFSR) 		c5, Data Fault Status Register on page 3-63
 *    Instruction Fault Status Register (IFSR)	c5, Instruction Fault Status Register on page 3-65
 *    Data Fault Address Register (DFAR) 		c6, Data Fault Address Register on page 3-67
 *    Instruction Fault Address Register (IFAR) c6, Instruction Fault Address Register on page 3-68
 *    TLB operations 							c8, TLB operations on page 3-75
 *    TLB Lockdown Registers 					c10, TLB Lockdown Registers on page 3-98
 *    Primary Region Remap Register 			c10, Memory Region Remap Registers on page 3-100
 *    Normal Memory Remap Register 				c10, Memory Region Remap Registers on page 3-100
 *    FCSE PID Register 						c13, FCSE PID Register on page 3-120
 *    Context ID Register 						c13, Context ID Register on page 3-122
 *
 *
 *    -----------------------------------------------------------------------------------
 *
 *    Layout of a TLB entry
 *
 *    Bits 31:20 = Base address for the section (i.e 0MB for the first entry, 1MB for the second and so on)
 *    Bits 19:12 = Don't care - should be 0
 *    Bits 11:10 = Access Permission - write as 11 to allow all accesses
 *    Bit 9 = Should be 0
 *    Bits 8:5 = Domain 0 - make it 0 for simplicity as you did above.
 *    Bit 4 = Must be '1'
 *    Bits 3:2 = '11' Cacheable and bufferable.
 *    Bits 1:0 Must be '10' for a section
 *
 */


void mmu_switch_on (void)
{

	/* Enable caches */
	enable_caches();
	debug("[MMU] switched on!\n");
}

void mmu_switch_off (void)
{
	/*
	 * Turn off I-cache and invalidate it
	 */
//	icache_disable();
//	invalidate_icache_all();

	/*
	 * turn off D-cache
	 * dcache_disable() in turn flushes the d-cache and disables MMU
	 */
	dcache_disable();

	/*
	 * After D-cache is flushed and before it is disabled there may
	 * be some new valid entries brought into the cache. We are sure
	 * that these lines are not dirty and will not affect our execution.
	 * (because unwinding the call-stack and setting a bit in CP15 SCTRL
	 * is all we did during this. We have not pushed anything on to the
	 * stack. Neither have we affected any static data)
	 * So just invalidate the entire d-cache again to avoid coherency
	 * problems for kernel
	 */
	invalidate_dcache_all();

	debug("[MMU] switched off!\n");
}

int do_mmu(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	switch (argc)
	{
	case 2:

		if (strcmp(argv[1], "on") == 0)
		{
			mmu_switch_on();
			/* printf("Switched on mmu ... \n"); */

		}
		else if (strcmp(argv[1], "off") == 0)
		{
			/* printf("Switch off mmu ... \n"); */
			mmu_switch_off();


		} else
		{
			printf("Usage:\n%s\n", cmdtp->usage);
		}
		break;

	default:
		printf("Usage:\n%s\n", cmdtp->usage);
		return 1;
		break;
	}
	return 0;
}

U_BOOT_CMD(	mmu, 2, 1, do_mmu,
	"MMU enable/disable",
	"on\n"
	"mmu off\n");


