/*
 * File: cmd_testroutines.c
 *
 * (C) Copyright 2011-2012
 * Ludwig Zenz, DH electronics GmbH , lzenz@dh-electronics.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * brief:		Collection of different HW-tests
 * description:	This test-routines are only for development.
 * 				They can be excluded for customer versions/builds (#undef CONFIG_CMD_TESTROUTINES).
 *
 */

#include <common.h>
#include <command.h>

#include <linux/ctype.h>
#include <asm/io.h>
#include <asm/errno.h>
#include <asm/arch/gpio.h>
#include <asm/arch/mux.h>
#include <asm/arch/omap3-regs.h>
#include <asm/arch/mmc_omap3.h>
#include <asm/arch/omap3_dma.h>

#include <linux/mtd/mtd.h>

#include <watchdog.h>
#include <malloc.h>
#include <asm/byteorder.h>
#include <jffs2/jffs2.h>
#include <nand.h>
#include <s_record.h>
#include <ata.h>
#include <part.h>
#include <net.h>
#include <fat.h>

#include <i2c.h>		   // use i2c functionality
#include <mcp7941x_rtc.h>  // defines to use the rtc mcp79411

#include <dh_settings.h>

DECLARE_GLOBAL_DATA_PTR;

#ifdef	CMD_MEM_DEBUG
#define	PRINTF(fmt,args...)	printf (fmt ,##args)
#else
#define PRINTF(fmt,args...)
#endif

#if defined(CONFIG_CMD_TESTROUTINES)
//#define DEVELOPERTESTS

extern const unsigned long padconf_gpio[192]; // necessary to access padconf registers for each gpio

/*
 * configure a pad to gpio-mode (UNSAFE)
 */
int configure_gpio_testdh(int gpio)
{
	int error = 0;

	printf("\ngpio %i configure to gpio-mode!\n\n", gpio);

	error = omap_request_gpio(gpio);
	if (error == -EINVAL) { printf("Enter a number between 0 and 191!\n\n"); return -1; }
	if (padconf_gpio[gpio] == 0)	{ printf("This gpio is not available!\n\n"); return -1; }

	writew((IDIS | PTD | DIS | M4),padconf_gpio[gpio]);

	return 0;
}


/*
 * set a gpio output to specified state
 */
int set_gpio_testdh(int gpio, int state)
{
	int error = 0;
	unsigned short pad_conf = 0;

	printf("\ngpio %i configure to output-mode and set to %i\n\n", gpio, state);

	if (state != 0) {state = 1;} // only 0 and 1 are possible values

	error = omap_request_gpio(gpio);
	if (error == -EINVAL) { printf("Enter a number between 0 and 191!\n\n"); return -1; }
	if (padconf_gpio[gpio] == 0)	{ printf("This gpio is not available!\n\n"); }

	pad_conf = readw(padconf_gpio[gpio]);

	if ((pad_conf & 0x07) != M4 ) { printf("This pin is not configured for gpio-mode!\n\n"); return -1; }

	omap_set_gpio_direction(gpio,0);
	omap_set_gpio_dataout(gpio,state);

	return 0;
}

/*
 * set a gpio to input and read the state
 */
int get_gpio_testdh(int gpio)
{
	int error = 0;
	int input_st = 0;
	unsigned short pad_conf = 0;

	error = omap_request_gpio(gpio);
	if (error == -EINVAL) { printf("Enter a number between 0 and 191!\n\n"); return -1; }
	if (padconf_gpio[gpio] == 0)	{ printf("This gpio is not available!\n\n"); }

	pad_conf = readw(padconf_gpio[gpio]);

	if ((pad_conf & 0x07) != M4 ) { printf("This pin is not configured for gpio-mode!\n\n"); return -1; }

	omap_set_gpio_direction(gpio,1);
	input_st = omap_get_gpio_datain(gpio);

	printf("\ngpio %i is input and has state %i\n\n", gpio, input_st);

	return 0;
}


/*
 * Test the connected NAND-Flash
 */
int nand_testdh(void)
{
	vu_long	*start_ddr2, *end_ddr2, *nandpage_length;
	nand_erase_options_t opts;
	vu_long	len, offset, pattern;
	nand_info_t *nand;
	size_t size;
	loff_t off;
	int index, err = 0;

	static const ulong bitpattern[] = {
		0x55555555,	/* alternating 0/1 */
		0xaaaaaaaa,	/* alternating 1/0 */
	};


	//--------------------------------------------------------------------------
	// set start and stop address from board configuration
	start_ddr2 = (ulong *)CONFIG_SYS_MEMTEST_START;
	end_ddr2 = (ulong *)(CONFIG_SYS_MEMTEST_END);
	nandpage_length = (ulong *)(CONFIG_SYS_NANDTEST_LENGTH);
	len = ((ulong)nandpage_length)/sizeof(vu_long);

	size=(size_t)nandpage_length;
	off=(size_t)CONFIG_SYS_NANDTEST_START;
	//--------------------------------------------------------------------------
	// check if we have enough space in ram
	if (&(start_ddr2[2*len]) > end_ddr2)
	{
		printf("\nError: Configure a bigger memtest-space for nand-flash testing!!\n\n");
		return -1;
	}

	//--------------------------------------------------------------------------
	// operate on the current device
	if (nand_curr_device < 0 || nand_curr_device >= CONFIG_SYS_MAX_NAND_DEVICE || !nand_info[nand_curr_device].name)
	{
		puts("\nno devices available\n");
		return -1;
	}
	nand = &nand_info[nand_curr_device];

	for (index = 0; index < 2; index++) // do this once for every pattern
	{

		//--------------------------------------------------------------------------
		// create a block with test-pattern in ddr-ram
		// this has to have the size of one block of nand-flash
		pattern = bitpattern[index]; // choose first pattern
		for (offset = 0; offset < len; offset++)
		{
			start_ddr2[offset] = pattern;
		}

		//--------------------------------------------------------------------------
		// erase nand-flash

		// set erase options
		memset(&opts, 0, sizeof(opts));
				opts.offset = CONFIG_SYS_NANDTEST_START;
				opts.length = CONFIG_SYS_NANDTEST_LENGTH;
				opts.quiet  = 0;

		PRINTF("%s:%d: erase: start 0x%p length 0x%p\n",	__FUNCTION__, __LINE__, opts.offset, opts.length);
		if(nand_erase_opts(nand, &opts))
		{
			printf("\nError while erasing nand!\n");
			return -1;
		}

		udelay(1000); // Delay before writing

		//--------------------------------------------------------------------------
		// write to nand-flash
		PRINTF("%s:%d: write: start 0x%p length 0x%p\n",	__FUNCTION__, __LINE__, off, size);
		if(nand_write_skip_bad(nand, off, &size, (u_char *)start_ddr2))
		{
			printf("\nError while writing to nand!\n");
			return -1;
		}

		//--------------------------------------------------------------------------
		// read from nand-flash
		PRINTF("%s:%d: read: start 0x%p length 0x%p\n",	__FUNCTION__, __LINE__, off, size);
		if(nand_read_skip_bad(nand, off, &size, (u_char *)&(start_ddr2[len])))
		{
			printf("\nError while reading from nand!\n");
			return -1;
		}

		//--------------------------------------------------------------------------
		// compare the original with read value
		for (offset = 0; offset < len; offset++)
		{
			//if(start_ddr2[offset] != (&(start_ddr2[len]))[offset])
			if(pattern != (&(start_ddr2[len]))[offset])
			{
				printf("\nError!");
				err++;
			}
		}
	}
	return err;
}

/*
 * Perform a memory test. This routine is similar to the more complete alternative test which can be configure
 * with CONFIG_SYS_ALT_MEMTEST for the mtest command. The complete test loops until
 * interrupted by a failure of one of the sub-tests or test is finished.
 */
int ddr2_testdh(void)
{
	vu_long	*addr, *start, *end;
	ulong	val;
	ulong	readback;
	ulong	errs = 0;
	int iterations = 1;
	int iteration_limit;

	vu_long	len;
	vu_long	offset;
	vu_long	test_offset;
	vu_long	pattern;
	vu_long	temp;
	vu_long	anti_pattern;
	vu_long	num_words;
	vu_long dummy ;
	int	j;

	static const ulong bitpattern[] = {
		0x00000001,	/* single bit */
		0x00000003,	/* two adjacent bits */
		0x00000007,	/* three adjacent bits */
		0x0000000F,	/* four adjacent bits */
		0x00000005,	/* two non-adjacent bits */
		0x00000015,	/* three non-adjacent bits */
		0x00000055,	/* four non-adjacent bits */
		0xaaaaaaaa,	/* alternating 1/0 */
	};

	// set start and stop address from board configuration
	start = (ulong *)CONFIG_SYS_MEMTEST_START;
	end = (ulong *)(CONFIG_SYS_MEMTEST_END);

	// choose test pattern
	pattern = 0;

	// configure hardcoded iteration limit
	iteration_limit = 1;

	printf ("Testing %08x ... %08x:\n", (uint)start, (uint)end);
	PRINTF("%s:%d: start 0x%p end 0x%p\n",	__FUNCTION__, __LINE__, start, end);

	for (;;) {

		// check if count of test-loops reached
		if (iteration_limit && iterations > iteration_limit)
		{
			printf("Tested %d iteration(s) with %lu errors.\n",	iterations-1, errs);
			return errs != 0;
		}

		printf("Iteration: %6d\r", iterations);
		PRINTF("\n");
		iterations++;

		/*
		 * Data line test:
		 * write a pattern to the first location, write the 1's complement to a 'parking'
		 * address (changes the state of the data bus so a floating bus doen't give a false OK), and then
		 * read the value back. Note that we read it back into a variable because the next time we read it,
		 * it might be right (been there, tough to explain to the quality guys why it prints a failure when the
		 * "is" and "should be" are obviously the same in the error message).
		 *
		 * Rather than exhaustively testing, we test some
		 * patterns by shifting '1' bits through a field of
		 * '0's and '0' bits through a field of '1's (i.e.
		 * pattern and ~pattern).
		 */
		addr = start;
		for (j = 0; j < sizeof(bitpattern)/sizeof(bitpattern[0]); j++)
		{
			val = bitpattern[j];
			for(; val != 0; val <<= 1)
			{
				*addr  = val;
				dummy  = ~val; /* clear the test data off of the bus */
				readback = *addr;
				if(readback != val)
				{
					printf ("FAILURE (data line): expected %08lx, actual %08lx\n", val, readback);
					errs++;
				}
				*addr  = ~val;
				dummy  = val;
				readback = *addr;
				if(readback != ~val)
				{
					printf ("FAILURE (data line): Is %08lx, should be %08lx\n",	readback, ~val);
					errs++;
				}
			}
		}

		/*
		 * Based on code whose Original Author and Copyright
		 * information follows: Copyright (c) 1998 by Michael
		 * Barr. This software is placed into the public
		 * domain and may be used for any purpose. However,
		 * this notice must not be changed or removed and no
		 * warranty is either expressed or implied by its
		 * publication or distribution.
		 */

		/*
		 * Address line test
		 *
		 * Description: Test the address bus wiring in a memory region by performing a walking
		 *              1's test on the relevant bits of the address and checking for aliasing.
		 *              This test will find single-bit address failures such as stuck -high,
		 *              stuck-low, and shorted pins. The base address and size of the region are
		 *              selected by the caller.
		 *
		 * Notes:	For best results, the selected base address should have enough LSB 0's to
		 *              guarantee single address bit changes. For example, to test a 64-Kbyte
		 *              region, select a base address on a 64-Kbyte boundary. Also, select the
		 *              region size as a power-of-two if at all possible.
		 */
		len = ((ulong)end - (ulong)start)/sizeof(vu_long);
		pattern = (vu_long) 0xaaaaaaaa;
		anti_pattern = (vu_long) 0x55555555;

		PRINTF("%s:%d: length = 0x%.8lx\n",	__FUNCTION__, __LINE__,	len);
		/*
		 * Write the default pattern at each of the
		 * power-of-two offsets.
		 */
		for (offset = 1; offset < len; offset <<= 1)
		{
			start[offset] = pattern;
		}

		/*
		 * Check for address bits stuck high.
		 */
		test_offset = 0;
		start[test_offset] = anti_pattern; // change state of lsb-address-line
										   // write anti-pattern if lsb-address is defect you would write to offset=1 and not test_offset=0

		for (offset = 1; offset < len; offset <<= 1)
		{
			temp = start[offset];
			if (temp != pattern)
			{
			  printf ("\nFAILURE: Address bit stuck high @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx\n", (ulong)&start[offset], pattern, temp);
			  errs++;
			}
		}
		start[test_offset] = pattern;
		WATCHDOG_RESET();

		/*
		 * Check for addr bits stuck low or shorted.
		 */
		for (test_offset = 1; test_offset < len; test_offset <<= 1)
		{
			start[test_offset] = anti_pattern;

			for (offset = 1; offset < len; offset <<= 1)
			{
			  temp = start[offset];
			  if ((temp != pattern) && (offset != test_offset))
			  {
				printf ("\nFAILURE: Address bit stuck low or shorted @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx\n", (ulong)&start[offset], pattern, temp);
				errs++;
			  }
			}
			start[test_offset] = pattern;
		}

		/*
		 * Description:
		 * Test the integrity of a physical memory device by performing an
		 * increment/decrement test over the entire region. In the process every
		 * storage bit in the device is tested as a zero and a one. The base address
		 * and the size of the region are selected by the caller.
		 */

		num_words = ((ulong)end - (ulong)start)/sizeof(vu_long) + 1;

		/*
		 * Fill memory with a known pattern.
		 */
		for (pattern = 1, offset = 0; offset < num_words; pattern++, offset++)
		{
			WATCHDOG_RESET();
			start[offset] = pattern;
		}

		/*
		 * Check each location and invert it for the second pass.
		 */
		for (pattern = 1, offset = 0; offset < num_words; pattern++, offset++)
		{
			WATCHDOG_RESET();
			temp = start[offset];
			if (temp != pattern)
			{
			  printf ("\nFAILURE (read/write) @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx)\n", (ulong)&start[offset], pattern, temp);
			  errs++;
			}

			anti_pattern = ~pattern;
			start[offset] = anti_pattern;
		}

		/*
		 * Check each location for the inverted pattern and zero it.
		 */
		for (pattern = 1, offset = 0; offset < num_words; pattern++, offset++)
		{
			WATCHDOG_RESET();
			anti_pattern = ~pattern;
			temp = start[offset];
			if (temp != anti_pattern)
			{
			  printf ("\nFAILURE (read/write): @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx)\n",	(ulong)&start[offset], anti_pattern, temp);
			  errs++;
			}
			start[offset] = 0;
		}
	}

}

#ifdef DEVELOPERTESTS
/*
 * Initialize the GPMC interface on CS2 (DHCOM_CS_B) to access a external sram (AS7C34098A)
 * Be aware of the fact that we can address only 128 kByte
 */
int ext_sram_init(void)
{
	/* Disable Chip-Select O */
	*((uint*) 0x6E000078) 		= 0x00000800;

	/* initialize sram first - this config is tested with AS7C34098A - SRAM
	 * Limitations: You can address only 128 kByte
	 * We use CS 2 */
	*((uint*) 0x6E0000C0)		= 0x00001200;	// GPMC_CONFIG1_2
	*((uint*) 0x6E0000C4)		= 0x00090903;	// GPMC_CONFIG2_2
	*((uint*) 0x6E0000C8)		= 0x00080802;	// GPMC_CONFIG3_2
	*((uint*) 0x6E0000CC)		= 0x06040603/*0x06030603*/;	// GPMC_CONFIG4_2
	*((uint*) 0x6E0000D0)		= 0x01070707;	// GPMC_CONFIG5_2
	*((uint*) 0x6E0000D4)		= 0x87030200;	// GPMC_CONFIG6_2
	*((uint*) 0x6E0000D8)		= 0x00000F42;	// GPMC_CONFIG7_2

	// Set WriteProtect Pin to "active low mode"
	*((uint*) 0x6E000050) 	= 0x00000A00;

	/* now the sram memory is available at the address-space:
	 *   0x02000000 -- 0x0201FFFF
	 */
	printf("sram is available at:  0x02000000 -- 0x0201FFFF\n");

	return 0;
}

/*
 * Perform a memory test on external sram. The address-space is hardcoded!!
 * This routine is similar to the more complete alternative test which can be configure
 * with CONFIG_SYS_ALT_MEMTEST for the mtest command.
 */
int ext_sram_testdh(void)
{
	vu_long	*addr, *start, *end;
	ulong	val;
	ulong	readback;
	ulong	errs = 0;
	int iterations = 1;
	int iteration_limit;

	vu_long	len;
	vu_long	offset;
	vu_long	test_offset;
	vu_long	pattern;
	vu_long	temp;
	vu_long	anti_pattern;
	vu_long	num_words;
	vu_long dummy ;
	int	j;

// OLD Settings used with 428-300
	/* disable Chip-Select O - only one active CS at a time */
//	*((uint*) 0x6E000078) 		= 0x00000800;

	/* initialize sram first - this config is tested with AS7C34098A - SRAM
	 * Limitations: You can address only 128 kByte */
//	*((uint*) 0x6E0000C0)		= 0x00001200;	// GPMC_CONFIG1_2
//	*((uint*) 0x6E0000C4)		= 0x00090903;	// GPMC_CONFIG2_2
//	*((uint*) 0x6E0000C8)		= 0x00080802;	// GPMC_CONFIG3_2
//	*((uint*) 0x6E0000CC)		= 0x06030603;	// GPMC_CONFIG4_2
//	*((uint*) 0x6E0000D0)		= 0x01080808;	// GPMC_CONFIG5_2
//	*((uint*) 0x6E0000D4)		= 0x87030200;	// GPMC_CONFIG6_2
//	*((uint*) 0x6E0000D8)		= 0x00000F42;	// GPMC_CONFIG7_2

	// Set WriteProtect Pin to "active low mode"
//	*((uint*) 0x6E000050) 	= 0x00000A00;


	/* Disable Chip-Select O */
	*((uint*) 0x6E000078) 		= 0x00000800;

	/* initialize sram first - this config is tested with AS7C34098A - SRAM
	 * Limitations: You can address only 128 kByte
	 * We use CS 2 */
	*((uint*) 0x6E0000C0)		= 0x00001200;	// GPMC_CONFIG1_2
	*((uint*) 0x6E0000C4)		= 0x00090903;	// GPMC_CONFIG2_2
	*((uint*) 0x6E0000C8)		= 0x00080802;	// GPMC_CONFIG3_2
	*((uint*) 0x6E0000CC)		= 0x06040603/*0x06030603*/;	// GPMC_CONFIG4_2
	*((uint*) 0x6E0000D0)		= 0x01070707;	// GPMC_CONFIG5_2
	*((uint*) 0x6E0000D4)		= 0x87030200;	// GPMC_CONFIG6_2
	*((uint*) 0x6E0000D8)		= 0x00000F42;	// GPMC_CONFIG7_2

	// Set WriteProtect Pin to "active low mode"
	*((uint*) 0x6E000050) 	= 0x00000A00;


	static const ulong bitpattern[] = {
		0x00000001,	/* single bit */
		0x00000003,	/* two adjacent bits */
		0x00000007,	/* three adjacent bits */
		0x0000000F,	/* four adjacent bits */
		0x00000005,	/* two non-adjacent bits */
		0x00000015,	/* three non-adjacent bits */
		0x00000055,	/* four non-adjacent bits */
		0xaaaaaaaa,	/* alternating 1/0 */
	};

	// set start and stop address from board configuration -> A16 and A17 are not connected -> max 128 kByte
	start = (ulong *)(0x02000000);
	end = (ulong *)(0x0201FFFF);

	// choose test pattern
	pattern = 0;

	// configure hardcoded iteration limit
	iteration_limit = 100;

	printf ("Testing %08x ... %08x:\n", (uint)start, (uint)end);
	PRINTF("%s:%d: start 0x%p end 0x%p\n",	__FUNCTION__, __LINE__, start, end);

	for (;;) {

		// check if count of test-loops reached
		if (iteration_limit && iterations > iteration_limit)
		{
			printf("Tested %d iteration(s) with %lu errors.\n",	iterations-1, errs);
			return errs != 0;
		}

		printf("Iteration: %6d\r", iterations);
		PRINTF("\n");
		iterations++;

		/*
		 * Data line test:
		 * write a pattern to the first location, write the 1's complement to a 'parking'
		 * address (changes the state of the data bus so a floating bus doen't give a false OK), and then
		 * read the value back. Note that we read it back into a variable because the next time we read it,
		 * it might be right (been there, tough to explain to the quality guys why it prints a failure when the
		 * "is" and "should be" are obviously the same in the error message).
		 *
		 * Rather than exhaustively testing, we test some
		 * patterns by shifting '1' bits through a field of
		 * '0's and '0' bits through a field of '1's (i.e.
		 * pattern and ~pattern).
		 */
		addr = start;
		for (j = 0; j < sizeof(bitpattern)/sizeof(bitpattern[0]); j++)
		{
			val = bitpattern[j];
			for(; val != 0; val <<= 1)
			{
				*addr  = val;
				dummy  = ~val; /* clear the test data off of the bus */
				readback = *addr;
				if(readback != val)
				{
					printf ("FAILURE (data line): expected %08lx, actual %08lx\n", val, readback);
					errs++;
				}
				*addr  = ~val;
				dummy  = val;
				readback = *addr;
				if(readback != ~val)
				{
					printf ("FAILURE (data line): Is %08lx, should be %08lx\n",	readback, ~val);
					errs++;
				}
			}
		}

		/*
		 * Based on code whose Original Author and Copyright
		 * information follows: Copyright (c) 1998 by Michael
		 * Barr. This software is placed into the public
		 * domain and may be used for any purpose. However,
		 * this notice must not be changed or removed and no
		 * warranty is either expressed or implied by its
		 * publication or distribution.
		 */

		/*
		 * Address line test
		 *
		 * Description: Test the address bus wiring in a memory region by performing a walking
		 *              1's test on the relevant bits of the address and checking for aliasing.
		 *              This test will find single-bit address failures such as stuck -high,
		 *              stuck-low, and shorted pins. The base address and size of the region are
		 *              selected by the caller.
		 *
		 * Notes:	For best results, the selected base address should have enough LSB 0's to
		 *              guarantee single address bit changes. For example, to test a 64-Kbyte
		 *              region, select a base address on a 64-Kbyte boundary. Also, select the
		 *              region size as a power-of-two if at all possible.
		 */
		len = ((ulong)end - (ulong)start)/sizeof(vu_long);
		pattern = (vu_long) 0xaaaaaaaa;
		anti_pattern = (vu_long) 0x55555555;

		PRINTF("%s:%d: length = 0x%.8lx\n",	__FUNCTION__, __LINE__,	len);
		/*
		 * Write the default pattern at each of the
		 * power-of-two offsets.
		 */
		for (offset = 1; offset < len; offset <<= 1)
		{
			start[offset] = pattern;
		}

		/*
		 * Check for address bits stuck high.
		 */
		test_offset = 0;
		start[test_offset] = anti_pattern; // change state of lsb-address-line
										   // write anti-pattern if lsb-address is defect you would write to offset=1 and not test_offset=0

		for (offset = 1; offset < len; offset <<= 1)
		{
			temp = start[offset];
			if (temp != pattern)
			{
			  printf ("\nFAILURE: Address bit stuck high @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx\n", (ulong)&start[offset], pattern, temp);
			  errs++;
			}
		}
		start[test_offset] = pattern;
		WATCHDOG_RESET();

		/*
		 * Check for addr bits stuck low or shorted.
		 */
		for (test_offset = 1; test_offset < len; test_offset <<= 1)
		{
			start[test_offset] = anti_pattern;

			for (offset = 1; offset < len; offset <<= 1)
			{
			  temp = start[offset];
			  if ((temp != pattern) && (offset != test_offset))
			  {
				printf ("\nFAILURE: Address bit stuck low or shorted @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx\n", (ulong)&start[offset], pattern, temp);
				errs++;
			  }
			}
			start[test_offset] = pattern;
		}

		/*
		 * Description:
		 * Test the integrity of a physical memory device by performing an
		 * increment/decrement test over the entire region. In the process every
		 * storage bit in the device is tested as a zero and a one. The base address
		 * and the size of the region are selected by the caller.
		 */

		num_words = ((ulong)end - (ulong)start)/sizeof(vu_long) + 1;

		/*
		 * Fill memory with a known pattern.
		 */
		for (pattern = 1, offset = 0; offset < num_words; pattern++, offset++)
		{
			WATCHDOG_RESET();
			start[offset] = pattern;
		}

		/*
		 * Check each location and invert it for the second pass.
		 */
		for (pattern = 1, offset = 0; offset < num_words; pattern++, offset++)
		{
			WATCHDOG_RESET();
			temp = start[offset];
			if (temp != pattern)
			{
			  printf ("\nFAILURE (read/write) @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx)\n", (ulong)&start[offset], pattern, temp);
			  errs++;
			}

			anti_pattern = ~pattern;
			start[offset] = anti_pattern;
		}

		/*
		 * Check each location for the inverted pattern and zero it.
		 */
		for (pattern = 1, offset = 0; offset < num_words; pattern++, offset++)
		{
			WATCHDOG_RESET();
			anti_pattern = ~pattern;
			temp = start[offset];
			if (temp != anti_pattern)
			{
			  printf ("\nFAILURE (read/write): @ 0x%.8lx: expected 0x%.8lx, actual 0x%.8lx)\n",	(ulong)&start[offset], anti_pattern, temp);
			  errs++;
			}
			start[offset] = 0;
		}
	}

	/* Disable CS_B (GPMC_2) */
	*((uint*) 0x6E0000D8)		= 0x00000F02;	// GPMC_CONFIG7_2

	/* Enable Chip-Select O */
	*((uint*) 0x6E000078) 		= 0x00000840;

}
#endif

/*
 * Set rtc via i2c to required date.
 */
int i2c_set_rtc_testdh(int year, int month, int day, int hour, int minutes, int seconds)
{
	DISABLE_PRINTF() // disable printf while i2c operation

	// set control register
	i2c_reg_write(ADDR_RTCC_WRITE, ADDR_CTRL, 0x83);	// initialize CONTROL	register

	// set date
	i2c_reg_write(ADDR_RTCC_WRITE, ADDR_YEAR, (uchar)(0xFF & year));				// initialize YEAR	register
	i2c_reg_write(ADDR_RTCC_WRITE, ADDR_MNTH, (uchar)(0xFF & month));     			// initialize MONTH register
	i2c_reg_write(ADDR_RTCC_WRITE, ADDR_DATE, (uchar)(0xFF & day));     			// initialize DATE  register
	i2c_reg_write(ADDR_RTCC_WRITE, ADDR_HOUR, (uchar)(0xFF & hour));     			// initialize HOUR  register
	i2c_reg_write(ADDR_RTCC_WRITE, ADDR_MIN,  (uchar)(0xFF & minutes));     			// initialize MIN   register
	i2c_reg_write(ADDR_RTCC_WRITE, ADDR_SEC,  (uchar)(0xFF & seconds)|START_32KHZ);  	// init SEC register, set START bit

	// Check started flag of rtc
	if( !(0x80 & i2c_reg_read(ADDR_RTCC_READ, ADDR_SEC)))
	{
		SESSION_DEPENDED_PRINTF_ENABLE() // enable printf after i2c operation

		printf("Error: rtc could not be started!\n");
		return -1;
	}

	SESSION_DEPENDED_PRINTF_ENABLE() // enable printf after i2c operation

	return 0;
}

/*
 * Set rtc via i2c to required date.
 */
int i2c_read_rtc_testdh(void)
{
	uchar year, month, day, hour, minutes, seconds;

	DISABLE_PRINTF() // disable printf while i2c operation

	// set date
	year 	= 0xFF & i2c_reg_read(ADDR_RTCC_READ, ADDR_YEAR);   // read YEAR	register
	month 	= 0x1F & i2c_reg_read(ADDR_RTCC_READ, ADDR_MNTH);   // read MONTH register
	day 	= 0x3F & i2c_reg_read(ADDR_RTCC_READ, ADDR_DATE);   // read DATE  register
	hour 	= 0x3F & i2c_reg_read(ADDR_RTCC_READ, ADDR_HOUR);   // read HOUR  register
	minutes = 0x7F & i2c_reg_read(ADDR_RTCC_READ, ADDR_MIN);    // read MIN   register
	seconds = 0xFF & i2c_reg_read(ADDR_RTCC_READ, ADDR_SEC);  	// read SEC register

	SESSION_DEPENDED_PRINTF_ENABLE() // enable printf after i2c operation

	// Check started flag of rtc
	if( !(0x80 & seconds))
	{
		printf("\nWarning: You have to start the rtc first!\n");
	}

	seconds &= 0x7F; // unmask start-bit of rtc

	printf("\ndate read from rtc: 20%02x-%02x-%02x  time: %02x:%02x:%02x\n\n",year, month, day, hour, minutes, seconds);

	return 0;
}

/*
 * SD/microSD-card Interface test
 *
 * param:
 * 		int iInterfaceSelector (if 1 use onboard microSD, if 2 use ext SD-Card Interface)
 */
int SDcard_Interface_testdh(int iInterfaceSelector)
{
	vu_long	*start_ddr2, *end_ddr2;
	vu_long	len, offset;
	int err = 0;
	long size;
	block_dev_desc_t *dev_desc=NULL;
	int dev=0;
	int part=1;

	char *p_cCommand_string_mmcswitchOnboard  = "mmc switch 1";
	char *p_cCommand_string_mmcswitchExt  = "mmc switch 2";
	char *p_cCommand_string_mmc_if  = "mmc";
	char *p_cCommand_string_mmc_file  = "microSD_test.bmp";
	//char *p_cCommand_string_loadfile[5]  = {"fatload","mmc","0","81000000","microSD_test.bmp"};
	//char *p_cCommand_string_loadfile  = "fatload mmc 0 81000000 microSD_test.bmp";

	static const ulong bitpattern[] = {
			0x55555555,	/* alternating 0/1 */
			0xaaaaaaaa,	/* alternating 1/0 */
		};

	//--------------------------------------------------------------------------
	// set start and stop address from board configuration
	start_ddr2 = (ulong *)CONFIG_SYS_MEMTEST_START;
	end_ddr2 = (ulong *)(CONFIG_SYS_MEMTEST_END);

	//--------------------------------------------------------------------------
	// Switch to required SD-Card interfase
	if (iInterfaceSelector == 1)
		run_command(p_cCommand_string_mmcswitchOnboard,0);
	else
		run_command(p_cCommand_string_mmcswitchExt,0);

	//--------------------------------------------------------------------------
	// initialize microSD-card
	if (mmc_legacy_init(0) != 0)
	{
		printf("No MMC card found\n");
		return -1;
	}

	//--------------------------------------------------------------------------
	// load file "microSD_test.bmp" to ddr2

	/* Get the device environment via interface type and id-number */
	dev_desc = get_dev(p_cCommand_string_mmc_if,dev);
	if (dev_desc==NULL)
	{
		printf("\n** Invalid boot device **\n");
		return -1;
	}

	/* try to register the device in fat-system */
	if (fat_register_device(dev_desc,part)!=0)
	{
		printf("\n** Unable to use %s %d:%d for fatload **\n",p_cCommand_string_mmc_if,dev,part);
		return -1;
	}

	/* Read the file from fat device */
	size = file_fat_read (p_cCommand_string_mmc_file, (unsigned char *) start_ddr2, 0);
	if (size==-1) 		/* Error during read-process? */
	{
		printf("\n** Unable to read \"%s\" from %s %d:%d **\n",p_cCommand_string_mmc_file,p_cCommand_string_mmc_if,dev,part);
		return -1;
	}
	printf ("\n%ld bytes read\n", size);	/* reading x bytes was successful */


	//--------------------------------------------------------------------------
	//compare with testpattern

	len = size/4; // len count long and size counts bytes

	for (offset = 0; offset < len; offset = offset+2)
	{
		if(bitpattern[1] != start_ddr2[offset])
		{
			printf("\nError!");
			err++;
		}

		if(bitpattern[0] != start_ddr2[offset+1])
		{
			printf("\nError!");
			err++;
		}
	}

	return err;
}


/*
 * send ping request to test ethernet
 */
int ping_testdh(void)
{
	NetPingIP = string_to_ip(getenv("serverip"));

	if (NetLoop(PING) < 0)
	{
		printf("ping failed; host %s is not alive\n", getenv("serverip"));
		return 1;
	}

	printf("host %s is alive\n", getenv("serverip"));

	return 0;
}

#ifdef DEVELOPERTESTS
/*
 * this is to test if the time measurement works correct
 */
int dram_performance_testdh(unsigned long addr, unsigned long dest, unsigned long count)
{
	long timervalue, timervalue2 = 0;
	unsigned long bytes = count;
	unsigned long kb_sec = 0;

	timervalue = get_timer(0);

	count /= 4;

	/* put load on dram */
	while (count-- > 0)
	{
			*((ulong  *)dest) = *((ulong  *)addr);
			dest += 4;
			addr += 4;
	}

	timervalue = get_timer(0) - timervalue;

	printf("\nMeasured Time: %li ms\n", timervalue);

	kb_sec = (bytes * 1000) / (timervalue * 1024);

	printf("kByte/s = %li \n\n",kb_sec);

	/* --------------------------------------------------
	 * measure execution time */

	timervalue2 = get_timer(0);

	count = bytes;
	count /= 4;

	/* put load on dram */
	while (count-- > 0)
	{
		asm volatile("nop");
		dest += 4;
		addr += 4;
	}

	timervalue2 = get_timer(0) - timervalue2;

	printf("\nExecution Time: %li ms\n", timervalue2);

	kb_sec = (bytes * 1000) / ((timervalue - timervalue2) * 1024);

	printf("kByte/s = %li \n\n",kb_sec);

	return 0;
}

/*
 * this is to test dma
 */
int dram_dma_copy(unsigned long addr, unsigned long dest, unsigned long count)
{
	//long timervalue, timervalue2 = 0;
	//unsigned long bytes = count;
	//unsigned long kb_sec = 0;
	struct omap3_dma4 *omap3_dma4_base 	= (struct omap3_dma4*)OMAP3_SDMA_BASE_ADDR;

	/* setting up dma */
	omap3_dma4_base->channel[0].csdp 	= DATA_TYPE(DATATYPE_32BIT_SCALAR);
	omap3_dma4_base->channel[0].cen		= 5;
	omap3_dma4_base->channel[0].cfn		= 10;
	omap3_dma4_base->channel[0].cssa	= addr;
	omap3_dma4_base->channel[0].cdsa	= dest;
	omap3_dma4_base->channel[0].csei	= 1;
	omap3_dma4_base->channel[0].csfi	= 1;
	omap3_dma4_base->channel[0].cdei	= 1;
	omap3_dma4_base->channel[0].cdfi	= 1;
	omap3_dma4_base->channel[0].ccr		= DST_ADDR_MODE(OMAP_DMA_AMODE_POST_INC) | SRC_ADDR_MODE(OMAP_DMA_AMODE_POST_INC);

	/* start dma*/
	omap3_dma4_base->channel[0].ccr		|= CHA_ENABLE;

	return 0;
}


/*
 * this is to test if the time measurement works correct
 */
int timer_testdh(unsigned long timeus)
{
	long timervalue = 0;

	timervalue = get_timer(0);

	udelay(timeus);

	timervalue = get_timer(0) - timervalue;

	printf("\nMeasured Time: %li ms\n", timervalue);

	return 0;
}
#endif
/*
 * dh board test
 * loop with different hardware tests
 */
int board_testdh(void)
{
	ulong error_count = 0;
	ulong loop_count = 0;
	int temp_errors;

	for (;;) // endless test loop
	{
		loop_count++;

		//-------------------------------------------
		// read date from rtc - equals I2C test
		printf("\n\nread date and time from rtc ...\n");
		if (i2c_read_rtc_testdh())
		{
			error_count++;
		}
		udelay(1000000);

		//-------------------------------------------
		// run ddr2 test
		printf("\n\nexecute ddr2 test ...\n");
		if (ddr2_testdh() != 0)
		{
			error_count++;
		}

		//-------------------------------------------
		// run nand-flash test

		printf("\n\nexecute nand-flash test ...\n");
		temp_errors = nand_testdh();

		if (temp_errors < 0)
		{
			printf("\nFatal error during nand-flash test!\n");
			error_count++;
			//return -1; // don't exit loop even after errors
		}
		printf("nand-flash test completed with %i errors! \n\n", temp_errors);
		error_count += temp_errors;

		//-------------------------------------------
		// execute ping to serverip
		printf("\n\nexecute ping to serverip ...\n");
		if (ping_testdh() != 0)
		{
			error_count++;
		}

		//-------------------------------------------
		// test microSD-card
		printf("\n\nexecute microSD-card test ...\n");
		temp_errors = nand_testdh();

		if (temp_errors < 0)
		{
			printf("\nFatal error during microSD-card test!\n");
			error_count++;
			//return -1; // don't exit loop even after errors
		}
		printf("microSD-card test completed with %i errors! \n\n", temp_errors);
		error_count += temp_errors;

		//-------------------------------------------
		// display loop iterations and error count
		printf("DH BOARD-TEST loop %ld finished with %ld absolute errors! \n\n", loop_count, error_count);

	}

	return 0; // never reached!!
}

int do_dhtestroutines(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{

	int year, month, day, hour, minutes, seconds;
	int gpio, state;
	int errors;
	char *cmd;


	/* at least two arguments please */
	if (argc < 2)
		goto usage;

	cmd = argv[1];

	/* dh board test */
	if (strcmp(cmd, "board") == 0)
	{
		if (argc != 2)
			goto usage;
		printf("start dh board test ...\n\n");
		return board_testdh();
	}

#ifdef DEVELOPERTESTS
	unsigned long timeus, sourceaddr, destaddr, count;

	/* dh timer test */
	if (strcmp(cmd, "timer") == 0)
	{
		if (argc != 3)
		{
			printf("\n\n\ttestdh timer usecvalue\n\n");
			goto usage;
		}

		timeus = (int)simple_strtoul(argv[2], NULL, 10);
		printf("timer test ...\n\n");
		return timer_testdh(timeus);
	}

	/* dram test */
	if (strcmp(cmd, "dramper") == 0)
	{
		if (argc != 5)
		{
			printf("\n\n\ttestdh dramper sourceaddr destaddr bytecount\n\n");
			goto usage;
		}


		printf("dram performance test ...\n\n");

		sourceaddr = simple_strtoul(argv[2], NULL, 16);
		destaddr = simple_strtoul(argv[3], NULL, 16);
		count = simple_strtoul(argv[4], NULL, 16);

	if (count == 0) {
		printf ("Zero length ???\n");
		return 1;
	}
	 return dram_performance_testdh(sourceaddr, destaddr, count);
	}

	/* Test a DMA channel */
	if (strcmp(cmd, "dma") == 0)
	{
		if (argc != 5)
		{
			printf("\n\n\ttestdh dma sourceaddr destaddr bytecount\n\n");
			goto usage;
		}

		printf("dma test ...\n\n");

		sourceaddr = simple_strtoul(argv[2], NULL, 16);
		destaddr = simple_strtoul(argv[3], NULL, 16);
		count = simple_strtoul(argv[4], NULL, 16);

	if (count == 0) {
		printf ("Zero length ???\n");
		return 1;
	}
	 return dram_dma_copy(sourceaddr, destaddr, count);
	}
#endif

	/* run nand-flash test */
	if (strcmp(cmd, "nand") == 0)
	{
		printf("starting nand-flash test-routine ...\n\n");

		errors = nand_testdh();

		if (errors < 0)
		{
			printf("\nFatal Error during nand-flash test!\n");
			return -1;
		}
		printf("nand-flash test completed with %i errors \n\n", errors);
		return 0;
	}

	/* run ddr2 test */
	if (strcmp(cmd, "ddr2") == 0)
	{
		printf("starting ddr2 test-routine ...\n\n");
		return ddr2_testdh();
	}

#ifdef DEVELOPERTESTS
	/* run ddr2 test */
	if (strcmp(cmd, "sram") == 0)
	{
		printf("starting sram test-routine ...\n\n");
		ext_sram_testdh();

		/* Disable CS_B (GPMC_2) */
		*((uint*) 0x6E0000D8)	= 0x00000F02;	// GPMC_CONFIG7_2

		/* Enable Chip-Select O */
		*((uint*) 0x6E000078) 	= 0x00000840;

		// Set WriteProtect Pin to "active low mode"
		*((uint*) 0x6E000050) 	= 0x00000A10;

		return 0;
	}


	/* run ddr2 test */
	if (strcmp(cmd, "initsram") == 0)
	{
		printf("init ext sram ...\n\n");
		return ext_sram_init();
	}
#endif
	/* configure rtc and start rtc */
	if (strcmp(cmd, "setrtc") == 0)
	{
		if (argc != 8)
			goto usage;
		printf("set rtc to specified date and start ...\n\n");

		year = (int)simple_strtoul(argv[2], NULL, 16);
		month = (int)simple_strtoul(argv[3], NULL, 16);
		day = (int)simple_strtoul(argv[4], NULL, 16);
		hour = (int)simple_strtoul(argv[5], NULL, 16);
		minutes = (int)simple_strtoul(argv[6], NULL, 16);
		seconds = (int)simple_strtoul(argv[7], NULL, 16);

		return i2c_set_rtc_testdh(year, month, day, hour, minutes, seconds);
	}

	/* read date from rtc */
	if (strcmp(cmd, "readrtc") == 0)
	{
		if (argc != 2)
			goto usage;
		printf("read date from rtc ...\n\n");
		return i2c_read_rtc_testdh();
	}

	/* microSD test */
	if (strcmp(cmd, "microSD") == 0)
	{
		if (argc != 2)
			goto usage;
		printf("test microSD-card ...\n\n");

		errors = SDcard_Interface_testdh(1); /* execute microSD test */
		if (errors < 0)
		{
			printf("\nFatal Error during microSD test!\n");
			return -1;
		}
		printf("\n\nmicroSD test completed with %i errors!\n\n", errors);
		return 0;
	}

	/* external SD-Card interface test */
	if (strcmp(cmd, "extSD") == 0)
	{
		if (argc != 2)
			goto usage;
		printf("test external SD-card ...\n\n");

		errors = SDcard_Interface_testdh(2); /* execute microSD test */
		if (errors < 0)
		{
			printf("\nFatal Error during ext SD-interface test!\n");
			return -1;
		}
		printf("\n\nExternal SD-interface test completed with %i errors!\n\n", errors);
		return 0;
	}

	/* set a gpio to output and specified state */
	if (strcmp(cmd, "setgpio") == 0)
	{
		if (argc != 4)
			goto usage;

		gpio = (int)simple_strtoul(argv[2], NULL, 10);
		state = (int)simple_strtoul(argv[3], NULL, 2);

		return set_gpio_testdh(gpio,state);
	}

	/* set a gpio to output and specified state */
	if (strcmp(cmd, "getgpio") == 0)
	{
		if (argc != 3)
			goto usage;

		gpio = (int)simple_strtoul(argv[2], NULL, 10);

		return get_gpio_testdh(gpio);
	}

	/* set a pad to gpio mode */
	if (strcmp(cmd, "congpio") == 0)
	{
		if (argc != 3)
			goto usage;

		gpio = (int)simple_strtoul(argv[2], NULL, 10);

		return configure_gpio_testdh(gpio);
	}

usage:
	cmd_usage(cmdtp);
	return 1;
}

/* ====================================================================== */

U_BOOT_CMD(
  testdh,     CONFIG_SYS_MAXARGS, 0, do_dhtestroutines,
  "DH Testroutines",
  "-> DH Test Subsystem\n"
  "---------------------------\n\n"
  "testdh board         - complete board (endless loop)\n"
  "testdh nand          - run nandflash test\n"
  "testdh ddr2          - run ddr2 test\n"
  "testdh setrtc yy mm dd hour min sec\n"
  "testdh readrtc       - read date from rtc\n"
  "testdh microSD       - microSD reading \"microSD_test.bmp\"\n"
  "testdh extSD         - SD-Card reading \"microSD_test.bmp\"\n"
  "testdh setgpio # x   - set a gpio # to output an state x\n"
  "testdh getgpio #     - set a gpio # to input and read in\n"
  "testdh congpio #     - configure a pad as gpio # (UNSAFE)\n");

#endif/* CONFIG_CMD_TESTROUTINES */
