/*
 * 2008 (c) STMicroelectronics, Inc.
 * Author: Ryan Chen <Ryan.Chen at st.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Opsycon AB, Sweden.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <common.h>
#include <command.h>
#include <linux/ctype.h>
#include <net.h>
#include <dh_settings.h>
#include <wince_args.h>

DECLARE_GLOBAL_DATA_PTR;

/*
 * Windows CE Binary Image Data Format
 * The binary image (.bin) file format organizes data by sections. Each section contains 
 * a section header that specifies the starting address, length, and checksum for that 
 * section. 
 *
 * Refer to <http://msdn.microsoft.com/en-us/library/ms924510.aspx>
 *
 * The following table shows the .bin file format.
 * Field        Length (bytes)  Description 
 * Sync bytes (optional)  7     Byte 0 is B, indicating a .bin file format. 
 *                Bytes 1-6 are reserved and set to 0, 0, 0, F, F, \n.
 * Run-time image address 4     Physical starting address of the run-time image. 
 * Run-time image length  4     Physical length, in bytes, of the run-time image. 
 * Record Address     4     Physical starting address of data record. 
 *                If this value is zero, the Record Address is the end of 
 *                the file, and record length contains the starting address 
 *                of the run-time image.
 * Record length      4     Length of record data, in bytes. 
 * Record checksum    4     Signed 32-bit sum of record data bytes. 
 * Record data      Record length Record data. 
 */
 
 BSP_ARGS *g_pBSPArgs;
 
#define WINCE_IMAGE_SYNC_SIZE 7
#define WINCE_IMAGE_NK_BIN_SYNC     "B000FF\n"
#define WINCE_IMAGE_NK_GZ_SYNC      "nk.nb0"

#define DEFAULT_MAC_ADDRESS	{0x2020,0x3040,0x5060}
u16 EthMacAddress[] = DEFAULT_MAC_ADDRESS;

typedef struct {
    unsigned char sync_bytes[WINCE_IMAGE_SYNC_SIZE];
    unsigned int img_addr;
    unsigned int img_length;
} type_wince_nk_bin_image_header;

typedef struct {
    unsigned char reserved[10];
    unsigned char sync_bytes[6];
} type_wince_nk_gz_image_header;

typedef struct {
    u32 last_record_addr;
    u32 last_record_length;
    u32 record_length;
} wince_record_struct;

typedef enum {
    NO_VALID_NK_FILE = 0,
    NK_BIN_FILE = 1,
    NK_GZ_FILE = 2,
} NK_TYPE_INFO;

#ifdef DH_NAND_FLASH    
extern int do_nand(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[]);
#endif /* DH_NAND_FLASH */   
extern int do_mmu(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int do_mem_mw (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int do_unzip ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
#if defined(CONFIG_MX25)
extern int do_mmcswitch ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
#else
extern int do_mmc ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);
extern int dh_get_sdram_size_mb(void);
#endif

int check_sum(unsigned char * buf, int len, unsigned int checksum)
{
    unsigned int count,i;

    for (i = 0,count = 0 ; i < len ; i++)
        count += buf[i];

    if (count == checksum)
        return 0;

    return 1;
}

/* ======================================================================
 * Determine if a valid WinCE image exists at the given memory location.
 * First looks at the image header field, the makes sure that it is
 * WinCE image.
 * ====================================================================== */
int valid_wince_image (unsigned long addr)
{
    type_wince_nk_bin_image_header *p_nk_bin = (type_wince_nk_bin_image_header *)addr;
    type_wince_nk_gz_image_header *p_nk_gz = (type_wince_nk_gz_image_header *)addr;

    if(strcmp((char *)p_nk_bin->sync_bytes, (char *)WINCE_IMAGE_NK_BIN_SYNC) == 0)
        return 1;
    
    if(strcmp((char *)p_nk_gz->sync_bytes, (char *)WINCE_IMAGE_NK_GZ_SYNC) == 0)  
        return 2;

    return 0;
}

/* ======================================================================
 * A very simple WinCE image loader, assumes the image is valid, returns the
 * entry point address.
 * ====================================================================== */
wince_record_struct load_wince_image (unsigned long addr)
{
    wince_record_struct return_struct = {0,0,0};
    unsigned char *p = (unsigned char *)addr;
    u32 start_addr, total_length;
    u32 record_addr, record_length, record_checksum;
    u32 i = 0;
    u32 j = 0;

    if(valid_wince_image(addr) == 0)
        return return_struct;

    printf("WINCE image is found: ");
    p += WINCE_IMAGE_SYNC_SIZE;
    start_addr = (u32)(p[3]<<24) + (u32)(p[2]<<16) + (u32)(p[1]<<8) + (u32)p[0];
    p += 4;
    total_length = (u32)(p[3]<<24) + (u32)(p[2]<<16) + (u32)(p[1]<<8) + (u32)p[0];
    printf("\n--> BootWinCE: Start Address = 0x%x @ Total Length = 0x%x", start_addr, total_length);
    p += 4;
    printf("\n--> BootWinCE: Loading: ");

    /* read each records */
    while(1) 
    {
        record_length = (u32)(p[7]<<24) + (u32)(p[6]<<16) + (u32)(p[5]<<8) + (u32)p[4];
        record_checksum = (u32)(p[11]<<24) + (u32)(p[10]<<16) + (u32)(p[9]<<8) + (u32)p[8];
        record_addr = (u32)(p[3]<<24) + (u32)(p[2]<<16) + (u32)(p[1]<<8) + (u32)p[0];
        if(record_addr == 0)
            break;  

        if(check_sum((unsigned char *)&p[12], record_length, record_checksum) != 0) 
        {
            printf("\n--> BootWinCE ERROR: Checksum Error!");
            return return_struct;
        }
        memcpy ((void *)record_addr, (const void *)&p[12], (unsigned long)record_length);
        /* printf("Region %d: Loading from 0x%x to 0x%x @ Length 0x%x\n", i, (unsigned int)&p[12], \
          (unsigned int)record_addr, record_length);*/
        if ( j < 50)
        { 
            printf(".");
            j++;
        }
        else
        {
            j = 0;
            printf("\n                        ");
        }
        p = p + 12 + record_length;
        i++;
        return_struct.last_record_addr = record_addr;
        return_struct.last_record_length = record_length;
    }

    /* the lastest checksun should be zero */
    if(record_checksum != 0) 
    {
        printf("\n--> BootWinCE ERROR: Checksum Error!");
        return return_struct;
    }

    if( j != 0)
    {
        printf("\n");
    }

    return_struct.record_length = record_length;
    /* the lastest length is entry address */
    return return_struct;
}

/* ======================================================================
 * Interpreter command to boot WinCE from a memory image.  The image can
 * be an WinCE image.  WinCE image do not need the
 * bootline and other parameters.
 * ====================================================================== */
int do_bootwince (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
    char cNKGZImageSDRAMAddress[9]          = {BOOTCE_NK_GZ_IMAGE_SDRAM_ADDRESS};
  
    unsigned long ulNKGZImageSDRAMAddress   = simple_strtoul (cNKGZImageSDRAMAddress, NULL, 16);
    unsigned long ulSDRAMNKGZBufferAddress  = BOOTCE_NK_GZ_SDRAM_BUFFER_ADDRESS;
    unsigned long ulSDRAMNKBINBufferAddress = BOOTCE_NK_BIN_SDRAM_BUFFER_ADDRESS;    
    
    char *p_cReadNANDFlashContent[5]        = {"nand", "read", "12345678", "12345672", "12345673"};
#if defined(CONFIG_MX25)
    char *p_cSwitchToMicroSDSlot[5]         = {"mmcswitch", "2"};  
#else
    char *p_cSwitchToMicroSDSlot[3]         = {"mmc", "switch", "1"};
#endif
    char *mmu_on[2]                         = {"mmu","on"};
    char *mmu_off[2]                        = {"mmu","off"};
    char *unzip_nk_gz_image[3]              = {"unzip"," ",cNKGZImageSDRAMAddress};
    char *mw_ce_memory_config[4]            = {"mw.b","12345678","00","100000"};  
    char *sWinCEImageType;
    char *p_ENVOSImageFlashAddress          = getenv ("nandflashOSaddr");
    
    char *p_cOalArgsSignature				= { OAL_ARGS_SIGNATURE };

    unsigned long ulFilesize                = 0;
    unsigned long addr                      = 0;
    unsigned long help_addr                 = 0;   
    int ret_value                           = 0; 
    
    wince_record_struct return_struct       = {0,0,0};  
    NK_TYPE_INFO eWinCEImageType            = NO_VALID_NK_FILE;
    uchar enetaddr[6];

    /*
     * Check the loadaddr variable.
     * If we don't know where the image is then we're done.
     */
    if (argc < 2)
    {
        addr = load_addr;
    }
    else
    {
#if defined(CONFIG_CMD_NET)
        /* Check to see if we need to tftp the image ourselves before starting */
        if ((argc == 2) && (strcmp (argv[1], "tftp") == 0)) 
        {
            if (NetLoop (TFTP) <= 0)
                return 1;
            addr = load_addr;    
            printf ("\n--> BootWinCE: Automatic boot of WinCE image at address 0x%08lx ... ", addr);
        }
        else
#endif    
        {
            addr = simple_strtoul (argv[1], NULL, 16);
        }
    }

    if((gd->dh_ulFlags & BOOT_AFTER_UPDATE_FLAG) == 0)
    {
        printf ("\n--> BootWinCE INFO: Don't start WinEC 7 image after update\n");  
        return 1;    
    }

    printf ("\n--> BootWinCE: Try to load WinEC 7 Image"); 

    // Try to load Image from NAND Flash
    if(addr == simple_strtoul (p_ENVOSImageFlashAddress, NULL, 16))
    {
        printf ("\n--> BootWinCE: WinEC 7 Image is in NAND Flash memory");   
        ulFilesize = simple_strtoul (getenv("winceimagesize"), NULL, 16);
        sWinCEImageType = getenv("winceimagetype");
        
        // WinCE NAND Image is valid
        if((ulFilesize != 0) && (strcmp (sWinCEImageType, "no_wince") != 0))
        {
            printf ("\n--> BootWinCE: Copy WinEC 7 Image from NAND Flash to SDRAM Buffer");  
            sprintf (p_cReadNANDFlashContent[3], "%08x", (unsigned int)addr); 
            sprintf (p_cReadNANDFlashContent[4], "%08x", (unsigned int)ulFilesize); 
            
            // Load nk.gz file
            if(strcmp (sWinCEImageType, "nk.gz") == 0)
            {
                sprintf (p_cReadNANDFlashContent[2], "%08x", (unsigned int)ulSDRAMNKGZBufferAddress); 
                // Set address to SDRAM Buffer address
                addr = ulSDRAMNKGZBufferAddress;
                // Check File size
                if(ulFilesize > BOOTCE_NK_GZ_SDRAM_BUFFER_SIZE)
                {
                    printf ("\n--> BootWinCE ERROR: nk.gz Image size (0x%08x bytes) is to large for SDRAM Buffer\n", (unsigned int)ulFilesize);  
                    return 1;
                }
            }
            // Load nk.bin file
            else
            {
                sprintf (p_cReadNANDFlashContent[2], "%08x", (unsigned int)ulSDRAMNKBINBufferAddress); 
                // Set address to SDRAM Buffer address
                addr = ulSDRAMNKBINBufferAddress;
                // Check File size
                if(ulFilesize > BOOTCE_NK_BIN_SDRAM_BUFFER_SIZE)
                {
                    printf ("\n--> BootWinCE ERROR: nk.bin Image size (0x%08x bytes) is to large for SDRAM Buffer\n", (unsigned int)ulFilesize);  
                    return 1;
                }
            }
            
            do_mmu(NULL, 0, 2, mmu_on);
            DISABLE_PRINTF()
#ifdef DH_NAND_FLASH    
            ret_value = do_nand(NULL, 0, 5, p_cReadNANDFlashContent);
#endif /* DH_NAND_FLASH */  
            do_mmu(NULL, 0, 2, mmu_off);

            SESSION_DEPENDED_PRINTF_ENABLE()
        }
        // WinCE NAND FLash Image not valid
        else
        {
            printf ("\n--> BootWinCE INFO: No valid WinEC 7 Image found on NAND Flash\n");  
            SET_DH_GD_FLAG(NO_OSIMAGE_FOUND_FLAG);
            return 1;            
        }
    }
    // Image not in NAND Flash and SDRAM
    else if((addr < PHYS_SDRAM_1) || (addr > (PHYS_SDRAM_1 + PHYS_SDRAM_1_SIZE)))
    {
        printf ("\n--> BootWinCE ERROR: Wrong WinEC 7 Image address\n");      
        return 1;
    }
    
    // Image is in SDRAM
    else
    {
        printf ("\n--> BootWinCE: WinEC 7 Image is in SDRAM");      
    }
    
    
    /*
     * If the data at the load address is an WinCE image, then
     * treat it like an WinCE image. Otherwise, return 1
     */
    eWinCEImageType = valid_wince_image (addr);
   
    switch (eWinCEImageType)
    {
        case NK_BIN_FILE:		// Load nk.bin File
            printf ("\n==> BootWinCE: Start to load WinEC 7 nk.bin image"); 
            DISABLE_PRINTF()
            do_mmu(NULL, 0, 2, mmu_on);
            SESSION_DEPENDED_PRINTF_ENABLE()
            return_struct = load_wince_image (addr);
            addr = return_struct.record_length;
        break;

        case NK_GZ_FILE:		// Load nk.gz File
            printf ("\n==> BootWinCE: Start to load WinEC 7 nk.gz image"); 
            sprintf (unzip_nk_gz_image[1], "%08lx",addr);
            DISABLE_PRINTF()
            do_mmu(NULL, 0, 2, mmu_on);
            ret_value = do_unzip (NULL, 0, 3, unzip_nk_gz_image);	
            SESSION_DEPENDED_PRINTF_ENABLE()
            if(ret_value != 0)            
            {
                printf ("\n--> BootWinCE ERROR: Can't unzip WinEC 7 Image\n");    
                return 1;
            }
        break;
    
        default:  // Not an WinCE Image found
            puts ("\n--> BootWinCE: ## Not an WinCE image, exit!\n");
            SET_DH_GD_FLAG(NO_OSIMAGE_FOUND_FLAG);
            return 1;
            /* leave addr as load_addr */
		break;
    }

    // Copy Arguments Block from NAND Flash to SDRAM
    /*printf ("\n--> BootWinCE: Copy Arguments Block from Flash (0xA0180000) to SDRAM (0x800FF000)");
    
    sprintf (p_cReadNANDFlashContent[2], "%08x", (unsigned int)BOOTCE_ARGUMENTS_SDRAM_ADDRESS);     
    sprintf (p_cReadNANDFlashContent[3], "%08x", (unsigned int)BOOTCE_ARGUMENTS_FLASH_ADDRESS); 
    sprintf (p_cReadNANDFlashContent[4], "%08x", (unsigned int)(ulNKGZImageSDRAMAddress - BOOTCE_ARGUMENTS_SDRAM_ADDRESS)); 
    DISABLE_PRINTF()
    ret_value = do_nand(NULL, 0, 5, p_cReadNANDFlashContent);*/
    SESSION_DEPENDED_PRINTF_ENABLE()
    
    
    
    // Initialize the BSP args structure.
    //
    g_pBSPArgs = (BSP_ARGS *) BOOTCE_ARGUMENTS_SDRAM_ADDRESS;

    memset((void *)g_pBSPArgs, 0, sizeof(BSP_ARGS));
    g_pBSPArgs->header.signature  = *((unsigned long *)p_cOalArgsSignature);
    g_pBSPArgs->header.oalVersion = OAL_ARGS_VERSION;
    g_pBSPArgs->header.bspVersion = BSP_ARGS_VERSION;
    g_pBSPArgs->kitl.flags             = OAL_KITL_FLAGS_VMINI;
    g_pBSPArgs->kitl.devLoc.IfcType    = Internal;
    g_pBSPArgs->kitl.devLoc.BusNumber  = 0;
    g_pBSPArgs->kitl.ipAddress = 0;
    g_pBSPArgs->kitl.ipMask = 0;
    eth_getenv_enetaddr("ethaddr", enetaddr);
    EthMacAddress[2] = (enetaddr[5] << 8) | enetaddr[4];
    EthMacAddress[1] = (enetaddr[3] << 8) | enetaddr[2];
    EthMacAddress[0] = (enetaddr[1] << 8) | enetaddr[0];
    memcpy(g_pBSPArgs->mac,EthMacAddress,sizeof(g_pBSPArgs->mac)); 
    g_pBSPArgs->coldBoot = 1;
    g_pBSPArgs->updateMode = 0;
    g_pBSPArgs->deviceID = 0;
    g_pBSPArgs->oalFlags = 0;
    g_pBSPArgs->dispRes = OMAP_LCD_DEFAULT;
    g_pBSPArgs->ECCtype = 1; // 0 = 1bit; 1 = 4bit; 2 = 8bit
    g_pBSPArgs->opp_mode = 4;   //opp_mode == 4
                                // MPU[600Mhz @ 1.3500V]
                                //opp_mode == 3
                                // MPU[550Mhz @ 1.2750V]
                                //opp_mode == 2
                                // MPU[500Mhz @ 1.2000V]
                                //opp_mode == 1
                                // MPU[250Mhz @ 1.000V]
                                //opp_mode == 0
                                // MPU[125Mhz @ 0.975V] 

    g_pBSPArgs->SDRAMSize = dh_get_sdram_size_mb() * 1024 * 1024;  // DDR2 RAM size in Bytes

    g_pBSPArgs->dh_board_settings.cLength = gd->dh_board_settings.cLength;                  // Settings block length
    g_pBSPArgs->dh_board_settings.cDisplayID = gd->dh_board_settings.cDisplayID;            // Display unique ID 
    g_pBSPArgs->dh_board_settings.wValidationID = gd->dh_board_settings.wValidationID;      // Settings Block validation ID

    g_pBSPArgs->dh_board_settings.wYResolution = gd->dh_board_settings.wYResolution;        // Display y resolution
    g_pBSPArgs->dh_board_settings.wXResolution = gd->dh_board_settings.wXResolution;        // Display x resolution

    g_pBSPArgs->dh_board_settings.wLCDConfigFlags = gd->dh_board_settings.wLCDConfigFlags;  // Display configuration flags
    g_pBSPArgs->dh_board_settings.wPixelClock = gd->dh_board_settings.wPixelClock;          // Display pixel clock

    g_pBSPArgs->dh_board_settings.wVPulseWidth = gd->dh_board_settings.wVPulseWidth;        // VSYNC pulse width
    g_pBSPArgs->dh_board_settings.wHPulseWidth = gd->dh_board_settings.wHPulseWidth;        // HSYNC pulse width

    g_pBSPArgs->dh_board_settings.wHBackPorch = gd->dh_board_settings.wHBackPorch;          // HSYNC back porch
    g_pBSPArgs->dh_board_settings.wHFrontPorch = gd->dh_board_settings.wHFrontPorch;        // HSYNC front porch

    g_pBSPArgs->dh_board_settings.wVBackPorch = gd->dh_board_settings.wVBackPorch;          // VSYNC back porch
    g_pBSPArgs->dh_board_settings.wVFrontPorch = gd->dh_board_settings.wVFrontPorch;        // VSYNC front porch   

    g_pBSPArgs->dh_board_settings.cACBiasTrans = gd->dh_board_settings.cACBiasTrans;        // AC Bias transitions per interrupt
    g_pBSPArgs->dh_board_settings.cACBiasFreq = gd->dh_board_settings.cACBiasFreq;          // AC Bias frequency
    g_pBSPArgs->dh_board_settings.cDatalines = gd->dh_board_settings.cDatalines;            // Number of display datalines

    g_pBSPArgs->dh_board_settings.wGPIODir = gd->dh_board_settings.wGPIODir;                // DHCOM GPIO default direction
    g_pBSPArgs->dh_board_settings.wGPIOState = gd->dh_board_settings.wGPIOState;            // DHCOM GPIO default state

    g_pBSPArgs->dh_board_settings.wHWConfigFlags = gd->dh_board_settings.wHWConfigFlags;    // Hardware configuration flags      
    
    /* delete old Windows Embedded memory configuration in SDRAM */ 
    if(eWinCEImageType == NK_BIN_FILE)
    {
        return_struct.last_record_length = return_struct.last_record_addr + return_struct.last_record_length;
  
        sprintf (mw_ce_memory_config[1],"%08lx",(unsigned long)return_struct.last_record_length);
        printf ("\n--> BootWinCE: Delete existing WinEC 7 memory configuration");
        do_mem_mw (NULL, 0, 4, mw_ce_memory_config);
        printf ("\n--> BootWinCE: Delete 0x100000 Bytes from address 0x%08lx --> done", (unsigned long)return_struct.last_record_length);
    }

    // Switch to MicroSD Slot
    DISABLE_PRINTF()
    do_mmc(NULL, 0, 3, p_cSwitchToMicroSDSlot);
    SESSION_DEPENDED_PRINTF_ENABLE()
    printf ("\n--> BootWinCE: MicroSD Slot selected");
    
    // If no debugging is needed --> Start Windows Embdedded CE
    if(!((argc == 3) && (strcmp (argv[2],"debug") == 0)))
    {
        if(eWinCEImageType == NK_GZ_FILE)
        {
            addr = ulNKGZImageSDRAMAddress;
        }
  	
        printf ("\n--> BootWinCE: ## Starting WinEC 7 at 0x%08lx ...\n\n", addr);
        do_mmu(NULL, 0, 2, mmu_off);
   
        ((void (*)(void)) addr) ();
    }
  
    // Else --> Debug Windows Embdedded CE Image
    /* Don't start the Windows Embdedded CE Image */
    help_addr = addr - 0x1000;
    if(eWinCEImageType == NK_GZ_FILE)
    {
        help_addr = ulNKGZImageSDRAMAddress;
    }
    printf ("\n--> BootWinCE: ## WinEC 7 copied to address 0x%08lx ...\n\n", help_addr);
    do_mmu(NULL, 0, 2, mmu_off);

    return 0;
}

/* ====================================================================== */

U_BOOT_CMD(
  bootwince,      3,      0,      do_bootwince,
  "Boot Windows CE from a binary (.bin) image",
  "[address] - load address of Windows CE BIN image.\n"
  "          tftp - load first image via TFTP.\n"
  "          [address] debug - only load Windows CE, don't start.\n"
);

/* end of file */

