/*
 * Common LCD routines for supported CPUs
 *
 * (C) Copyright 2001-2002
 * Wolfgang Denk, DENX Software Engineering -- wd@denx.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/************************************************************************/
/* ** HEADER FILES							*/
/************************************************************************/

/* #define DEBUG */

#include <config.h>
#include <common.h>
#include <command.h>
#include <stdarg.h>
#include <linux/types.h>
#include <stdio_dev.h>
#if defined(CONFIG_POST)
#include <post.h>
#endif
#include <lcd.h>
#include <watchdog.h>
/* DH added includes */
#include <nand.h>
#include <dh_settings.h>
#include <bmp_layout.h>		/* Get logo data, width and height	*/

#if defined(CONFIG_PXA250)
#include <asm/byteorder.h>
#endif

#if defined(CONFIG_MPC823)
#include <lcdvideo.h>
#endif

#if defined(CONFIG_SPLASH_IS_IN_MMC)
#include <mmc.h>
#endif

#if defined(CONFIG_ATMEL_LCD)
#include <atmel_lcdc.h>
#endif

#ifdef CONFIG_OMAP
#include <asm/arch/am35xfb.h>
#endif

extern int do_nand(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[]);

/************************************************************************/
/* ** FONT DATA								*/
/************************************************************************/
#include <video_font.h>		/* Get font data, width and height	*/

/************************************************************************/
/* ** LOGO DATA								*/
/************************************************************************/
#ifdef CONFIG_LCD_LOGO
# include <bmp_layout.h>		/* Get logo data, width and height	*/
# if (CONSOLE_COLOR_WHITE >= BMP_LOGO_OFFSET)
#  error Default Color Map overlaps with Logo Color Map
# endif
#endif

DECLARE_GLOBAL_DATA_PTR;

#ifdef CONFIG_LCD
char lcd_cmap[256];
#endif

ulong lcd_setmem (ulong addr);

static void lcd_drawchars (ushort x, ushort y, uchar *str, int count);
static inline void lcd_puts_xy (ushort x, ushort y, uchar *s);
static inline void lcd_putc_xy (ushort x, ushort y, uchar  c);

static int lcd_init (void *lcdbase);

static int lcd_clear (cmd_tbl_t * cmdtp, int flag, int argc, char *argv[]);
static void *lcd_logo (void);

static int lcd_getbgcolor (void);
static void lcd_setfgcolor (int color);
static void lcd_setbgcolor (int color);

char lcd_is_enabled = 0;

#ifdef CONFIG_AM35XX_LCD
extern void lcd_reinit_to_new_bmpformat(uchar *bmp_header);
#endif

#ifdef	NOT_USED_SO_FAR
static void lcd_getcolreg (ushort regno, ushort *red, ushort *green, ushort *blue);
static int lcd_getfgcolor (void);
#endif	/* NOT_USED_SO_FAR */


/************************************************************************/

/*----------------------------------------------------------------------*/

static void console_scrollup (void)
{
	/* Copy up rows ignoring the first one */
	memcpy (CONSOLE_ROW_FIRST, CONSOLE_ROW_SECOND, CONSOLE_SCROLL_SIZE);

	/* Clear the last one */
	memset (CONSOLE_ROW_LAST, COLOR_MASK(lcd_color_bg), CONSOLE_ROW_SIZE);
}

/*----------------------------------------------------------------------*/

static inline void console_newline (void)
{
	++console_row;
	console_col = 0;

	/* Check if we need to scroll the terminal */
	if (console_row >= CONSOLE_ROWS)
	{
		/* Scroll everything up */
		console_scrollup () ;
		--console_row;
	}
}

/*----------------------------------------------------------------------*/

static inline void console_back (void)
{
	if (--console_col < 0)
	{
		console_col = CONSOLE_COLS-1 ;

		if (--console_row < 0) { console_row = 0; }
	}

	lcd_putc_xy (console_col * VIDEO_FONT_WIDTH, console_row * VIDEO_FONT_HEIGHT, ' ');
}

/*----------------------------------------------------------------------*/

void lcd_putc (const char c)
{
	if (!lcd_is_enabled)
	{
		serial_putc(c);
		return;
	}

	switch (c)
	{
	case '\r':	console_col = 0;
			return;

	case '\n':	console_newline();
			return;

	case '\t':	/* Tab (8 chars alignment) */
			console_col +=  8;
			console_col &= ~7;

			if (console_col >= CONSOLE_COLS)
			{
				console_newline();
			}
			return;

	case '\b':	console_back();
			return;

	default:	lcd_putc_xy (console_col * VIDEO_FONT_WIDTH, console_row * VIDEO_FONT_HEIGHT, c);
			if (++console_col >= CONSOLE_COLS)
			{
				console_newline();
			}
			return;
	}
	/* NOTREACHED */
}




/*----------------------------------------------------------------------*/

void lcd_puts (const char *s)
{
	if (!lcd_is_enabled)
	{
		serial_puts (s);
		return;
	}

	while (*s)
	{
		lcd_putc (*s++);
	}
}

/*----------------------------------------------------------------------*/

void lcd_printf(const char *fmt, ...)
{
	va_list args;
	char buf[CONFIG_SYS_PBSIZE];

	va_start(args, fmt);
	vsprintf(buf, fmt, args);
	va_end(args);

	lcd_puts(buf);
}

/************************************************************************/
/* ** Low-Level Graphics Routines					                    */
/************************************************************************/

static void lcd_drawchars (ushort x, ushort y, uchar *str, int count)
{
	uchar *dest;
	ushort off, row;

	dest = (uchar *)(lcd_base + y * lcd_line_length + x * (panel_info.vl_bpix) / 8);
	off  = x * (panel_info.vl_bpix) % 8;

	for (row=0;  row < VIDEO_FONT_HEIGHT;  ++row, dest += lcd_line_length)
	{
		uchar *s = str;
		uchar *d = dest;
		int i;

		for (i=0; i<count; ++i) {
			uchar c, bits;

			c = *s++;
			bits = video_fontdata[c * VIDEO_FONT_HEIGHT + row];

			#ifdef CONFIG_RGB16_FRAMEBUFFER
			for (c=0; c<VIDEO_FONT_WIDTH; ++c)
			{
				*d++ = (bits & 0x80) ? lcd_color_fg : lcd_color_bg;
				*d++ = (bits & 0x80) ? lcd_color_fg : lcd_color_bg;
				bits <<= 1;
			}
			#endif

			#ifdef CONFIG_RGB24_FRAMEBUFFER
			for (c=0; c<VIDEO_FONT_WIDTH; ++c)
			{
				*d++ = (bits & 0x80) ? lcd_color_fg : lcd_color_bg;
				*d++ = (bits & 0x80) ? lcd_color_fg : lcd_color_bg;
				*d++ = (bits & 0x80) ? lcd_color_fg : lcd_color_bg;
				bits <<= 1;
			}
			#endif
		}
	}
}

/*----------------------------------------------------------------------*/


static inline void lcd_puts_xy (ushort x, ushort y, uchar *s)
{
#if defined(CONFIG_LCD_LOGO) && !defined(CONFIG_LCD_INFO_BELOW_LOGO)
	lcd_drawchars (x, y+BMP_LOGO_HEIGHT, s, strlen ((char *)s));
#else
	lcd_drawchars (x, y, s, strlen ((char *)s));
#endif
}

/*----------------------------------------------------------------------*/

static inline void lcd_putc_xy (ushort x, ushort y, uchar c)
{
#if defined(CONFIG_LCD_LOGO) && !defined(CONFIG_LCD_INFO_BELOW_LOGO)
	lcd_drawchars (x, y+BMP_LOGO_HEIGHT, &c, 1);
#else
	lcd_drawchars (x, y, &c, 1);
#endif
}

/************************************************************************/
/**  Small utility to check that you got the colours right		*/
/************************************************************************/
#ifdef LCD_TEST_PATTERN

#define	N_BLK_VERT	2
#define	N_BLK_HOR	3

static int test_colors[N_BLK_HOR*N_BLK_VERT] = {
	CONSOLE_COLOR_RED,	CONSOLE_COLOR_GREEN,	CONSOLE_COLOR_YELLOW,
	CONSOLE_COLOR_BLUE,	CONSOLE_COLOR_MAGENTA,	CONSOLE_COLOR_CYAN,
};

static void test_pattern (void)
{
	ushort v_max  = panel_info.vl_row;
	ushort h_max  = panel_info.vl_col;
	ushort v_step = (v_max + N_BLK_VERT - 1) / N_BLK_VERT;
	ushort h_step = (h_max + N_BLK_HOR  - 1) / N_BLK_HOR;
	ushort v, h;
	uchar *pix = (uchar *)lcd_base;

	printf ("[LCD] Test Pattern: %d x %d [%d x %d]\n",
		h_max, v_max, h_step, v_step);

	/* WARNING: Code silently assumes 8bit/pixel */
	for (v=0; v<v_max; ++v) {
		uchar iy = v / v_step;
		for (h=0; h<h_max; ++h) {
			uchar ix = N_BLK_HOR * iy + (h/h_step);
			*pix++ = test_colors[ix];
		}
	}
}
#endif /* LCD_TEST_PATTERN */


/************************************************************************/
/* ** GENERIC Initialization Routines					*/
/************************************************************************/

int drv_lcd_init (void)
{
	debug ("[LCD] drv_lcd_init() called\n");

	struct stdio_dev lcddev;
	int rc = 0;

	lcd_base = (void *)(gd->fb_base);

#ifdef CONFIG_AM35XX_LCD
	lcd_line_length = (panel_info.vl_col * panel_info.vl_bpix) / 8;
#else
	lcd_line_length = (panel_info.vl_col * NBITS (panel_info.vl_bpix)) / 8;
#endif
	lcd_init (lcd_base);		/* LCD initialization */


	/* Device initialization */
	memset (&lcddev, 0, sizeof (lcddev));

	strcpy (lcddev.name, "lcd");
	lcddev.ext   = 0;			/* No extensions */
	lcddev.flags = DEV_FLAGS_OUTPUT;	/* Output only */
	lcddev.putc  = lcd_putc;		/* 'putc' function */
	lcddev.puts  = lcd_puts;		/* 'puts' function */

	rc = stdio_register (&lcddev);

	return (rc == 0) ? 1 : rc;
}

/*----------------------------------------------------------------------*/
static int lcd_clear (cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{

	debug ("[LCD] lcd_clear() called\n");


	/* Setting the palette */
	lcd_initcolregs();

#ifndef CONFIG_SYS_WHITE_ON_BLACK
	lcd_setfgcolor (CONSOLE_COLOR_BLACK);
	lcd_setbgcolor (CONSOLE_COLOR_WHITE);
#else
	lcd_setfgcolor (CONSOLE_COLOR_WHITE);
	lcd_setbgcolor (CONSOLE_COLOR_BLACK);
#endif	/* CONFIG_SYS_WHITE_ON_BLACK */

#ifdef	LCD_TEST_PATTERN
	test_pattern();
#else
	/* set framebuffer to background color */
	memset ((char *)lcd_base,
		COLOR_MASK(lcd_getbgcolor()),
		lcd_line_length*panel_info.vl_row);
#endif
	/* Paint the logo and retrieve LCD base address */
	debug ("[LCD] Drawing the logo...\n");
	lcd_logo ();

	console_col = 0;
	console_row = 0;

	return (0);
}

U_BOOT_CMD(
	cls,	1,	1,	lcd_clear,
	"clear screen",
	""
);

/*----------------------------------------------------------------------*/
/*
 * Enable or disable console output on lcd-display
 *
 * Author: Ludwig Zenz
 */
static int lcd_console_enable (cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
	debug ("[LCD] lcd_console_enable() called\n");

	if (lcd_is_enabled)	/* Toggle enable flag each call */
	{
		lcd_is_enabled = 0;
	}
	else
	{
		lcd_is_enabled = 1;
	}
	return (0);
}

U_BOOT_CMD(
	lcdcon,	1,	1,	lcd_console_enable,
	"enable or disable (toggle) console via lcd-display",
	""
);

/*----------------------------------------------------------------------*/

static int lcd_init (void *lcdbase)
{
	/* Initialize the lcd controller */
	debug ("[LCD] Initializing LCD framebuffer at %p\n", lcdbase);

	lcd_ctrl_init (lcdbase);
#ifdef CONFIG_AM35XX_LCD
	if (gd->dh_board_settings.wHWConfigFlags & SETTINGS_HW_LCD_MODE_FLAG)
	{
		lcd_is_enabled = 1;
	}
	else lcd_is_enabled = 0;
#else
	lcd_is_enabled = 1;
#endif

	lcd_clear (NULL, 1, 1, NULL);	/* dummy args */
	lcd_enable ();

	/* Initialize the console */
	console_col = 0;
#ifdef CONFIG_LCD_INFO_BELOW_LOGO
	console_row = 7 + BMP_LOGO_HEIGHT / VIDEO_FONT_HEIGHT;
#else
	console_row = 1;	/* leave 1 blank line below logo */
#endif

	return 0;
}


/************************************************************************/
/* ** ROM capable initialization part - needed to reserve FB memory	*/
/************************************************************************/
/*
 * This is called early in the system initialization to grab memory
 * for the LCD controller.
 * Returns new address for monitor, after reserving LCD buffer memory
 *
 * Note that this is running from ROM, so no write access to global data.
 */
ulong lcd_setmem (ulong addr)
{
	ulong size, realsize;
#ifdef CONFIG_AM35XX_LCD

	nand_info_t *nand;
	nand = &nand_info[nand_curr_device];		/* Get nand-info structure from current device */

	int line_length = (panel_info.vl_col * panel_info.vl_bpix) / 8;

	debug ("LCD panel info: %d x %d, %d bit/pix\n",	panel_info.vl_col, panel_info.vl_row, panel_info.vl_bpix );
#else
	int line_length = (panel_info.vl_col * NBITS (panel_info.vl_bpix)) / 8;

	debug ("LCD panel info: %d x %d, %d bit/pix\n",	panel_info.vl_col, panel_info.vl_row, NBITS (panel_info.vl_bpix) );
#endif



	realsize = line_length * panel_info.vl_row;

#ifdef CONFIG_AM35XX_LCD
	/* Round up to nearest full BLOCK */
	size = realsize / nand->writesize;
	size = (size+1) * nand->writesize;

#else
	/* Round up to nearest full page */
	size = (size + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1);
#endif

	/* Allocate pages for the frame buffer. */
	addr -= size;

	debug ("Reserving %ldk for LCD Framebuffer at: %08lx\n", size>>10, addr);

	return (addr);
}

/*----------------------------------------------------------------------
 * Set the foreground color
 */
static void lcd_setfgcolor (int color)
{
	debug ("[LCD] lcd_setfgcolor() called \n");

#if defined(CONFIG_ATMEL_LCD) | defined(CONFIG_AM35XX_LCD)
	lcd_color_fg = color;
#else
	lcd_color_fg = color & 0x0F;
#endif
}

/*----------------------------------------------------------------------
 * Set the background color
 */
static void lcd_setbgcolor (int color)
{
	debug ("[LCD] lcd_setbgcolor() called \n");

#if defined(CONFIG_ATMEL_LCD) | defined(CONFIG_AM35XX_LCD)
	lcd_color_bg = color;
#else
	lcd_color_bg = color & 0x0F;
#endif
}

/*----------------------------------------------------------------------*/

#ifdef	NOT_USED_SO_FAR
static int lcd_getfgcolor (void)
{
	return lcd_color_fg;
}
#endif	/* NOT_USED_SO_FAR */

/*----------------------------------------------------------------------*/

static int lcd_getbgcolor (void)
{
	return lcd_color_bg;
}

/*----------------------------------------------------------------------*/

/************************************************************************/
/* ** Chipset depending Bitmap / Logo stuff...                          */
/************************************************************************/
#ifdef CONFIG_LCD_LOGO
void bitmap_plot (int x, int y)
{
#ifdef CONFIG_ATMEL_LCD
	uint *cmap;
#else
	ushort *cmap;
#endif
	ushort i, j;
	uchar *bmap;
	uchar *fb;
	ushort *fb16;
#if defined(CONFIG_PXA250)
	struct pxafb_info *fbi = &panel_info.pxa;
#elif defined(CONFIG_MPC823)
	volatile immap_t *immr = (immap_t *) CONFIG_SYS_IMMR;
	volatile cpm8xx_t *cp = &(immr->im_cpm);
#endif

	debug ("Logo: width %d  height %d  colors %d  cmap %d\n",
		BMP_LOGO_WIDTH, BMP_LOGO_HEIGHT, BMP_LOGO_COLORS,
		(int)(sizeof(bmp_logo_palette)/(sizeof(ushort))));

	bmap = &bmp_logo_bitmap[0];
	fb   = (uchar *)(lcd_base + y * lcd_line_length + x);

#ifdef CONFIG_AM35XX_LCD
	if (panel_info.vl_bpix < 12)
#else
	if (NBITS(panel_info.vl_bpix) < 12)
#endif
	{
		/* Leave room for default color map */
#if defined(CONFIG_PXA250)
		cmap = (ushort *)fbi->palette;
#elif defined(CONFIG_MPC823)
		cmap = (ushort *)&(cp->lcd_cmap[BMP_LOGO_OFFSET*sizeof(ushort)]);
#elif defined(CONFIG_ATMEL_LCD)
		cmap = (uint *) (panel_info.mmio + ATMEL_LCDC_LUT(0));
#endif

		WATCHDOG_RESET();

		/* Set color map */
		for (i=0; i<(sizeof(bmp_logo_palette)/(sizeof(ushort))); ++i) {
			ushort colreg = bmp_logo_palette[i];
#ifdef CONFIG_ATMEL_LCD
			uint lut_entry;
#ifdef CONFIG_ATMEL_LCD_BGR555
			lut_entry = ((colreg & 0x000F) << 11) |
				    ((colreg & 0x00F0) <<  2) |
				    ((colreg & 0x0F00) >>  7);
#else /* CONFIG_ATMEL_LCD_RGB565 */
			lut_entry = ((colreg & 0x000F) << 1) |
				    ((colreg & 0x00F0) << 3) |
				    ((colreg & 0x0F00) << 4);
#endif
			*(cmap + BMP_LOGO_OFFSET) = lut_entry;
			cmap++;
#else /* !CONFIG_ATMEL_LCD */
#ifdef  CONFIG_SYS_INVERT_COLORS
			*cmap++ = 0xffff - colreg;
#else
			*cmap++ = colreg;
#endif
#endif /* CONFIG_ATMEL_LCD */
		}

		WATCHDOG_RESET();

		for (i=0; i<BMP_LOGO_HEIGHT; ++i) {
			memcpy (fb, bmap, BMP_LOGO_WIDTH);
			bmap += BMP_LOGO_WIDTH;
			fb   += panel_info.vl_col;
		}
	}
	else { /* true color mode */
		fb16 = (ushort *)(lcd_base + y * lcd_line_length + x);
		for (i=0; i<BMP_LOGO_HEIGHT; ++i) {
			for (j=0; j<BMP_LOGO_WIDTH; j++) {
				fb16[j] = bmp_logo_palette[(bmap[j])];
				}
			bmap += BMP_LOGO_WIDTH;
			fb16 += panel_info.vl_col;
		}
	}

	WATCHDOG_RESET();
}
#endif /* CONFIG_LCD_LOGO */

/*----------------------------------------------------------------------*/

#if defined(CONFIG_CMD_BMP) || defined(CONFIG_SPLASH_SCREEN)

/*
 * Display the BMP file located at address bmp_image.
 * Only uncompressed.
 */

#ifdef CONFIG_SPLASH_SCREEN_ALIGN
#define BMP_ALIGN_CENTER	0x7FFF
#endif

/*
 * Copy to
 */
int lcd_display_bitmap(ulong bmp_image, int x, int y)
{
	ushort i, j;
	uchar *fb;
	bmp_image_t *bmp=(bmp_image_t *)bmp_image;
	uchar *bmap;
	ushort padded_line;
	unsigned long bmp_width, bmp_height, byte_width;
	unsigned long panel_width = panel_info.vl_col;
	unsigned long panel_height = panel_info.vl_row;
	unsigned colors, bpix, bmp_bpix;
	unsigned long compression;

	if (!((bmp->header.signature[0]=='B') && (bmp->header.signature[1]=='M')))
	{
		printf ("[LCD] Error: no valid bmp image at %lx\n", bmp_image);
		return 1;
	}

	bmp_width 	= le32_to_cpu(bmp->header.width);
	bmp_height 	= le32_to_cpu(bmp->header.height);
	bmp_bpix 	= le16_to_cpu(bmp->header.bit_count);
	colors 		= 1 << bmp_bpix;
	compression = le32_to_cpu(bmp->header.compression);

	bpix = panel_info.vl_bpix; /* read panel configuration */

	if ((bpix != 8) && (bpix != 16) && (bpix != 24)) {
		printf ("[LCD] Error: %d bit/pixel mode, but BMP has %d bit/pixel\n", bpix, bmp_bpix);
		return 1;
	}

	if ((panel_info.vl_col < bmp_width) || (panel_info.vl_row < bmp_height))
	{
		printf ("[LCD] Error: BMP resolution is to high. Use %dx%d or smaller!\n",	panel_info.vl_col, panel_info.vl_row);
		return 1;
	}


	debug ("[LCD] Display-bmp: %d x %d  with %d colors\n", (int)bmp_width, (int)bmp_height, (int)colors);

	/*
	 *  BMP format for Monochrome assumes that the state of a
	 * pixel is described on a per Bit basis, not per Byte.
	 *  So, in case of Monochrome BMP we should align widths
	 * on a byte boundary and convert them from Bit to Byte
	 * units.
	 *  Probably, PXA250 and MPC823 process 1bpp BMP images in
	 * their own ways, so make the converting to be MCC200
	 * specific.
	 */


	padded_line = (bmp_width&0x3) ? ((bmp_width&~0x3)+4) : (bmp_width);

#ifdef CONFIG_SPLASH_SCREEN_ALIGN
	if (x == BMP_ALIGN_CENTER)
		x = max(0, (panel_info.vl_col - bmp_width) / 2);
	else if (x < 0)
		x = max(0, panel_info.vl_col - bmp_width + x + 1);

	if (y == BMP_ALIGN_CENTER)
		y = max(0, (panel_info.vl_row - bmp_height) / 2);
	else if (y < 0)
		y = max(0, panel_info.vl_row - bmp_height + y + 1);
#endif /* CONFIG_SPLASH_SCREEN_ALIGN */

	/* check if picture fits into panel - if not, display only the possible part */
	if ( (x + bmp_width)  > panel_width) { bmp_width = panel_width - x; }

	if ( (y + bmp_height) > panel_height) { bmp_height = panel_height - y; }

	/* get pointer to bmp data */
	bmap = (uchar *)bmp + le32_to_cpu (bmp->header.data_offset);


	/* get pointer of frame buffer */
	fb   = (uchar *) (lcd_base + (y + bmp_height - 1) * lcd_line_length + x * BYTES_PER_PIXEL);

	/* copy data into framebuffer - handle bits per pixel variants */
	switch (bmp_bpix)
	{
	case 1: /* pass through */
	case 8:

		if (bpix != 16)
			byte_width = bmp_width;
		else
			byte_width = bmp_width * 2;

		for (i = 0; (i < bmp_height) && (i < 1080); ++i)
		{
			WATCHDOG_RESET();
			for (j = 0; (j < bmp_width) && (j < 1920); j++)
			{

				#ifdef CONFIG_RGB16_FRAMEBUFFER
				*(fb)    = ((bmp->color_table[*(bmap)].blue) >> 3) & 0x1F ;
				*(fb++) |= ((bmp->color_table[*(bmap)].green)   << 3) & 0xE0;
				*(fb)    = ((bmp->color_table[*(bmap)].green) >> 5) & 0x07 ;
				*(fb++) |=  (bmp->color_table[*(bmap++)].red) & 0xF8;
				#endif

				#ifdef CONFIG_RGB24_FRAMEBUFFER
				*(fb++) = bmp->color_table[*(bmap)].blue;
				*(fb++) = bmp->color_table[*(bmap)].green;
				*(fb++) = bmp->color_table[*(bmap++)].red;
				#endif
			}
			bmap += (padded_line - bmp_width);
			fb   -= (bmp_width * BYTES_PER_PIXEL + lcd_line_length);
		}
		break;

	case 16:
		for (i = 0; (i < bmp_height) && (i < 1080); ++i)
		{
			WATCHDOG_RESET();

			for (j = 0; (j < bmp_width) && (j < 1920); j++)
			{
				#ifdef CONFIG_RGB16_FRAMEBUFFER
				*(fb++) = *(bmap++);
				*(fb++) = *(bmap++);
				#endif

				#ifdef CONFIG_RGB24_FRAMEBUFFER
				*(fb++) = (*(bmap) & 0x1F) << 3 ;
				*(fb) 	= ((*(bmap++) & 0xE0) >> 3);
				*(fb++)  |= ((*(bmap) & 0x07) << 5);
				*(fb++) = *(bmap++) & 0xF8;
				#endif
			}
			bmap += (padded_line - bmp_width) * 2;
			fb   -= (bmp_width * BYTES_PER_PIXEL + lcd_line_length);
		}
		break;

	case 24:
		for (i = 0; (i < bmp_height) && (i < 1080); ++i)
		{
			WATCHDOG_RESET();

			for (j = 0; (j < bmp_width) && (j < 1920); j++)
			{
				#ifdef CONFIG_RGB16_FRAMEBUFFER
				*(fb)    = (*(bmap++) >> 3) & 0x1F ;
				*(fb++) |= (*(bmap)   << 3) & 0xE0;
				*(fb)    = (*(bmap++) >> 5) & 0x07 ;
				*(fb++) |=  *(bmap++) & 0xF8;
				#endif

				#ifdef CONFIG_RGB24_FRAMEBUFFER
				*(fb++) = *(bmap++);
				*(fb++) = *(bmap++);
				*(fb++) = *(bmap++);
				#endif
			}

			bmap += (padded_line - bmp_width) * 3;
			fb   -= (bmp_width * BYTES_PER_PIXEL + lcd_line_length);
		}
		break;


	default:
		break;
	};

	return (0);
}
#endif

/*
 * write splashimage to framebuffer
 */
static void *lcd_logo (void)
{
#ifdef CONFIG_SPLASH_SCREEN
	char *s;
	ulong addr;
	static int do_splash = 1;

	if (do_splash && (s = getenv("rambufferaddr")) != NULL) {
		int x = 0, y = 0;
		do_splash = 0;

		addr = simple_strtoul (s, NULL, 16);
#ifdef CONFIG_SPLASH_SCREEN_ALIGN
		if ((s = getenv ("splashpos")) != NULL) {
			if (s[0] == 'm')
				x = BMP_ALIGN_CENTER;
			else
				x = simple_strtol (s, NULL, 0);

			if ((s = strchr (s + 1, ',')) != NULL) {
				if (s[1] == 'm')
					y = BMP_ALIGN_CENTER;
				else
					y = simple_strtol (s + 1, NULL, 0);
			}
		}
#endif /* CONFIG_SPLASH_SCREEN_ALIGN */

#ifdef CONFIG_VIDEO_BMP_GZIP
		bmp_image_t *bmp = (bmp_image_t *)addr;
		unsigned long len;

		if (!((bmp->header.signature[0]=='B') &&
		      (bmp->header.signature[1]=='M'))) {
			addr = (ulong)gunzip_bmp(addr, &len);
		}
#endif

		if (lcd_display_bitmap (addr, x, y) == 0) {
			return ((void *)lcd_base);
		}
	}
#endif /* CONFIG_SPLASH_SCREEN */

#ifdef CONFIG_LCD_LOGO
	bitmap_plot (0, 0);
#endif /* CONFIG_LCD_LOGO */

#ifdef CONFIG_LCD_INFO
	console_col = LCD_INFO_X / VIDEO_FONT_WIDTH;
	console_row = LCD_INFO_Y / VIDEO_FONT_HEIGHT;
	lcd_show_board_info();
#endif /* CONFIG_LCD_INFO */

#if defined(CONFIG_LCD_LOGO) && !defined(CONFIG_LCD_INFO_BELOW_LOGO)
	return ((void *)((ulong)lcd_base + BMP_LOGO_HEIGHT * lcd_line_length));
#else
	return ((void *)lcd_base);
#endif /* CONFIG_LCD_LOGO && !CONFIG_LCD_INFO_BELOW_LOGO */
}
/************************************************************************/
/************************************************************************/
