/*
 * File: omap_gpmc_prefetch.c
 *
 * (C) Copyright 2011-2012
 * Ludwig Zenz, DH electronics GmbH , lzenz@dh-electronics.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 *  Created on: 24.04.2012
 */

#include <common.h>
#include <asm/io.h>
#include <malloc.h>
#include <watchdog.h>
#include <linux/err.h>
#include <linux/mtd/compat.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <asm/arch/dma.h>

#include <asm/arch/omap_bch_soft.h>
#include <asm/arch/omap_gpmc_prefetch.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#include <asm/io.h>
#include <asm/errno.h>

/**
 * gpmc_prefetch_reset - disables and stops the prefetch engine
 */
int gpmc_prefetch_reset(int cs)
{
	unsigned int config1;

	/* check if the same module/cs is trying to reset */
	config1 = readl(GPMC_PREFETCH_CONFIG1);

	if (((config1 >> CS_NUM_SHIFT) & 0x7) != cs) return -EINVAL;

	/* Stop the PFPW engine */
	writel(0x0, GPMC_PREFETCH_CONTROL);

	/* Reset/disable the PFPW engine */
	writel(0x0, GPMC_PREFETCH_CONFIG1);

	return 0;
}


/**
 * gpmc_prefetch_enable - configures and starts prefetch transfer
 * @cs: cs (chip select) number
 * @fifo_th: fifo threshold to be used for read/ write
 * @dma_mode: dma mode enable (1) or disable (0)
 * @u32_count: number of bytes to be transferred
 * @is_write: prefetch read(0) or write post(1) mode
 */
int gpmc_prefetch_enable(int cs, int fifo_th, int dma_mode, unsigned int u32_count)
{

	if (fifo_th > PREFETCH_FIFOTHRESHOLD_MAX)
	{
		printf("[NAND]: ERROR -> FiFothreshold is max %d", PREFETCH_FIFOTHRESHOLD_MAX);
		return -1;
	}
	else if (!(readl(GPMC_PREFETCH_CONTROL)))
	{
		/* Set the amount of bytes to be prefetched */
		writel(u32_count, GPMC_PREFETCH_CONFIG2);

		/* Set dma/mpu mode, the prefetch read / post write and
		 * enable the engine. Set which cs is has requested for.
		 */
		writel( ((cs << CS_NUM_SHIFT) | PREFETCH_FIFOTHRESHOLD(fifo_th) | ENABLE_PREFETCH | (dma_mode << DMA_MPU_MODE) ) , GPMC_PREFETCH_CONFIG1);

		/*  Start the prefetch engine */
		writel(0x1, GPMC_PREFETCH_CONTROL);
	}
	else
	{
		printf("[NAND]: ERROR -> Prefetch-Engine is busy!!");
		return -1; /* Busy */
	}

	return 0;
}


/**
 * nand_read_buf_dma_prefetch - [OMAP3 specific] read chip data into buffer
 * @mtd:	MTD device structure
 * @buf:	buffer to store date
 * @len:	number of bytes to read
 *
 * @author: LZ
 *
 * Read function for 8bit buswith.
 */
inline void nand_read_buf_dma_prefetch(struct mtd_info *mtd, uint8_t *buf, int len)
{
	struct dma4_chan cfg;
	int ret;
	int dma_cha = 0x0;


	debug("[NAND]: configure DMA ...\n");
	omap3_dma_init(dma_cha);

	/* setup DMA channel */
	omap3_dma_get_conf_chan(dma_cha, &cfg);

	cfg.csdp = CSDP_DATA_TYPE_32BIT | CSDP_SRC_BURST_EN_64BYTES | CSDP_DST_BURST_EN_64BYTES
			| CSDP_DST_ENDIAN_LOCK_ADAPT | CSDP_DST_ENDIAN_LITTLE | CSDP_SRC_ENDIAN_LOCK_ADAPT
			| CSDP_SRC_ENDIAN_LITTLE;

	cfg.cfn = (len >> ELEMENTS_32BIT_IN_FIFO_SHIFTER); /* number of frames to be transfered (64 bytes per frame )*/

	cfg.ccr = CCR_FRAME_SYNC | CCR_READ_PRIORITY_HIGH /* not sure about this */ | CCR_SRC_AMODE_CONSTANT
			/*| | CCR_CONST_FILL_ENABLE*/ /* not sure about this */ | CCR_DST_AMODE_POST_INC
			/* | CCR_PREFETCH */ /* NOT to enable */ | CCR_SYNC_SOURCE;


	/* DMA_SYNCHRO_CONTROL_UPPER depends on the channel number */
	cfg.ccr &= ~((0x03 << 19) | 0x1f);
	cfg.ccr |= (4 & 0x1f) ;

	cfg.cicr = (1 << 5) | (1 << 11) | (1 << 10) | (1 << 8); /* Channel Interrupt Control Register */

	ret = omap3_dma_conf_chan(dma_cha, &cfg);
	if (ret != 0)
		printf("[NAND]: ERROR during DMA channel configuration");

	/* setup DMA transfer */
	ret = omap3_dma_conf_transfer(dma_cha, (uint32_t *)NAND_CS0_FIFO, (uint32_t *)buf, ELEMENTS_32BIT_IN_FIFO);


	/*  configure and start prefetch transfer */
	ret = gpmc_prefetch_enable(0x0, PREFETCH_FIFOTHRESHOLD_MAX, 0x1, len);
	if (ret)
		printf("[NAND]: ERROR -> PFPW engine is busy");


	/* start DMA transfer from Prefetch-FIFO to buf */
	ret = omap3_dma_start_transfer(dma_cha);
	if (ret != 0)
		printf("[NAND]: ERROR -> could not start DMA transfer");

	/* wait for DMA transfer to finish */
	ret = omap3_dma_wait_for_transfer(dma_cha);
	if (ret != 0)
		printf("[NAND]: ERROR during waiting for DMA transfer to finish");

	/* disable and stop the PFPW engine */
	gpmc_prefetch_reset(0x0);

	return;
}

