/*
 * (C) Copyright 2008-2009 Freescale Semiconductor, Inc.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <common.h>
#include <spi.h>
#include <spi_flash.h>
#include <asm/errno.h>
#include <linux/types.h>
#include <malloc.h>

#include <asm/arch/am35x_spi_nor.h>

static u8 g_tx_buf[32];
static u8 g_rx_buf[32];

extern void spi_io_set_WP_pin(void);

#define WRITE_ENABLE(a)			 spi_nor_cmd_1byte(a, WREN)
#define ENABLE_WRITE_STATUS(a)	 spi_nor_cmd_1byte(a, EWSR)

struct spi_flash_params {
	u8		idcode1;
	u32		block_size;
	u32		block_count;
	u32		device_size;
	const char	*name;
};

struct am35x_spi_flash {
	const struct spi_flash_params *params;
	struct spi_flash flash;
};

static inline struct am35x_spi_flash *
to_am35x_spi_flash(struct spi_flash *flash)
{
	return container_of(flash, struct am35x_spi_flash, flash);
}

static const struct spi_flash_params spi_flash_table[] = {
	{
		.idcode1		= 0x27,
		.block_size		= SZ_64K,
		.block_count	= 64,
		.device_size	= SZ_64K * 64,
		.name			= "Atmel AT25DF321A - 4MB",
	},
	{
		.idcode1		= 0x20,
		.block_size		= SZ_64K,
		.block_count	= 64,
		.device_size	= SZ_64K * 64,
		.name			= "Numonyx M25P32 - 4MB",
	},
	{
		.idcode1		= 0xBF,
		.block_size		= SZ_64K,
		.block_count	= 64,
		.device_size	= SZ_64K * 64,
		.name			= "SST SST25VF032B - 4MB",
	},
};

static s32 spi_nor_flash_query(struct spi_flash *flash, void* data)
{
	u8 au8Tmp[4] = { 0 };
	u8 *pData = (u8 *)data;

	g_tx_buf[0] = JEDEC_ID;

	if (spi_xfer_dh(flash->spi, (4 << 3), g_tx_buf, au8Tmp, SPI_XFER_BEGIN | SPI_XFER_END  ))
	{
		return -1;
	}

	printf("JEDEC ID: 0x%02x:0x%02x:0x%02x\n",
			au8Tmp[2], au8Tmp[1], au8Tmp[0]);

	pData[0] = au8Tmp[2];
	pData[1] = au8Tmp[1];
	pData[2] = au8Tmp[0];

	return 0;
}

static s32 spi_nor_status(struct spi_flash *flash)
{
    u32 *cmd = (u32 *)g_tx_buf;
	*cmd = (RDSR << 24);

	if (spi_xfer_dh(flash->spi, 2 << 3, g_tx_buf, g_rx_buf, SPI_XFER_BEGIN | SPI_XFER_END  ) != 0)
	{
		printf("Error: %s(): %d\n", __func__, __LINE__);
		return 0;
	}
	return g_rx_buf[0];
}

/*!
 * Erase a block_size data from block_addr offset in the flash
 */
static int spi_nor_erase_page(struct spi_flash *flash, void *page_addr)
{
	u32 addr = (u32)page_addr;

	if ((addr & 512) != 0) {
		printf("Error - page_addr is not "
				"512 Bytes aligned: %p\n",
				page_addr);
		return -1;
	}

	/* now do the block erase */
	if (spi_xfer_dh(flash->spi, 4 << 3, g_tx_buf, g_rx_buf,
			SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) {
		return -1;
	}

	while (spi_nor_status(flash) & RDSR_BUSY)
		;

	return 0;
}

static int spi_nor_flash_read(struct spi_flash *flash, u32 offset,
		size_t len, void *buf)
{
	struct am35x_spi_flash *am35x_sf = to_am35x_spi_flash(flash);
	u32 *cmd = (u32 *)g_tx_buf;
	u32 max_rx_sz = (MAX_SPI_BYTES) - 4;
	u8 *d_buf = (u8 *)buf;
	u8 *s_buf;
	s32 s32remain_size = len;
	int i;

	if (!(flash->spi))
		return -1;

	printf("Reading SPI NOR flash 0x%x [0x%x bytes] -> ram 0x%p\n",
		offset, len, buf);
	debug("%s(from flash=0x%08x to ram=%p len=0x%x)\n",
		__func__,
		offset, buf, len);

	if (len == 0)
		return 0;

	*cmd = (READ << 24) | ((u32)offset & 0x00FFFFFF);

	for (; s32remain_size > 0;
			s32remain_size -= max_rx_sz, *cmd += max_rx_sz) {
		debug("Addr:0x%p=>Offset:0x%08x, %d bytes transferred\n",
				d_buf,
				(*cmd & 0x00FFFFFF),
				(len - s32remain_size));
		debug("%d%% completed\n", ((len - s32remain_size) * 100 / len));

		if (s32remain_size < max_rx_sz) {
			debug("100%% completed\n");

			if (spi_xfer_dh(flash->spi, (s32remain_size + 4) << 3,
				g_tx_buf, g_rx_buf,
				SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) {
				printf("Error: %s(%d): failed\n",
					__FILE__, __LINE__);
				return -1;
			}
			/* throw away 4 bytes (5th received bytes is real) */
			s_buf = g_rx_buf + 4;

			/* now adjust the endianness */
			for (i = s32remain_size; i >= 0; i -= 4, s_buf += 4) {
				if (i < 4) {
					if (i == 1) {
						*d_buf = s_buf[0];
					} else if (i == 2) {
						*d_buf++ = s_buf[1];
						*d_buf++ = s_buf[0];
					} else if (i == 3) {
						*d_buf++ = s_buf[2];
						*d_buf++ = s_buf[1];
						*d_buf++ = s_buf[0];
					}
					printf("SUCCESS\n\n");
					return 0;
				}
				/* copy 4 bytes */
				*d_buf++ = s_buf[3];
				*d_buf++ = s_buf[2];
				*d_buf++ = s_buf[1];
				*d_buf++ = s_buf[0];
			}
		}

		/* now grab max_rx_sz data (+4 is
		*needed due to 4-throw away bytes */
		if (spi_xfer_dh(flash->spi, (max_rx_sz + 4) << 3,
			g_tx_buf, g_rx_buf,
			SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) {
			printf("Error: %s(%d): failed\n", __FILE__, __LINE__);
			return -1;
		}
		/* throw away 4 bytes (5th received bytes is real) */
		s_buf = g_rx_buf + 4;
		/* now adjust the endianness */
		for (i = 0; i < max_rx_sz; i += 4, s_buf += 4) {
			*d_buf++ = s_buf[3];
			*d_buf++ = s_buf[2];
			*d_buf++ = s_buf[1];
			*d_buf++ = s_buf[0];
		}

		if ((s32remain_size % am35x_sf->params->block_size) == 0)
			printf(".");
	}
	printf("SUCCESS\n\n");

	return -1;
}

static int spi_nor_flash_write(struct spi_flash *flash, u32 offset,
		size_t len, const void *buf)
{
	u8 *s_buf = (u8 *)buf;
	int trans_bytes = 0;
    int page_trans_bytes = 0;
    int bytes_to_transfer = 0;
	int i = 0;
	int m_page_trans_bytes = 0;
    int m_len = 0;
    int page_align_diff = 0;
    u32 *cmd = (u32 *)g_tx_buf;
    int act_addr = 0;
    
	if (!(flash->spi))
		return -1;

	if (len == 0)
		return 0;

	printf("Writing SPI NOR flash 0x%x [0x%x bytes] <- ram 0x%p\n",
		offset, len, buf);
	debug("%s(flash addr=0x%08x, ram=%p, len=0x%x)\n",
			__func__, offset, buf, len);

    /* Set Write protect pin */
    spi_io_set_WP_pin();
    
    for(trans_bytes = 0; trans_bytes < (int)len;)
    {
        m_len = len;
        
        // Offset is not page size aligned
        page_align_diff = offset - ((offset / SPI_FLASH_PAGE_SIZE) * SPI_FLASH_PAGE_SIZE);
        if(page_align_diff > 0)
        {
            len = SPI_FLASH_PAGE_SIZE - page_align_diff;
            if( len > m_len)
            {
                len = m_len;
            }
        }
        else
        { 
            if((len - trans_bytes) >= SPI_FLASH_PAGE_SIZE)
            {
                len = SPI_FLASH_PAGE_SIZE;
            }
            else
            {
                len = m_len - trans_bytes;
            }
        }
        
        for(page_trans_bytes = 0, act_addr = offset; page_trans_bytes < (int)len;)
        {
            /* Send Write Enable instruction */
            *cmd = (WREN << 24);
            if (spi_xfer_dh(flash->spi, 1 << 3, g_tx_buf, g_rx_buf,
                    SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) {
                printf("Error: %s(%d): failed\n", __FILE__, __LINE__);
                return -1;
            }

            /* Check if Write enable bit ist set */
            if (!(spi_nor_status(flash) & STAT_WEL))
            {
                printf("SPI Error: Can't set Write Enable Latch\n");
                return -1;
            }
            
            /* Wait till SPI Flash status is busy */
            while ((spi_nor_status(flash) & STAT_WIP));       
        
            // Set address and opcode
            g_tx_buf[0] = (act_addr & 0xFF);
            g_tx_buf[1] = (act_addr >> 8) & 0xFF;
            g_tx_buf[2] = (act_addr >> 16) & 0xFF;
            g_tx_buf[3] = PAGE_PROG; /*Opcode */        
        

            // Set Write data
            if((int)len - page_trans_bytes >= 28)
            {
                for (i = 4; i < (28 + 4); i += 4) 
                {
                    g_tx_buf[i + 3] = s_buf[trans_bytes++];
                    g_tx_buf[i + 2] = s_buf[trans_bytes++];
                    g_tx_buf[i + 1] = s_buf[trans_bytes++];
                    g_tx_buf[i] = s_buf[trans_bytes++];  
                    page_trans_bytes += 4;
                }     

                if (spi_xfer_dh(flash->spi, 32 << 3, g_tx_buf, g_rx_buf,
                             SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) {
                    printf("Error: %s(%d): failed\n", __FILE__, __LINE__);
                    return -1;
                }
                
                act_addr = act_addr + 28;
            } 
            else
            {
                m_page_trans_bytes = page_trans_bytes;
                for (bytes_to_transfer = 0, i = 4; i < ((int)len - m_page_trans_bytes + 4); i += 4) 
                {
                    if(((int)len - page_trans_bytes) < 4)
                    {
                        switch((int)len - page_trans_bytes)
                        {
                            case 1:
                                g_tx_buf[i + 3] = s_buf[trans_bytes++];
                                page_trans_bytes += 1;
                                bytes_to_transfer += 1;
                            break;
                             
                            case 2:
                                g_tx_buf[i + 3] = s_buf[trans_bytes++];
                                g_tx_buf[i + 2] = s_buf[trans_bytes++];
                                page_trans_bytes += 2;
                                bytes_to_transfer +=2;
                            break;
                             
                            case 3:
                                g_tx_buf[i + 3] = s_buf[trans_bytes++];
                                g_tx_buf[i + 2] = s_buf[trans_bytes++];
                                g_tx_buf[i + 1] = s_buf[trans_bytes++];
                                page_trans_bytes += 3;
                                bytes_to_transfer += 3;
                            break;
                        }
                    }                  
                    else
                    {
                        g_tx_buf[i + 3] = s_buf[trans_bytes++];
                        g_tx_buf[i + 2] = s_buf[trans_bytes++];
                        g_tx_buf[i + 1] = s_buf[trans_bytes++];
                        g_tx_buf[i] = s_buf[trans_bytes++];  
                        page_trans_bytes += 4;
                        bytes_to_transfer += 4;
                    }
                }     

                if (spi_xfer_dh(flash->spi, (bytes_to_transfer + 4) << 3, g_tx_buf, g_rx_buf,
                             SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) {
                    printf("Error: %s(%d): failed\n", __FILE__, __LINE__);
                    return -1;
                }
                
                act_addr = act_addr + (int)len - m_page_trans_bytes;
            }
                    
            /* Wait till SPI Flash status is busy */
            while ((spi_nor_status(flash) & STAT_WIP));   

        }
        
        offset += (int)len;
        len = m_len;
    }
    
	printf("SUCCESS\n\n");

	return 0;    
}

static int spi_nor_flash_erase(struct spi_flash *flash, u32 offset,
		size_t len)
{
    u32 *cmd = (u32 *)g_tx_buf;
    u32 i;
    struct am35x_spi_flash *am35x_sf = to_am35x_spi_flash(flash);
    
    if(len != am35x_sf->params->block_size * (int)(len / am35x_sf->params->block_size))
    {
        printf("Error: SPI erase len is not block aligned\n");
        return -1;
    }
    
    /* Set Write protect pin */
    spi_io_set_WP_pin();    

    for(i = 1; i <= (len / am35x_sf->params->block_size); i++, offset=offset+am35x_sf->params->block_size)
    {
        /* Send Write Enable instruction */
        *cmd = (WREN << 24);
        if (spi_xfer_dh(flash->spi, 1 << 3, g_tx_buf, g_rx_buf,
                SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) {
            printf("Error: %s(%d): failed\n", __FILE__, __LINE__);
            return -1;
        }

        /* Check if Write enable bit ist set */
        if (!(spi_nor_status(flash) & STAT_WEL))
        {
            printf("SPI Error: Can't set Write Enable Latch\n");
            return -1;
        }
        
        /* Wait till SPI Flash status is busy */
        while ((spi_nor_status(flash) & STAT_WIP));      
    
        // Set address and opcode
        g_tx_buf[0] = (offset & 0xFF);
        g_tx_buf[1] = (offset >> 8) & 0xFF;
        g_tx_buf[2] = (offset >> 16) & 0xFF;
        g_tx_buf[3] = ERASE_64K; /*Opcode */        

        if (spi_xfer_dh(flash->spi, 4 << 3,
                g_tx_buf, g_rx_buf,
                SPI_XFER_BEGIN | SPI_XFER_END  ) != 0) 
        {
            printf("Error: %s(%d): failed\n",__FILE__, __LINE__);
            return -1;
        }  
        
        /* Wait till SPI Flash status is busy */
        while ((spi_nor_status(flash) & STAT_WIP));  
    }
    
    printf("SUCCESS\n\n");
    
	return 0;
}

struct spi_flash *spi_flash_probe(unsigned int bus, unsigned int cs, unsigned int max_hz, unsigned int spi_mode)
{
	struct spi_slave *spi = NULL;
	const struct spi_flash_params *params = NULL;
	struct am35x_spi_flash *am35x_sf = NULL;
	u8  idcode[4] = { 0 };
	u32 i = 0;
	s32 ret = 0;

	if (CONFIG_SPI_FLASH_CS != cs) {
		printf("Invalid cs for SPI NOR.\n");
		return NULL;
	}

	spi = spi_setup_slave(bus, cs, max_hz, spi_mode);

	if (!spi) {
		debug("SF: Failed to set up slave\n");
		return NULL;
	}

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("SF: Failed to claim SPI bus: %d\n", ret);
		goto err_claim_bus;
	}

	am35x_sf = (struct am35x_spi_flash *)malloc(sizeof(struct am35x_spi_flash));

	if (!am35x_sf) {
		debug("SF: Failed to allocate memory\n");
		spi_free_slave(spi);
		return NULL;
	}

	am35x_sf->flash.spi = spi;

	/* Read the ID codes */
	ret = spi_nor_flash_query(&(am35x_sf->flash), idcode);
	if (ret)
		goto err_read_id;

	for (i = 0; i < ARRAY_SIZE(spi_flash_table); ++i) {
		params = &spi_flash_table[i];
		if (params->idcode1 == idcode[1])
			break;
	}

	if (i == ARRAY_SIZE(spi_flash_table)) {
		debug("SF: Unsupported DataFlash ID %02x\n",
				idcode[1]);

		goto err_invalid_dev;
	}

	am35x_sf->params = params;

	am35x_sf->flash.name = params->name;
	am35x_sf->flash.size = params->device_size;

	am35x_sf->flash.read  = spi_nor_flash_read;
	am35x_sf->flash.write = spi_nor_flash_write;
	am35x_sf->flash.erase = spi_nor_flash_erase;

	debug("SF: Detected %s with block size %lu, "
			"block count %lu, total %u bytes\n",
			params->name,
			params->block_size,
			params->block_count,
			params->device_size);

	return &(am35x_sf->flash);

err_read_id:
	spi_release_bus(spi);
err_invalid_dev:
	if (am35x_sf)
		free(am35x_sf);
err_claim_bus:
	if (spi)
		spi_free_slave(spi);
	return NULL;
}

void spi_flash_free(struct spi_flash *flash)
{
	struct am35x_spi_flash *am35x_sf = NULL;

	if (!flash)
		return;

	am35x_sf = to_am35x_spi_flash(flash);

	if (flash->spi) {
		spi_free_slave(flash->spi);
		flash->spi = NULL;
	}

	free(am35x_sf);
}

